/** @file  playlist.c
 *  @brief context-menu action "make playlist"
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <fcntl.h>      /* creat() */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "gui/common/session.h"
#include "gui/common/alert.h"      /* form_*() */
#include "meta.h"       /* struct _meta */
#include "util.h"       /* estring_*() */
#include "action.h"     /* action_file_dst() */
#include "file.h"       /* file_exists() */
#include "mime.h"       /* MIME_DELIMITER */



/******************************************************************************/



/**
 * context-menu action: make playlist
 * @return an error-code
 * @param serial    the serial number of the session that called us
 * @param filename  the file to act on
 * @param md        the file's meta-data
 * @param ac        the context of this action (selection et al.)
 */
int efun(long serial,char *filename,meta **md,action_context *ac) {
  noir_sess   *ns=noir_sess_get_by_serial(serial);
  int          ret=NOIR_FAIL;
  action_file *sf;
  estring     *msg=estring_new(2048);
  char        *data,*path;
  const char  *mime;

  if(!filename)
    return NOIR_FAIL;

  if(!(path=strdup(filename)))
    return NOIR_OOM;

  if(!msg) {
    free(path);
    return NOIR_OOM; }



  /* create playlist */
  while((sf=ac->selection)) { /* get next item in selection */
    ac->selection=sf->next;   /* remove item from selection */

    mime=NULL;                /* get its MIME-type */
    if(sf->md) {
      meta *m=meta_in_group(&sf->md,"MIME",NULL);
      if(m)
        mime=m->val; }
    if(!mime)
      mime=file_get_mime(sf->tag);

    if(mime) {
      char   *p=strchr(mime,MIME_DELIMITER);
      size_t  cl=strlen(mime);
      if(p)
        cl-=(p-mime);

      if(strncasecmp(mime,"audio/",cl))  /* if it's not audio/..., ignore it */
        E(3,"ignoring \"%s\" (%s)\n",sf->tag,mime);
      else {                             /* otherwise, add it */
        E(2,"adding   \"%s\" (%s)\n",sf->tag,mime);

        if((p=strrchr(sf->tag,DIR_DELIMITER)))  /* make path relative */
          p++;
        else
          p=sf->tag;

        estring_appendf(msg,"%s\n",p); }}       /* add relative path */

    action_file_dst(sf); }



  if(estring_length(msg)&&(data=ESTRING_GET_CSTRING(msg))) {  /* do we have a playlist? */
    char *fn=NULL,*p=strrchr(path,DIR_DELIMITER);
    if(p) {
      *p='\0';
      if((p=strrchr(path,DIR_DELIMITER))) {
        mode_t     mode=S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH;
        file_info *pa=file_info_fetch(NULL,path);

        if(pa&&pa->intrinsic)   /* inherit permissions from dir if possible */
          mode=pa->intrinsic->mode&mode;

        esprintf(&fn,"%s/%s.m3u",path,p+1);

        if(fn) {
          E(1,"playlist \"%s\"\n",fn);

          if(file_exists(fn)&&(form_yes_no(ns,"Playlist already exists, overwrite?")!=NOIR_SUCC))
            ;
          else if(file_save(fn,mode,estring_length(msg),data)!=NOIR_SUCC)
            form_alert(ns,"Couldn't save %s...",fn);

          free(fn); }}}}
  else
    ret=NOIR_IGNORED;

  estring_free(msg,TRUE);
  free(path);

  return ret; }



/******************************************************************************/
