/** @file  evidence.c
 *  @brief Evidence. GTK2 file-manager. by Azundris. GPL applies.
 */

#include "config.h"

#include "gui/common/session_private.h"
#include "gui/common/session.h"
#include "gui/common/typebuffer.h"
#include "gui/common/menu.h"
#include "gui/common/desktop.h"

#ifdef HAVE_EVAS2
#  include "gui/gtk2/icon_view/gevas_light.h"
#  include "gui/gtk2/icon_view/gevas_bind.h"
#  include "gui/efl/common/theme.h"
#endif

#ifdef HAVE_GTK2
#  include <gtk/gtk.h>

#  include "gui/gtk2/common/icon.h"      /* icon_init(), icon_exit(), icon_path_load() */
#  include "gui/gtk2/common/shelf.h"     /* struct noir_shelf */

#  include "gui/gtk2/tree_view/tree_view.h"      /* struct noir_treeview, tree_model_setup() */
#  include "gui/gtk2/browser_view/browser_view.h"   /* struct noir_browser */
#  include "gui/gtk2/icon_view/icon_view.h"      /* struct noir_iconview */

#  include "gui/gtk2/common/notebook.h"       /* notebook_setup() */
#  include "gui/gtk2/about/about.h"          /* about_init() */
#  include "gui/gtk2/common/dnd.h"            /* dnd_init() */

#  include "gui/gtk2/common/tooltip.h"        /* info_tooltip_cancel() */

#  include "gui/gtk2/common/clipboard.h"      /* selection_update(), clipboard_get/set() */
#endif

#include "gui/common/alert.h"          /* form_alert(), form_yes_no(), form_dial(), info_init(), info_exit() */

#include <stdlib.h>         /* getenv() */
#include <string.h>
#include <stdio.h>

#include <gdk/gdkx.h>

#ifdef HAVE_ECORE
#  include <Ecore.h>
#endif

#include "ipc_main.h"
#include "keysyms.h"        /* GDK_* */
#include "evidence_private.h"
#include "file.h"           /* file_*() */
#include "model.h"
#include "thumbnailer/thumbnail.h"      /* thumbnail_init/exit() */

#include "util.h"           /* now() */
#include "setup.h"          /* paths_init(), get_setup_file() */
#include "gui/common/options.h"   /* view_name[], parse_options() */

#include "gui/common/menu.h"           /* menu_root_context(), menu_lib_gc() */
#include "mime.h"           /* MIME_INTL_MKD */



q_head           *queue=NULL;

char             *home=NULL,
                 *temp=NULL,
                 *appName=PACKAGE;  /* FAM wants this spelling */
static char       cwd[PATH_MAX+1];
int               go2view_mapping[START_IN_CREDITS+1];
int               debug=0;

pthread_t         main_thread;



/*****************************************************************************/
/* X helpers */



#ifdef HAVE_X

/**
 * get X display.
 */
Display *get_display(void) {
  return GDK_DISPLAY(); }



/**
 * get X root window.
 */
Window get_root_window(void) {
  return GDK_ROOT_WINDOW(); }



/**
 * get X window from GTK-widget.
 * @param w - the widget
 */
Window get_widget_window(void *w) {
  return GDK_WINDOW_XWINDOW(((GtkWidget *)w)->window); }

#endif






/*****************************************************************************/
/* path menu -- a menu containing all paths leading up to this one */



/**
 * a path was selected from path-menu, cd to it.
 * @param item - the selected entry, "data" prop holds index into session->cwd
 * @param data - the session
 */
static gint path_menu_selected(GtkWidget *item,gpointer data) {
  noir_sess *ns=(noir_sess *)data;
  gint       ret;
  char      *p;

  if(!ns||!(p=noir_sess_get_cwd(ns))||!*p)
    return FALSE;
  if(!(p=strdup(p)))
    return FALSE;

  ret=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(item),"data"));

  if(ret) {
    char *x=p;
    while(ret--) {
      if(!(x=strchr(x+1,DIR_DELIMITER)))
        x=p+strlen(p); }
    *x='\0'; }
  else
    p[1]='\0';

  noir_sess_focus(ns,p);                       /* update views */

  free(p);
  return TRUE; }



/**
 * add a path to the path-menu.
 * @param ns - session
 * @param menu - the GTK menu
 * @param path - the path
 * @param el - the index into session->cwd ("data" prop, cf. path_menu_selected())
 */
static void path_menu_add_path(noir_sess *ns,GtkWidget *menu,char *path,int el) {
  GtkWidget *item=gtk_menu_item_new_with_label(path);
  g_object_set_data(G_OBJECT(item),"data",GINT_TO_POINTER(el));
  g_signal_connect(item,"activate",G_CALLBACK(path_menu_selected),ns);
  gtk_menu_shell_append((GtkMenuShell *)menu,item); }



/**
 * show menu containing all paths leading up to cwd.
 * @param ns - session
 * @param button - what mouse-button caused the menu to open?
 * @param when - when did the click occur?
 */
int menu_path(noir_sess *ns,int button,long when) {
  char *p,*path;
  char  s;
  GtkWidget *menu;
  int el=0;

  if(!ns||!(p=noir_sess_get_cwd(ns))||!*p)
    return NOIR_FAIL;
  if(!(path=strdup(p)))
    return NOIR_OOM;

  menu=gtk_menu_new();
  p=path;
  while((p=strchr(p,'/'))) {
    if(p==path)
      p++;
    s=*p;
    *p='\0';
    path_menu_add_path(ns,menu,path,el++);
    *(p++)=s; }

  if(el>1)
    path_menu_add_path(ns,menu,path,el);
  gtk_widget_show_all(menu);
  gtk_menu_popup(GTK_MENU(menu),NULL,NULL,NULL,NULL,button,when);

  free(path);
  return NOIR_SUCC; }






/*****************************************************************************/



/**
 * callback for GTK notebook's "switch-page" signal.
 * @param notebook - the GTK notebook, or NULL
 * @param page_ptr - the GTK notebook page, or NULL
 * @param page_index - the page we're switching to
 * @param data - the session
 */
static int go2view_activate_page(GtkNotebook *notebook,GtkNotebookPage *page_ptr,
                                 guint page_index,gpointer data) {
  noir_sess     *ns=(noir_sess *)data;
  noir_treeview *tv=(noir_treeview *)noir_sess_client_get_data(ns,CLIENT_TREEVIEW);
  noir_iconview *iv=(noir_iconview *)noir_sess_client_get_data(ns,CLIENT_ICONVIEW);
  noir_browser  *bv=(noir_browser *) noir_sess_client_get_data(ns,CLIENT_BROWSERVIEW);
  client_funs   *cf=NULL;
  int            page_no=(int)page_index;

  E(2,"page no. %d, %s\n",page_no,
    ((page_no>=0)&&(page_no<4))?view_name[page_no]:"???");

  if(tv)  tv->client->flags|=CF_PAUSED;
  if(bv)  bv->client->flags|=CF_PAUSED;
  if(iv)  iv->client->flags|=CF_PAUSED;

  if(page_no==go2view_mapping[START_IN_CREDITS])
    return NOIR_SUCC;

  if(page_no==go2view_mapping[START_IN_TREEVIEW])
    cf=tv?tv->client:NULL;
  else if(page_no==go2view_mapping[START_IN_BROWSER])
    cf=bv?bv->client:NULL;
  else if(page_no==go2view_mapping[START_IN_ICONVIEW])
    cf=iv?iv->client:NULL;
  else
    E(0,"unknown page #%d!?\n",page_no);

  ns->client=cf;
  if(cf) {
    cf->flags&=(~CF_PAUSED);
    if(cf->set_selection)
      cf->set_selection((void *)ns); }

  return noir_sess_focus(ns,noir_sess_get_cwd(ns)); }



/**
 * setup go2view sub-system.
 * @param ns - the session
 */
void go2view_setup(noir_sess *ns) {
#ifdef WITH_TREEVIEW
  noir_treeview *tv=(noir_treeview *)noir_sess_client_get_data(ns,CLIENT_TREEVIEW);
#endif
#ifdef WITH_ICONVIEW
  noir_iconview *iv=(noir_iconview *)noir_sess_client_get_data(ns,CLIENT_ICONVIEW);
#endif
#ifdef WITH_BROWSERVIEW
  noir_browser  *bv=(noir_browser *) noir_sess_client_get_data(ns,CLIENT_BROWSERVIEW);
#endif
  int            c;

  for(c=0;c<=START_IN_CREDITS;c++)
    go2view_mapping[c]=-1;

  c=0;

#ifdef WITH_TREEVIEW
  if(tv&&tv->tree_view) {
    go2view_mapping[START_IN_TREEVIEW]=c++;
    E(2,"tree-view enabled...  (%d)\n",c-1); }
  else
    E(0,"could not enable tree-view... (%p,%p)\n",tv,tv?tv->tree_view:NULL);
#endif

#ifdef WITH_BROWSERVIEW
  if(bv&&bv->browser_view) {
    go2view_mapping[START_IN_BROWSER]=c++;
    E(2,"browser-view enabled...  (%d)\n",c-1); }
  else
    E(0,"could not enable browser-view... (%p,%p)\n",bv,bv?bv->browser_view:NULL);
#endif

#ifdef WITH_ICONVIEW
  if(iv&&iv->icon_scroll) {
    go2view_mapping[START_IN_ICONVIEW]=c++;
    E(2,"icon-view enabled...  (%d)\n",c-1); }
  else
    E(0,"could not enable icon-view... (%p,%p)\n",iv,iv?iv->icon_scroll:NULL);
#endif

  if(ns->about) {
    go2view_mapping[START_IN_CREDITS]=c++;
    E(2,"credits-screen enabled...  (%d)\n",c-1); }}



/**
 * go to a different page in the notebook of a given session.
 * @return TRUE for success, FALSE otherwise.
 * @param ns - the session
 * @param thepage - what page in notebook to go to
 */
int notebook_go2view(noir_sess *ns,int thepage) {
  int page=-1;

  if(thepage>START_IN_CREDITS) {
    E(0,"(%d): parameter out of range (0..%d)\n",thepage,START_IN_CREDITS);
    return FALSE; }
  page=go2view_mapping[thepage];

  if((page<0)||
     (page>gtk_notebook_page_num(GTK_NOTEBOOK(ns->notebook),ns->about))) {
      E(0,"(sess-no: #%ld, view-no: #%d-\"%s\") failed (%d-\"%s\")...\n",ns->serial,thepage,view_name[thepage],page,(page==-1)?"view not available":"unknown error");
    return FALSE; }

  ns->start_in=thepage;
  gtk_notebook_set_current_page(GTK_NOTEBOOK(ns->notebook),page);

#ifdef HAVE_EVAS2
  { noir_iconview *iv=(noir_iconview *)noir_sess_client_get_data(ns,CLIENT_ICONVIEW);
    if((thepage==START_IN_ICONVIEW)&&(iv->icon_view)&&
       GTK_WIDGET_MAPPED(iv->icon_view))
      gtk_widget_grab_focus(iv->icon_view); }
#endif

  return TRUE; }






/*****************************************************************************/



/**
 * end program by destroying all its windows.
 */
void gui_quit(void) {
  noir_sess *a=noir_sess_get_1st(),*b;

  while(a) {
    b=a->next;
    if(a->window)
      gtk_widget_destroy(a->window);
    a=b; }}






/*****************************************************************************/
/* various event-handlers */


static int evidence_shutdown(void *data) {
  char *p=strrchr(temp,'-');

  /* delete tmp-dirs, if any */
  if(p) {
    while(*(++p))
      *p='?';
    p=file_glob(temp); }
  file_rm(p?p:temp); /* synch */
  if(p)
    free(p);
  free(temp);

  /* expiry old thumbnails */
  if(!(random()%8))
    thumbnail_disk_gc();

  mainloop_quit();

  return FALSE; }



/**
 * callback for a window's "destroy" event.
 * @param w - the widget that was destroyed
 * @param ns - the session it belonged to
 */
void closed(GtkWidget *w,noir_sess *ns) {
  info_tooltip_cancel(NULL,NULL,NULL);

  if(!ns) {
    E(0,"destroyed window that belonged to no session??\n");
    return; }

  /* the window is already being destroyed, so drop window-reference to it --
     otherwise, noir_sess_drop() will try to be helpful and kill it for us. */
  ns->window=NULL;
  if(noir_sess_drop(ns)==NOIR_SUCC)
    timeout_add(0,evidence_shutdown,NULL); }



/**
 * fallback callback for mouse-clicks that no better handler exists for.
 * @param widget - the widget that received the event
 * @param ev - the actual event
 * @param data - the session
 */
static gboolean win_clicked(GtkWidget *widget,GdkEventButton *ev,gpointer data) {
  noir_sess     *ns=(noir_sess *)data;
#ifdef HAVE_GNOMECANVAS
  noir_iconview *iv=(noir_iconview *)noir_sess_client_get_data(ns,CLIENT_ICONVIEW);
#endif

  if((ev->type!=GDK_BUTTON_RELEASE)&&(ev->type!=GDK_BUTTON_PRESS))
    return FALSE;

  if(ns&&ev) {
    if(ev->button==2) {
      if(ns->typebuffer_open)
        clipboard_get(GDK_SELECTION_PRIMARY,ns);
      else
        menu_path(ns,ev->button,ev->time);
      return TRUE; }

    else if((ev->type==GDK_BUTTON_RELEASE)&&(ev->button==3)) {
      if(ev->state&GDK_CONTROL_MASK) {
        menu_root_context(ns,ns->cwd,3);
        return TRUE; }
#ifdef HAVE_GNOMECANVAS
      if(GTK_WIDGET_MAPPED(iv->icon_scroll)) {
        gint no=-1;
        if((no=gnome_icon_list_get_icon_at(GNOME_ICON_LIST(iv->icon_view),ev->x,ev->y))>=0) {
          icon_view_selected(GNOME_ICON_LIST(iv->icon_view),no,(GdkEvent *)ev,(gpointer)ns);
          return TRUE; }}
#endif
/*    menu_root_context(ns,ns->cwd,ev->button); */
      return TRUE; }}

  return FALSE; }



/**
 * fallback callback for keystrokes that no better handler exists for.
 * @param widget - the widget that received the event
 * @param ev - the actual event
 * @param data - the session
 */
static gboolean win_keyed(GtkWidget *widget,GdkEventKey *ev,gpointer data) {
  noir_sess *ns=(noir_sess *)data;

  if(!ns)
    return FALSE;

  if(ev->state&(GDK_MOD1_MASK|GDK_CONTROL_MASK)) {
    if((ev->keyval>=GDK_F1)&&(ev->keyval<=GDK_F10))
      return notebook_go2view(ns,ev->keyval-GDK_F1);

    switch(ev->keyval) {

    /* need to handle in case there is no note-book */
    case GDK_i:
      if((ev->state&GDK_CONTROL_MASK))
        return FALSE;
      notebook_go2view(ns,START_IN_ICONVIEW);
      break;
    case GDK_t:
      if((ev->state&GDK_CONTROL_MASK))
        return FALSE;
      notebook_go2view(ns,START_IN_TREEVIEW);
      break;
    case GDK_a:
      if((ev->state&GDK_CONTROL_MASK))
        return FALSE;
      notebook_go2view(ns,START_IN_CREDITS);
      break;

    case GDK_z:    /* ^Z - iconify */
/*    if(!ns->fullscreen) */
        gtk_window_iconify(GTK_WINDOW(ns->window));
      break;

    case GDK_n: { /* ^N - new folder */
      int ret;
      if((ret=menu_action_by_mime(noir_sess_get_serial(ns),MIME_INTL_MKD,noir_sess_get_cwd(ns),NULL))!=NOIR_SUCC)
        E(0,"could not open properties dialog (type \"" MIME_INTL_MKD "\"): %d\n",ret); }
      break;

    case GDK_d:    /* ^D - duplicate session */
      noir_sess_create(SESSION_COPY,ns,NULL);
      break;

    case GDK_x:    /* ^X - cut */
      if(ns->typebuffer_open) {
        clipboard_set(GDK_SELECTION_CLIPBOARD,ns->typebuffer);
        ns->typebuffer[0]='\0';
        typebuffer_set(ns); }
      else if(ns->selection) {
        gchar *cb=gtk_clipboard_wait_for_text(gtk_clipboard_get(GDK_SELECTION_PRIMARY));
        gtk_clipboard_set_text(gtk_clipboard_get(GDK_SELECTION_CLIPBOARD),cb,-1);
        g_free(cb);
        free(ns->selection);
        ns->selection=NULL;
        selection_update(ns,NULL); }
      else
	form_alert(ns,"No selection to cut to clipboard...");
      break;
    case GDK_c:    /* ^C - copy */
      if(ns->typebuffer_open)
        clipboard_set(GDK_SELECTION_CLIPBOARD,ns->typebuffer);
      else if(ns->selection) {
        gchar *cb=gtk_clipboard_wait_for_text(gtk_clipboard_get(GDK_SELECTION_PRIMARY));
        gtk_clipboard_set_text(gtk_clipboard_get(GDK_SELECTION_CLIPBOARD),cb,-1);
        g_free(cb); }
      else
	form_alert(ns,"No selection to copy to clipboard...");
      break;
    case GDK_b:
    case GDK_v:
      if((ev->state&GDK_CONTROL_MASK)) {  /* ^V  - paste */
        if(ns->typebuffer_open)
          clipboard_get((ev->keyval==GDK_b)?GDK_SELECTION_PRIMARY:GDK_SELECTION_CLIPBOARD,ns);
        else
          form_alert(ns,"Cannot paste to view, only to typebuffer..."); }
      else if(ev->keyval==GDK_b)
        notebook_go2view(ns,START_IN_BROWSER);
      break;

    case GDK_l:    /* ^L - location */
    case GDK_o:    /* ^O - open */
      ns->typebuffer_open=TRUE;
      info_tooltip_cancel(NULL,NULL,NULL);
      typebuffer_set(ns);
      return TRUE;

    case GDK_q:    /* ^Q - quit */
      if(form_yes_no(ns,"Exit file-manager?")==NOIR_SUCC)
        gui_quit();
      return TRUE;
    case GDK_u: {  /* ^U - close window */
      int but=1;
      noir_sess *s=noir_sess_get_1st();

      if(!s||!noir_sess_get_next(s))    /* last window */
        but=form_dial(ns,ALERT_QUESTION,ALERT_BUTTONS_CUSTOM,"You are about to close the file-manager's last window.\nWould you really like to exit evidence?","_Quit evidence|_Iconify window|_Cancel");

      else if(ns->fullscreen) {
        if((but=form_dial(ns,ALERT_QUESTION,ALERT_BUTTONS_CUSTOM,"You are about to close the file-manager's desktop.","_Quit evidence|_Remove desktop|_Iconify desktop|_Cancel"))==1) {
          gui_quit();
          return TRUE; }
        but--;      }

      switch(but) {
        case 1:
          noir_sess_close(ns);
          break;
        case 2:
          gtk_window_iconify(GTK_WINDOW(ns->window));
          break; }}
      return TRUE;

    case GDK_w:    /* ^W - next Window */
      if(ns->next)
        gtk_window_present(GTK_WINDOW(ns->next->window));
      else
        gtk_window_present(GTK_WINDOW(noir_sess_get_1st()->window));
      return TRUE; }}

  /***************************************************************************/

  else if(ns->typebuffer_open) {      /* TYPEBUFFER ("shell") */
    size_t l=strlen(ns->typebuffer);

    if(ev->keyval==GDK_Tab) {         /* tab-completion */
      typebuffer_tab(ns);
      if(ns->typebuffer_entered<1)
        ns->typebuffer_entered=l; }
    else {                            /* editing */
      switch(ev->keyval) {
        case GDK_BackSpace:
          if(l)
            ns->typebuffer[l-1]='\0';
          typebuffer_set(ns);
          break;
        case GDK_Return:
          typebuffer_do(ns);
          typebuffer_close(ns);
          break;
        case GDK_Escape:
          typebuffer_close(ns);
          break;
        default:                      /* input */
          if(ev->length&&((l+(ev->length)+1)<TB_MAX)) {
            E(1,"ev->length: %d,  strlen(): %d\n",ev->length,strlen(ev->string));
            strcat(ns->typebuffer,ev->string);
            typebuffer_set(ns); }
          break; }

      if(l!=strlen(ns->typebuffer)) {
        ns->typebuffer_entered=-1;
        if(ns->tabbuffer) {
          free(ns->tabbuffer);
          ns->tabbuffer=NULL; }}

      return TRUE; }}

  else if((ev->keyval==GDK_less)||(ev->keyval==GDK_slash)||
          (ev->keyval==GDK_Escape)||(ev->keyval==GDK_apostrophe)) {
    ns->typebuffer_open=TRUE;
    info_tooltip_cancel(NULL,NULL,NULL);
    /* special case, this annoyed me no end */
    if(ev->keyval==GDK_slash)
      strcpy(ns->typebuffer,"/");
    else if(ev->keyval==GDK_apostrophe)
      strcpy(ns->typebuffer,"'");
    typebuffer_set(ns); }

  else if(ev->keyval==GDK_asciitilde)
    noir_sess_focus(ns,home);

  E(3,"s: %d, k: 0x%x\n",ev->state,ev->keyval);

  return FALSE; }






/*****************************************************************************/



/**
 * update all transparency pixmaps in a window.
 *
 * for now, this will be zero or more of icon-view and shelf, ie whatever is
 * present and uses evas.  evas_bg_update() will call bg_upd() if present in
 * the background plug-in; the plug-in may update the background in whatever
 * manner it sees fit. if it supports pseudo-transparency, it may call
 * transparency_get_pixmap() in evas2.c to obtain a suitable pixmap.
 * @param ns - the session
 */
static inline int window_trans_update(noir_sess *ns) {
#ifdef HAVE_EVAS2
  { icon_list *il;
    GtkWidget *widget;
    noir_iconview *iv=(noir_iconview *)noir_sess_client_get_data(ns,CLIENT_ICONVIEW);
    noir_shelf    *sv=(noir_shelf *)   noir_sess_client_get_data(ns,CLIENT_SHELF);

    if(iv&&(widget=iv->icon_view)&&(il=ICON_LIST_GET_FROM_WIDGET(widget))) {
      /* must... force.... redraw...   (gevas_q not enough) */
      gevas_bg_update(il,il->vis_w,il->vis_h);
      evas_obscured_rectangle_add(gevas_get_evas((GtkgEvas *)il->canvas),0,0,il->vis_w,il->vis_h);
      evas_obscured_clear((gevas_get_evas((GtkgEvas *)il->canvas)));
      if(GTK_WIDGET_MAPPED(widget))
        evas_render(gevas_get_evas((GtkgEvas *)il->canvas)); }

    if(sv&&(widget=sv->shelf_view)&&(il=ICON_LIST_GET_FROM_WIDGET(widget))) {
      /* must... force.... redraw...   (gevas_q not enough) */
      gevas_bg_update(il,il->vis_w,il->vis_h);
      evas_obscured_rectangle_add(gevas_get_evas((GtkgEvas *)il->canvas),0,0,il->vis_w,il->vis_h);
      evas_obscured_clear((gevas_get_evas((GtkgEvas *)il->canvas)));
      if(GTK_WIDGET_MAPPED(widget))
        evas_render(gevas_get_evas((GtkgEvas *)il->canvas)); }}
#endif
  return NOIR_SUCC; }



/**
 * listen for properites on the root-window changing.
 *
 * for now, we're only interested in the rootmap ("wallpaper") changing,
 * because we need to update our transparency pixmap when that happens.
 * this will cycle through all windows we have as we don't get this event
 * on a per-window basis.  (event-filter)
 * @param x - the X-event, or NULL
 * @param ev - the property event
 * @param data - NULL
 */
static gboolean root_propertied(GdkXEvent *x,GdkEventProperty *ev,gpointer data) {
  char *atom;
  if((atom=gdk_atom_name(ev->atom))) {
    if(!strcmp(atom,"_XROOTPMAP_ID")) {
      noir_sess *sess=noir_sess_get_1st();
      E(3,"called\n");
      while(sess) {
        /* maybe check whether the window actually has transparency relative
           to this root pixmap here */
        window_trans_update(sess);
        sess=sess->next; }}
    g_free(atom); }
  return GDK_FILTER_CONTINUE; }


#ifdef HAVE_X

/**
 * keep track of which window was last focused (session->focused).
 *
 * event-filter.
 * @param gx - the X-event
 * @param ev - the property event, or NULL
 * @param data - the session
 */
static gboolean win_event(GdkXEvent *gx,GdkEvent *ev,gpointer data) {
  noir_sess *ns=(noir_sess *)data;
  XAnyEvent *xe=(XAnyEvent *)gx;
  if(xe->type==FocusIn)
    ns->focused=xe->serial;
  return GDK_FILTER_CONTINUE; }

#endif



/**
 * callback for configure-events on the main window.
 *
 * for now, we're only interested because we'll need to update the transparency
 * pixmap when the window is moved.
 * @param widget - the widget that received the event
 * @param ev - the actual event
 * @param data - the session
 */
static gboolean win_configured(GtkWidget *widget,GdkEventConfigure *ev,gpointer data) {
  window_trans_update((noir_sess *)data);
  return FALSE; }






/*****************************************************************************/



/**
 * pre-selected items at start-up.
 *
 * called from main.
 * @param ns - the session
 * @param sel - the selection to make. items are separated by DELIMITER_
 */
static inline int view_sel(noir_sess *ns,char *sel) {
  char *p,*q,*r=NULL,*s,*m=NULL;
  int   virt=0,l=-1,cnt=0;

  if(!sel||!*sel)
    return NOIR_FAIL;

  p=sel;
  do {
    cnt++;
    q=p;
    if((p=strchr(q,DELIMITER_)))
      *p='\0';
    if((s=strrchr(q,DIR_DELIMITER))) {
      if((s-q)>l) {
        r=q;
        l=s-q; }
      else if(((s-q)!=l)||strncmp(q,r,l+1))
        virt=1; }
    if(p)
      *(p++)=DELIMITER_;
  } while(p);

  if(virt) {
    E(2,"selecting files from multiple directories => making virt-folder...\n");
    E(0,"sorry, selecting files from multiple directories would require virtual folders, which we don't have *yet*...\n"); }
  else {
    if((m=strdup(r))) {
      m[l?l:1]='\0';
      E(2,"selecting files from single directory => selecting in \"%s\"...\n",m);

      if(ns->selection)
        free(ns->selection);
      ns->selection=strdup(sel);
      selection_update(ns,NULL);
      noir_sess_focus(ns,m); }}

  if(m)
    free(m);

  return NOIR_SUCC; }






/*****************************************************************************/



/**
 * create new session.
 * @return positive serial number of the new session, or a (negative) error-code
 * @param sm - mode: SESSION_CREATE -- use preinit'd struct referred to by template||
 *             SESSION_COPY   -- allocate, init using template
 * @param template - template to copy (SESSION_COPY), or struct to work in (SESSION_CREATE)
 * @param dir - working dir for the new session
 */
long noir_sess_create(sn_mode sm,noir_sess *template,const char *dir) {
  noir_sess *ns=NULL;
  int        page,nodeco;

  E(1,"opening (%c) \"%s\"...\n",sm==SESSION_COPY?'C':'N',dir?dir:(template->cwd?template->cwd:"<none!?>"));

  if(!template)
    return NOIR_NODATA;

  if(sm==SESSION_COPY) {
    if(!(ns=noir_sess_new()))
      return NOIR_OOM;
    noir_sess_copy(ns,template); }
  else
    ns=template;

  if(dir&&(dir!=ns->cwd)) {
    if(ns->cwd)
      free(ns->cwd);
    ns->cwd=strdup(dir); }

  if(!ns->cwd) {
    E(0,"no diretory!?\n");
    if((sm==SESSION_COPY)&&ns)
      free(ns);
    return NOIR_NODATA; }

  /* main-window */
  ns->window=gtk_window_new(GTK_WINDOW_TOPLEVEL);

#ifdef HAVE_EVAS2
  if(ns->theme) {
    if(!((noir_theme *)template->theme)->name||!(((noir_theme *)ns->theme)->name=strdup(((noir_theme *)template->theme)->name)))
      goto free_theme;
    else if(!((noir_theme *)template->theme)->path||(((noir_theme *)ns->theme)->path=strdup(((noir_theme *)template->theme)->path)))
      goto view_load_theme;
    free(((noir_theme *)ns->theme)->name);
   free_theme:
    free(ns->theme);
    ns->theme=NULL; }

 view_load_theme:
  if(!ns->theme||(theme_init(ns,cwd)!=NOIR_SUCC)) {
    if(ns&&ns->theme&&((noir_theme *)ns->theme)->name&&strcmp(((noir_theme *)ns->theme)->name,"default")) {
      E(0,"could not load theme \"%s\" (%s) {%s}, trying to fall back to default...\n",(ns&&ns->theme&&((noir_theme *)ns->theme)->name)?((noir_theme *)ns->theme)->name:"???",(ns->theme&&((noir_theme *)ns->theme)->path)?((noir_theme *)ns->theme)->path:"???",cwd);
/*    free(((noir_theme *)ns->theme)->name); */
      ((noir_theme *)ns->theme)->name=strdup("default");
      goto view_load_theme; }

    ns->start_in=START_IN_BROWSER;

    E(0,"could not load theme \"%s\" (%s) {%s}, disabling canvas...\n",(ns&&ns->theme&&((noir_theme *)ns->theme)->name)?((noir_theme *)ns->theme)->name:"???",(ns->theme&&((noir_theme *)ns->theme)->path)?((noir_theme *)ns->theme)->path:"???",cwd); }

  else
#endif
  {
#ifdef HAVE_EVAS2
    if(ns->fullscreen)
      ((noir_theme *)ns->theme)->sb_new=NULL;
    E(2,"theme \"%s\" (%s) initialized...\n",((noir_theme *)ns->theme)->name,((noir_theme *)ns->theme)->path);
#endif

    /* icon-view */
    icon_view_init(ns);

    /* shelf */
    if(!(ns->without_shelf||ns->fullscreen))
      shelf_init(ns); }

  /* tree */
#ifdef WITH_TREEVIEW
  tree_view_init(ns);
#endif

  /* browser */
#ifdef WITH_BROWSERVIEW
  browser_view_init(ns);
#endif

#ifdef HAVE_EVAS2
  { char *about=NULL;
    if(ns->theme&&((noir_theme *)ns->theme)->path&&((noir_theme *)ns->theme)->name) {
      char  *exts[]={ ".png", ".jpg", ".gif", NULL };
      char **ext=exts;
      while(!ns->about&&*ext) {
        if((about=theme_ms(ns->theme,((noir_theme *)ns->theme)->path,((noir_theme *)ns->theme)->name,"","about",*ext))) {
          ns->about=about_init(ns,about);
        free(about); }
        ext++; }}}
#else
  ns->about=about_init(NULL);
#endif

  /* assemble main-window */

  ns->container=notebook_setup(ns);
  if(ns->without_notebook)
    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(ns->notebook),FALSE);
  gtk_notebook_set_show_border(GTK_NOTEBOOK(ns->notebook),FALSE);

  gtk_signal_connect(GTK_OBJECT(ns->notebook),"switch-page",G_CALLBACK(go2view_activate_page),(gpointer)ns);

  { noir_shelf *sv=(noir_shelf *)noir_sess_client_get_data(ns,CLIENT_SHELF);
    /* setup paned */
    if(sv&&sv->shelf_scroll) {
      GtkWidget *paned=gtk_vpaned_new();
      if((ns->container=paned)) {
        gtk_paned_pack2(GTK_PANED(paned),sv->shelf_scroll,TRUE,TRUE);
        gtk_widget_set_size_request(sv->shelf_scroll,-1,10);
        gtk_paned_pack1(GTK_PANED(paned),ns->notebook,TRUE,FALSE);
        gtk_widget_set_size_request(ns->notebook,-1,88); }}}

  gtk_container_add(GTK_CONTAINER(ns->window),ns->container);

  g_signal_connect(G_OBJECT(ns->container),"button_press_event",   G_CALLBACK(win_clicked),   (gpointer)ns); /* yes, notebook is exactly what we want to connect it to. trust me. */
  g_signal_connect(G_OBJECT(ns->container),"button_release_event", G_CALLBACK(win_clicked),   (gpointer)ns); /* yes, notebook is exactly what we want to connect it to. trust me. */

  g_signal_connect(G_OBJECT(ns->window),   "destroy",              G_CALLBACK(closed),        (gpointer)ns);
  g_signal_connect(G_OBJECT(ns->window),   "key-press-event",      G_CALLBACK(win_keyed),     (gpointer)ns);

  g_signal_connect(G_OBJECT(ns->window),   "configure-event",      G_CALLBACK(win_configured),(gpointer)ns);

  gdk_window_set_events(gdk_get_default_root_window(),GDK_PROPERTY_CHANGE_MASK);
  gdk_window_add_filter(gdk_get_default_root_window(),(GdkFilterFunc)root_propertied,NULL);

  if(!ns->geometry||!gtk_window_parse_geometry(GTK_WINDOW(ns->window),ns->geometry))
    ns->geometry=0;

  nodeco=ns->without_decorations;
  if(ns->fullscreen) {
    nodeco=TRUE;
    gtk_window_set_default_size(GTK_WINDOW(ns->window),gdk_screen_width(),gdk_screen_height());
    gtk_window_move(GTK_WINDOW(ns->window),0,0);
    gtk_window_set_wmclass(GTK_WINDOW(ns->window),"desktop_window","evidence");
    gtk_window_maximize(GTK_WINDOW(ns->window)); }
  else
    gtk_window_set_default_size(GTK_WINDOW(ns->window),800,600);

  gtk_window_set_decorated(GTK_WINDOW(ns->window),!nodeco);
  gtk_widget_show_all(ns->window);

  if(ns->fullscreen)
#ifdef HAVE_X
    window_as_desk(GDK_WINDOW_XDISPLAY(ns->window->window),
                   GDK_WINDOW_XWINDOW(ns->window->window),
                   gdk_screen_width(),gdk_screen_height());
#else
    E(0,"cannot flag window as desktop, compiled with --enable-x=no\n");
#endif

  if(ns->serial==1)
    go2view_setup(ns);

#ifdef HAVE_EVAS2
  if(ns->theme) {
    page=template->start_in;
    if(!notebook_go2view(ns,START_IN_ICONVIEW)) {
      E(0,"go2view failed, critical at this point...  : (\n"); }
    else
      theme_idle_start(ns);
    template->start_in=page; }
#endif

  dnd_init(ns);

  /* go */
  if(go2view_activate_page(NULL,NULL,go2view_mapping[ns->start_in],ns)==NOIR_NOTFOUND) {
    E(0,"couldn't activate desired view %d\n",ns->start_in);
    if(noir_sess_focus(ns,cwd)!=NOIR_SUCC) {
      if(noir_sess_focus(ns,home)!=NOIR_SUCC)
        noir_sess_focus(ns,"/"); }}

  /* desktop is always in iconview, but if the user gave some other
     mode, we'll save that for further windows... */
  page=ns->start_in;
  if(ns->fullscreen)
    ns->start_in=START_IN_ICONVIEW;

  if(ns->start_in==START_IN_ICONVIEW)
    icon_view_exposed(((noir_iconview *)noir_sess_client_get_data(ns,CLIENT_ICONVIEW))->icon_view,NULL,ns);

  notebook_go2view(ns,template->start_in);

  ns->start_in=page;

#ifdef HAVE_X
  gdk_window_add_filter(ns->window->window,(GdkFilterFunc)win_event,(gpointer)ns);
#endif

  E(1,"new session: %ld-|%s|\n",ns->serial,ns->cwd);

  return ns->serial; }






/*****************************************************************************/



/**
 * marshal garbage collections.
 * @param data - reserved
 */
static int marshal_gc(void *data) {
  static int  state=0;

  switch(++state) {
  case 1:
    E(3,"see whether there are unused meta-data providers\n");
    meta_lib_gc();
    break;

  case 2:
    E(3,"see whether there are unused menu-action providers\n");
    menu_lib_gc();
    break;

  default:
    state=0; }

  return TRUE; }  /* TRUE == keep handler */



/*****************************************************************************/



/**
 * set application icon on a session.
 * @return NOIR_SUCC, or an error-code
 * @param ns - the session
 */
int set_app_icon(noir_sess *ns) {
  int    ret=NOIR_FAIL;
  gchar *ip=get_setup_file("filemanager.png");  /* "evidence_icon.png" */

  if(ip) {
    GdkPixbuf *pb=gdk_pixbuf_new_from_file(ip,NULL);
    if(pb) {
      if(ns&&ns->window) {
        gtk_window_set_icon(GTK_WINDOW(ns->window),pb);
        ret=NOIR_SUCC; }
      else
        E(0,"PixBuf %p, Session %p, Window %p\n",pb,ns,ns?ns->window:NULL); }
    else
      E(0,"could not create pixbuf from my icon??\n");
    g_free(ip); }
  else
    E(0,"could not find my icon...\n");

  return ret; }



/*****************************************************************************/



#ifdef HAVE_ECORE
static int ecore_gtk_event_handler(void *data,Ecore_Fd_Handler *fd_handler) {
  while(gdk_events_pending())
    g_main_context_iteration(NULL,TRUE);
  return 1; }

static int ecore_gtk_update_handler(void *data) {
  g_main_context_iteration(NULL,TRUE);
  return TRUE; }

#if 0
static int ecore_timer(void *data) {
  E(0,"timer %ld\n",time(NULL));
  return TRUE; }
#endif

static int ecore_exit(void *data,int ev_type,void *ev) {
  ecore_main_loop_quit();
  return 1; }
#endif



/*****************************************************************************/



/**
 * is this the first time the user starts this version?
 * @return TRUE or FALSE
 */
int new_features_p() {
  char *vf=file_user_setup("version");
  int   updated=FALSE;

  if(!vf)
    updated=TRUE;
  else {
    int  l;
    char buf[PATH_MAX+1];
    
    if((l=readlink(vf,buf,PATH_MAX))<0)
      updated=TRUE;
    else {
      buf[l]='\0';
      E(1,"last used version \"%s\", this is \"%s\".\n",buf,VERSION);
      if(strcmp(buf,VERSION))
        updated=TRUE; }

    if(updated) {
      unlink(vf);
      symlink(VERSION,vf); }}
  return updated; }



/**
 * create an instance of evidence, and run it : )
 */
int main(int argc,char *argv[]) {
  show_dir  *dirs;
  char      *dir,*sel,*ctype;
  char       env_dflt_locale[]="LC_CTYPE=de_DE";
  char      *env_prefix=NULL;
  noir_sess *ns;
  void      *marshal_gc_handle;

#ifdef HAVE_ECORE
  Ecore_Fd_Handler   *ecore_gtk_event_handle=NULL;
  void               *ecore_gtk_update_handle=NULL;
#endif

  if(!(ns=noir_sess_new())) {
    E(0,"out of memory...\n");
    return 1; }

  /* GTK init */
  if(!(ctype=getenv("LC_CTYPE"))||!strlen(ctype)||!strcmp(ctype,"C")) {
    E(1,"LC_CTYPE not set or \"C\", falling back to \"%s\"...\n",env_dflt_locale);
    putenv(env_dflt_locale); }

#ifdef HAVE_GTHREAD
  g_thread_init(NULL);
#endif

  /* this break horribly with GTK>2.2.x for some reason */
#if 0
  if(!XInitThreads()) {
    E(0,"XInitThreads() failed...\n");
    return 1; }
#endif

  gtk_init(&argc,&argv);

  main_thread=pthread_self();

  home=getenv("HOME");
  if((dir=getenv("PWD"))&&(strlen(dir)<PATH_MAX))
    strcpy(cwd,dir);  /* avoid realpath() where possible */
  else
    getcwd(cwd,PATH_MAX);

  paths_init(ns->iconset,cwd,argv[0]);
  if(prefix) {
    esprintf(&env_prefix,"PREFIX=%s",prefix);
    if(env_prefix)
      putenv(env_prefix); }  /* valgrind says this leaks, POSIX says this is right? */


  ns->cwd=strdup(cwd);
  parse_options(ns,argc,argv,&dirs,&sel);
  if(dirs)
    dir=dirs->dir;
  if(dir) {
    if(ns->cwd)
      free(ns->cwd);
    ns->cwd=strdup(dir); }

  E(1,"Evidence " VERSION " (" __DATE__ ") coming up...\n\nstarting in directory \"%s\"...\n",ns->cwd);

  if(new_features_p())
    E(0,"\n\n -- UPDATE! -- \n\n");

  srandom(now(0));

  { char *t=NULL;
    esprintf(&t,"/tmp/evidence-%s-XXXXXX",get_username(geteuid()));
    if(!t) {
      E(0,"out of memory!?\n");
      exit(1); }
    if((temp=mkdtemp(t)))
      E(1,"tmpdir is \"%s\"...\n",temp);
    else {
      perror("tmpdir-creation failed");
      free(t); }}

  /* init */

  if(!(queue=model_init(get_hostname()))) {
    E(0,"fatal -- could not initialize model!?\n");
    goto no_model; }

  ipc_init(); /* do this before file_init() so the latter can override some signal()s */
  file_init(queue,"lib/evidence/backend/fsxs.so");
  icon_init();
  meta_init();
  menu_init(ns);
  typebuffer_init(ns);
  thumbnail_init();

#ifdef WITH_TREEVIEW
  /* set up tree model */
  { noir_treeview tv;
    memset(&tv,0,sizeof(noir_treeview));
    tv.session=ns;
    tree_model_setup(&tv); }
#endif

  /* open all requested directories */
  { int       c=0;
    show_dir *t;

    noir_sess_create(SESSION_CREATE,ns,NULL);
    while(dirs) {
      if(c)
        noir_sess_create(SESSION_COPY,ns,dirs->dir);
      c++;
      t=dirs;
      dirs=dirs->next;
      free(t->dir);
      free(t); }}

  /* set application icon */
  set_app_icon(ns);

  if(sel) {
    view_sel(ns,sel);
    free(sel); }

  marshal_gc_handle=timeout_add(GC_INTERVAL_MAIN,marshal_gc,NULL);

/* file_monitor_add_with_cb(ns,NOIR_MONITOR_FILE,"/home/azundris/m1/m2/m3"); */

#ifdef HAVE_GTHREAD
  gdk_threads_enter();
#endif

#if defined(HAVE_ECORE)&&defined(HAVE_X)
  ecore_init();
  ecore_app_args_set(argc,(const char **)argv);

  { Display *d=get_display();
    if((ecore_gtk_event_handle=ecore_main_fd_handler_add(ConnectionNumber(d),ECORE_FD_READ,ecore_gtk_event_handler,d,NULL,NULL))) {
      ecore_gtk_update_handle=idle_enter_add(0,ecore_gtk_update_handler,NULL);
/*    timeout_add(500,ecore_timer,NULL); */
      ecore_event_handler_add(ECORE_EVENT_SIGNAL_EXIT,ecore_exit,NULL);
      ecore_main_loop_begin(); }
    else
      E(0,"FAIL ECORE\n"); }
#else

/*  E(0,"giving plugins a moment...  "); sleep(1); if(debug>=0) fprintf(stderr,"go!\n"); */
  gtk_main();

#endif

#ifdef HAVE_GTHREAD
  gdk_threads_leave();
#endif

  while(noir_sess_get_1st()) {
    noir_sess *sa=noir_sess_get_1st();
    E(1,"destroying remaining session #%ld\n",sa->serial);
    noir_sess_close(sa); }

#ifdef HAVE_ECORE
  ecore_main_fd_handler_del(ecore_gtk_event_handle);
  idle_enter_remove(&ecore_gtk_update_handle);
  ecore_shutdown();
#endif

#if !defined(HAVE_GNOMECANVAS)&&!defined(HAVE_EVAS2)
#  warning compiling without a canvas
#endif

  ipc_exit();
  typebuffer_exit(ns);
  model_exit(queue);
 no_model:
  menu_exit(ns);
  thumbnail_exit();
  file_exit();
  meta_exit();
  icon_exit();
  paths_exit();

  return 0; }
