/** @file  focus.c
 *  @brief load (where necessary) and display a directory and all directories leading up to it.
 */

#include <stdlib.h>         /* free */
#include <string.h>         /* strstr() et al. */
#include <stdio.h>          /* fprintf() et al. */
#include <sys/stat.h>       /* S_ISDIR() */
#include <errno.h>          /* errno */
#include <unistd.h>         /* chdir() */

#include "session.h"
#include "session_private.h"
#include "model.h"          /* path_get(), CLIENT_PAUSED() */
#include "alert.h"          /* form_alert() */
#include "file.h"           /* file_exists() */
#include "monitors.h"       /* file_monitor_*() */

#include "gui/gtk2/common/tooltip.h"        /* info_tooltip_cancel() */

#include "gui/efl/common/theme.h"



/****************************************************************************/



/**
 * check whether the new path exists and consists of directories only.
 * @return NOIR_SUCC if we don't object to newdir becoming ns->cwd.
 * @param ns     - the session we want to change the CWD for
 * @param olddir - the CWD for that session
 * @param newdir - the directory we want to become the new CWD
 */
static inline int focus_is_path_valid(noir_sess *ns,char *olddir,char *newdir) {
  gint  isdir;
  char *e,*p;
  long  err=0;

#ifndef WITH_DOT_DIR
  while((e=strstr(newdir,"/./")))
    memmove(e,e+2,strlen(e+2)+1);
#endif

  while((e=strstr(newdir,"/../"))) {
    *e='\0';
    if((p=strrchr(newdir,'/')))
      memmove(p,e+3,strlen(e+3)+1);
    else
      memmove(e,e+3,strlen(e+3)+1); }

  if((p=strrchr(newdir,'/'))) {
    if(!strcmp(p,"/..")) {
      *p='\0';
      if((p=strrchr(newdir,'/')))
        *p='\0'; }
#ifndef WITH_DOT_DIR
    else if(!strcmp(p,"/."))
      *p='\0';
#endif
    else if(!p[1]&&(p>newdir))   /* discard trailing / */
      *p='\0'; }

  if(!S_ISDIR(isdir=file_exists(newdir))||chdir(newdir)) {
    char *e;
    int   et;
    err=errno;
    do {
      et=isdir;
      if((e=strrchr(newdir,'/')))
        *e='\0';
    } while((e>newdir)&&!S_ISDIR(isdir=file_exists(newdir)));

    E(0,"(,\"%s\"): \"%s\" last valid dir...\n",olddir,!strcmp(olddir,newdir)?"<none>":newdir);

    if(e) {
      char *ets=(err==EACCES)?"is inaccessible":(et?"is not a directory":"not found");
      *e='/';
      form_alert(ns,"%s %s!",newdir,ets); }

    return NOIR_NOTFOUND; }

  return NOIR_SUCC; }



/****************************************************************************/



/**
 * change the current working-directory of a session.
 *
 * called by focus() to do the actual work.
 * @param ns - the session
 * @param tag - the desired new working-directory
 */
static inline int do_cwd(noir_sess *ns,char *tag) {
  file_info   *fi,fs;
  char        *p;

  if((p=strrchr(tag,DIR_DELIMITER))&&!strcmp(p+1,DIR_PARENT))
    return NOIR_IGNORED;
#ifndef WITH_DOT_DIR
  if(!strcmp(p+1,DIR_MYSELF))
    return NOIR_IGNORED;
#endif

  if(!(fi=file_info_fetch(NULL,tag))) {
    ns->cd=NULL;
    memset(&fs,0,sizeof(file_info));
    fi=&fs;
    fs.tag=tag; }
  else
    ns->cd=fi;

  E(2,"(%s) => %s\n",tag,fi->tag);

  noir_sess_tell_clients(ns,fi,MODEL_CWD,FF_NONE);  /* signal pre-cwd */

  path_get(queue,NULL,tag,PG_ENQUEUE_ALL);          /* signal _LAST */
  ns->cd=file_info_fetch(NULL,tag);

  noir_sess_tell_clients(ns,fi,MODEL_CWD,FF_NAME);  /* signal post-cwd */

  return NOIR_SUCC; }



/****************************************************************************/



/**
 * change the current working-directory of a session.
 *
 * provides the supporting structure to the actual work done in do_cwd().
 * @param ns - the session
 * @param dir - the desired new working-directory
 */
int noir_sess_focus(noir_sess *ns,char *dir) {
  file_annotation *fa;
  int              ret=NOIR_SUCC;
  char            *s=dir;

  /* this should only happen when a property-change forces a redraw
     (e.g. "icon-font") while in focus().  the canonical example is
     start-up, but this can also happen if a remote config-tool
     changes a prop. */
  if(ns->sema_focus) {
    E(2,"already in focus(), call ignored\n");
    return NOIR_IGNORED; }

  ns->sema_focus++;

  if(!dir||!*dir)
    return NOIR_NODATA;

  info_tooltip_cancel(NULL,NULL,NULL);

  /* update cwd */
  if(dir[0]!='/') {                           /* relative path */
    size_t l=strlen(dir)+strlen(ns->cwd)+2;
    if(!(dir=malloc(l)))
      return NOIR_OOM;
    snprintf(dir,l,"%s/%s",ns->cwd,s); }
  else if(!(dir=strdup(dir)))                 /* absolute path */
    return NOIR_OOM;

  if((ret=focus_is_path_valid(ns,ns->cwd,dir))!=NOIR_SUCC) {
    free(dir);
    goto done; }

  if((fa=file_annotation_get(dir))) {
    int reply=NOIR_SUCC;
    if((fa->lvl&FA_ROOT)&&(geteuid()!=0)) {
      if((reply=form_yes_no(ns,"You are about to enter a folder only relevant to\nthe system-administrator while logged in as a user. Proceed?"))==NOIR_SUCC)
        fa->lvl&=(~FA_ROOT); }
    if(fa->lvl&FA_MEDIA) {
      if((reply=form_yes_no(ns,"You are about to enter a folder that lives on a removable media\n(such as a CD, floppy disk, or USB stick). Proceed?"))==NOIR_SUCC)
        fa->lvl&=(~FA_MEDIA); }
    if(reply!=NOIR_SUCC)
      goto done; }

  E(1,"%s => %s\n",ns->cwd?ns->cwd:"",dir?dir:"");

  /* save old cwd in pwd etc. */
  if(!ns->cwd||strcmp(dir,ns->cwd)) {
    if(ns->pwd)
      free(ns->pwd);
    ns->pwd=ns->cwd; }
  else
    free(ns->cwd);
  ns->cwd=dir;

  /* do the actual cd */
  if(do_cwd(ns,ns->cwd)==NOIR_SUCC) {
    /* set monitors for directories along the path */
    file_info *fi=file_info_fetch(NULL,ns->cwd);
/*  file_monitor_remove_all(ns); */  /* check ### */
    while(fi) {
      file_monitor_add(ns,NOIR_MONITOR_DIR,fi->tag);
      fi=fi->parent; }}

 done:
  ns->sema_focus=0;

  return ret; }



/****************************************************************************/
