/** @file  typebuffer.c
 *  @brief handle "typebuffer" (micro-shell)
 */

#include "config.h"

#include "gui.h"

#include "session_private.h"

#include "mime.h"        /* MIME_* */
#include "file.h"        /* file_fork(), ... */
#include "util.h"        /* hashes, replace_variables() */
#include "setup.h"       /* get_setup_file(), get_parse_file(),
                            parse_line(), replace_variables() */
#include "alert.h"       /* form_alert(), form_yes_no() */
#include "typebuffer.h"

#include "gui/gtk2/common/tree/tree.h"
#include "gui/gtk2/tree_view/tree_view.h"
#include "gui/gtk2/icon_view/icon_view.h"

#ifdef HAVE_EVAS2
#  include "gui/efl/common/theme.h"
#  include "icon_list.h"
#  include "gui/gtk2/icon_view/gevas_light.h"
#  include "gui/gtk2/icon_view/gevas_bind.h"         /* ICON_LIST_GET_FROM_WIDGET() */
#  include "gui/efl/icon_view/evas_bind.h"
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <glob.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <ctype.h>
#include <pwd.h>         /* getpwnam() */

#ifdef HAVE_PCRE
#  include <pcre.h>
#  define RE_OPTS 0
#  define RE_ELS  30
#endif



static void *ht_handler=NULL;

typedef enum { HIST_GENERIC, HIST_URL } hist_type;

typedef struct _hist_entry {
  char               *line;
  time_t              timestamp;
  hist_type           type;
  struct _hist_entry *next;
} hist_entry;

hist_entry *history=NULL;
 



/****************************************************************************/
/* setup */



/**
 * add URL-handler
 * @param ns the session
 * @param prot the protocol ("http")
 * @param handler the handler (string with substitutions, %u, %f, ...)
 */
static int typebuffer_add_handler(noir_sess *ns,char *prot,char *handler) {
  hash_table_insert(ht_handler,strdup(prot),strdup(handler));
  return NOIR_SUCC; }



/**
 * initialize typebuffer subsystem for given session
 * @param ns the session
 */
int typebuffer_init(noir_sess *ns) {
  int ret=NOIR_SUCC;
  gchar *handlers=get_setup_file(PROTOCOL_HANDLERS);

  if(handlers) {
    FILE *fh;
    char l[MIME_MAX_LEN];
    char *p,*p3;

    if((fh=fopen(handlers,"r"))) {
      E(2,"URL-handlers in type-buffer:\n");

      if(!(ht_handler=hash_table_new(free,free))) {
        ret=NOIR_OOM;
        goto close; }

      while(fgets(l,MIME_MAX_LEN,fh)) {
        if(parse_line(l,&p,&p3,NULL,NULL)==2) {
          E(2,"%s" PROTOCOL_DELIMITER " => %s\n",p,p3);
          typebuffer_add_handler(ns,p,p3); }}
    close:
      fclose(fh); }
    else
      ret=NOIR_FAIL;

    g_free(handlers); }

  return ret; }



/**
 * deinitialize typebuffer subsystem for given session
 * @param ns the session
 */
int typebuffer_exit(noir_sess *ns) {
  if(ht_handler)
    hash_table_dst(&ht_handler);
  /* tb_data free()d in theme.c */
#if 0
  if(((noir_theme *)ns->theme)&&((noir_theme *)ns->theme)->tb_data) {
    free(((noir_theme *)ns->theme)->tb_data);
    ((noir_theme *)ns->theme)->tb_data=NULL; }
#endif
  return NOIR_SUCC; }



/****************************************************************************/
/* regex */



#ifdef HAVE_PCRE

/**
 * compile a regular expression
 * @return an error-code
 * @param ns      the session
 * @param re_txt  the cleartext-representation of the regex
 */
int re_add(noir_sess *ns,char *re_txt) {
  const char *error;
  int         erroffset;
  pcre       *re=pcre_compile(re_txt,RE_OPTS,&error,&erroffset,NULL);
  pcre_extra *pe;
  char       *n;

  if(!re) {
    form_alert(ns,"Invalid regular expression \"%s\".\nPerl-compatible syntax applies.",re_txt);
    return NOIR_FAIL; }

  pe=pcre_study(re,0,&error);
  if((n=strdup(re_txt))) {
    if(ns->pcre)  free(ns->pcre);
    if(ns->re)    free(ns->re);
    if(ns->pe)    free(ns->pe);
    ns->pcre=n;
    ns->re=re;
    ns->pe=pe;
    return NOIR_SUCC; }
  return NOIR_OOM; }



/**
 * check whether current regex matches given file
 * @return an error-code
 * @param ns  the session
 * @param fn  the file-name
 */
int re_match(noir_sess *ns,char *fn) {
  int rc;
  int ovector[RE_ELS];

  rc=pcre_exec(ns->re,ns->pe,fn,strlen(fn),0,RE_OPTS,ovector,RE_ELS);

  if(rc<0) {
    switch(rc) {
    case PCRE_ERROR_NOMATCH:
      return NOIR_NOTFOUND;
      break;
    default:
      E(0,"error %d...\n",rc);
      return NOIR_FAIL; }}

  return NOIR_SUCC; }
#endif



/****************************************************************************/
/* display */



/**
 * display current contents of a session's typebuffer
 * @param ns the session
 */
void typebuffer_set(noir_sess *ns) {
  noir_iconview *iv=(noir_iconview *)noir_sess_client_get_data(ns,CLIENT_ICONVIEW);
  char          *tmp=NULL;

  if(ns->typebuffer_open) {
    char *msg=(ns->typebuffer&&strlen(ns->typebuffer))?ns->typebuffer:WINDOW_TYPEBUFFER;
    char *utf8;

    esprintf(&tmp,"evidence shell: %s",msg);

    if(tmp) {
      if((utf8=e2utf8(tmp))) {
        gtk_window_set_title(GTK_WINDOW(ns->window),utf8);
        free(utf8); }
      free(tmp); }
#ifdef HAVE_EVAS2
    if((utf8=e2utf8(msg))) {
      ns->tb=((noir_theme *)ns->theme)->tb_set(ICON_LIST_GET_FROM_WIDGET(iv->icon_view),ns->tb,utf8);
      free(utf8); }
#endif
  }
  else {
#ifdef HAVE_EVAS2
    /* command might have discarded theme! ("quit" etc.) */
    if(((noir_theme *)ns->theme)&&((noir_theme *)ns->theme)->tb_dst&&ns->tb)
      ((noir_theme *)ns->theme)->tb_dst(&ns->tb);
#endif
    esprintf(&tmp,"evidence - %s",ns->cwd);
    if(tmp) {
      gtk_window_set_title(GTK_WINDOW(ns->window),tmp);
      free(tmp); }}
#ifdef HAVE_EVAS2
  if(((noir_theme *)ns->theme)&&((noir_theme *)ns->theme)->tb_redraw)
    ((noir_theme *)ns->theme)->tb_redraw(ICON_LIST_GET_FROM_WIDGET(iv->icon_view));
#endif
}



/**
 * close a session's typebuffer
 * @param ns the session
 */
void typebuffer_close(noir_sess *ns) {
  ns->typebuffer_open=FALSE;
  ns->typebuffer[0]='\0';
  typebuffer_set(ns); }



/****************************************************************************/
/* handle URL */



/**
 * contents assumed URL. find handler, start it with the URL.
 * @param ns the session
 * @param url the URL
 */
static int typebuffer_handle_URL(noir_sess *ns,char *url) {
  char *r,*h;
  int   ret=NOIR_SUCC;

  if(!ht_handler||!url||!*url)
    return NOIR_NODATA;

  if(!ns)
    return NOIR_FAIL;

  E(1,"\"%s\"\n",url);

  if(!(r=strstr(url,PROTOCOL_DELIMITER)))
    return NOIR_FAIL;

  *r='\0';
  h=(char *)hash_table_fetch(ht_handler,url);

  if(!h) {
    form_alert(ns,"No handler registered for protocol \"%s\"...",url);
    ret=NOIR_NOTFOUND; }

  *r=':';
  if(h)
    ret=file_run(ns,h,url,"");

  return ret; }



/****************************************************************************/
/* handle magic comma operator */



/**
 * contents assumed "'propname value"; modify theme.
 * @return an error-code
 * @param ns the session
 * @param tb the typebuffer-contents
 */
static int typebuffer_modify_theme(noir_sess *ns,char *tb) {
  char *p=&tb[1],*k,*v;

  k=&p[strlen(p)-1];
  while(isspace(*k))                  /* remove trailing space */
    *(k--)='\0';

  while(*p&&isspace(*p))              /* skip initial space */
    p++;

  if(!*p)                             /* nothing left?  go home */
    return NOIR_IGNORED;

  k=p;
  while(*p&&(strchr(":= ",*p)==NULL))
    p++;

  if(*p) {                            /* multiple words? */
    size_t l;

    *(p++)='\0';
    while(*p&&(isspace(*p)||(*p==':')||(*p=='=')))
      p++;

    theme_el_canonize_key(k,TRUE);

    if(*p&&(l=strlen(k))) {
      noir_theme_el *e;

      v=p;

      E(1,"(\"%s\",\"%s\")\n",k,v);

      if(!(e=theme_el_get(&((noir_theme *)ns->theme)->data,k))) {
        if(form_yes_no(ns,"No such entity to modify, append as new entity?")!=NOIR_SUCC)
          return NOIR_IGNORED; }
      if((l>4)&&!strcasecmp(k+l-4,"font")) {
        int ret=NOIR_FAIL;
#ifdef HAVE_EVAS2
        if((v=theme_font_find(((noir_theme *)ns->theme),v))) {
          if(file_exists(v))
            ret=theme_el_set(&((noir_theme *)ns->theme)->data,k,v);
          else
            form_alert(ns,"No such font \"%s\"...",v);
          free(v); }
#else
	form_alert(ns,"Can't set font for GNOME-icon-view...");
#endif
        return ret; }
      else
        return theme_parse_line(ns,k,v); }}

  else if(!strcmp(k,"save")||!strcmp(k,"dump"))
    return theme_dump((void *)ns);

  else {
    if((v=&k[strlen(k)-1])&&(*v=='?'))
      *v='\0';
    E(0,"%s\n",theme_el_get_as_string(&((noir_theme *)ns->theme)->data,k)); }

  return NOIR_IGNORED; }



/****************************************************************************/



/**
 * debug helper:  print out a single file-info
 * @return an error-code
 * @param ns the session
 * @param tb the typebuffer's contents
 */
static int typebuffer_dump_fi(noir_sess *ns,char *tb) {
  char      *n=&tb[1];
  file_info *fi=file_info_fetch(NULL,n);
  int        ret=NOIR_SUCC;

  if(fi) {
    E(0,"\"%s\" \"%s\"  $%x ($%x)...\n",fi->name,fi->tag,fi->intrinsic?fi->intrinsic->mode:0,fi->flags);
    model_ls(fi,0); }
  else {
    if(!model_find_by_name(NULL,n)) {
      E(0,"\"%s\" not found in store...\n",n);
      ret=NOIR_NOTFOUND; }}
  return ret; }



/****************************************************************************/
/* cd */



/**
 * cd -- change working directory
 * @return an error-code
 * @param ns the session
 * @param path the new wd
 */
int typebuffer_focus(noir_sess *ns,char *path) {
  int ret=noir_sess_focus(ns,path);
#if 0
  noir_iconview *iv=(noir_iconview *)noir_sess_client_get_data(ns,CLIENT_ICONVIEW);
  if(GTK_WIDGET_MAPPED(iv->icon_view)) {
    go2view(ns,START_IN_CREDITS);
    go2view(ns,START_IN_ICONVIEW); }
#endif
  return ret; }



/****************************************************************************/
/* parser helpers */



/**
 * get token
 * @param ns the session
 * @param begin start of current token
 * @param end end of current token
 * @param cook resolve $VARIABLES in input?
 */
static inline void typebuffer_get_token(noir_sess *ns,char **begin,char **end,int cook) {
  char *b=*begin,*e=*end,t=' ';

  while(*b==' ')                            /* skip spaces */
    b++;

  if(*b=='\"') {
    t='\"';
    b++; }

  e=b;                                      /* jump to end of element */
  while(*e&&*e!=t) {
    if((*e=='\\')&&e[1])                    /* skip escaped char */
      e+=2;
    else
      e++; }

  if(cook) {
    if((t=='\"')&&(*e==t))
      *(e++)='\0';
    while(*e==' ')                            /* delimit */
      *(e++)='\0';
    b=replace_variables(ns,b); }

  else {
    if((t=='\"')&&(*e==t))
      e++;
    while(*e==' ')                            /* delimit */
      e++; }

  E(1," ::%s::\n",b);

  *begin=b;
  *end=e; }



/**
 * assemble path
 * @return an error-code
 * @param buf an initialized buffer (changed)
 * @param a a string to append to the buffer, or NULL
 * @param b a string to append to the buffer, or NULL
 */
static int handle_relative(char *buf,char *a,char *b) {
  size_t la=a?strlen(a):0;
  size_t lb=b?strlen(b):0;

  if((la+lb+1)>PATH_MAX)
    return NOIR_PATHEX;

  strcat(strcat(buf,a?a:""),b?b:"");
  return NOIR_SUCC; }



/**
 * normalize a path (resolve /./ and /../)
 * @return an error-code
 * @param myfile the path to normalize
 * @param leave_single leave in single-dot entities "/./"
 */
static int path_normalize(char *myfile,int leave_single) {
  /* can't use realpath because we want to keep symlinks */
  char *t=myfile;
  while((t=strstr(t,"/..")))  /* eliminate /../ */
    if((!t[3]||t[3]=='/')&&(t>myfile)) {
      char *t2=t;
      do {
        t2--;
      } while((t2>myfile)&&(*t2!='/'));
      if(*t2=='/') {
        t+=3;
        memmove(t2,t,strlen(t)+1);
        t=t2; }
      else
        t++; }
    else
      t++;

  if(!leave_single) {
    t=myfile;
    while((t=strstr(t,"/."))) { /* eliminate /./ */
      if(!t[2]||t[2]=='/')
        memmove(t,t+2,strlen(t+2));
      t++; }}

  return NOIR_SUCC; }


/****************************************************************************/
/* history */



/**
 * add an item to shared typebuffer-history
 * @return an error-code
 * @param l the line to add to the history
 */
int typebuffer_history_add(char *l) {
  hist_entry *h;
  char       *p,*q;

  if(!l||!*l)
    return NOIR_IGNORED;

  if(!(h=malloc(sizeof(hist_entry))))
    return NOIR_OOM;

  memset(h,0,sizeof(hist_entry));

  if(!(h->line=strdup(l))) {
    free(h);
    return NOIR_OOM; }

  h->timestamp=now(0);
  p=strchr(l,' ');
  q=strstr(l,PROTOCOL_DELIMITER);
  h->type=(q&&(!p||(p>q)))?HIST_URL:HIST_GENERIC;
  h->next=NULL;

  if(!history)
    history=h;
  else {
    h->next=history;
    history=h;
/* todo: expire old entries */
  }

  return NOIR_SUCC; }



/****************************************************************************/
/* tab-completion */



/***********/
/* helpers */
/***********/

/**
 * qsort_helper: compare two entries in tab-completion array
 * @return -1,0,1 -- see qsort()
 * @param pa first entry
 * @param pb second entry
 */
static int tab_sort_fun(const void *pa, const void *pb) {
  char *a=*((char **)pa),*b=*((char **)pb);
  return strcmp(a,b); }



/**
 * find token in typebuffer in array of possible completions
 * @param ns the session
 * @param tb the token to complete
 * @param e entered -- the first e characters must match
 * @param n number of possible completions in array
 * @param s array of possible completions
 */
int tab(noir_sess *ns,char *tb,int e,int n,char **s) {
  int   i=0,l;

  l=(int)strlen(tb);

  while((i<n)&&strncmp(tb,s[i],l))
    i++;

  E(1,"looking for \"%s\"... e=%d, s[%d]=\"%s\"\n",tb,e,i,(i>=n)?"<undefined>":s[i]);

  if(i>=n)
    return NOIR_NOTFOUND;

  if(e>=0) {
    i++;
    if((i>=n)||strncmp(tb,s[i],e)) {
      i=0;
      while((i<n)&&strncmp(tb,s[i],e))
        i++; }}

  return (i<n)?i:NOIR_NOTFOUND; }



/**********************************/
/* create array of tab-completion */
/* options according to context   */
/**********************************/

/**
 * tab-complete a shell-command (complete from files in $PATH)
 * @return an error-code
 * @param ns the session
 */
static int tab_shell(noir_sess *ns) {
#define STM (2*PATH_MAX)
  int      c,n=0,m,ret=NOIR_IGNORED,
           flags=GLOB_MARK|GLOB_BRACE|GLOB_TILDE;
  char    *tb=ns->typebuffer,*l,*b,*e,*f=NULL;
  char     np[STM+3],tp[STM+3];
  glob_t   myglob;
  char   **s=NULL,**p;

  E(1,"\n");

  if(!tb||!*tb)
    return NOIR_NODATA;
  if(strlen(tb)>=(STM))
    return NOIR_OOM;

  for(e=NULL,b=tb;b;) {
    typebuffer_get_token(ns,&b,&e,FALSE);
    if(strlen(b)) {
      tb=b;
      b=e; }
    else
      b=NULL; }

  if((tb>ns->typebuffer)&&(tb[-1]=='\"'))
    tb--;
  e=tb;

  f=tb;
  typebuffer_get_token(ns,&f,&l,TRUE);
  if(f)
    tb=f;

  if(ns->tabbuffer)
    strcpy(np,ns->tabbuffer);
  else {
    if((tb[0]==DIR_DELIMITER)||(tb[0]=='~'))
      strcpy(np,tb);
    else
      snprintf(np,STM,"%s/%s",ns->cwd,tb);
    c=strlen(np);
    if(np[c-1]!='*')
      strcat(np,"*");

    path_normalize(np,FALSE);

    if(strlen(np)>=PATH_MAX) {
      ret=NOIR_FAIL;
      goto fail; }}

  l=NULL;
 retry:
  E(1,"glob \"%s\", last match \"%s\"\n",np,tb);

  if(!(c=glob(np,flags,NULL,&myglob)))
    n=myglob.gl_pathc;
  else if(c==GLOB_NOSPACE) {
    ret=NOIR_OOM;
    goto done; }

  if(!n) {
    char *x=np,*y;
    tp[0]='\0';
    while((y=strchr(x+1,DIR_DELIMITER))) {
      strncat(tp,x,y-x);
      if(!S_ISDIR(file_exists(tp))) {
        if(l!=x) {
          l=x;
          strcat(tp,"*");
          globfree(&myglob);
          if((c=glob(tp,flags,NULL,&myglob))) {
            ret=NOIR_NOTFOUND;
            goto done; }}}
      x=y; }
    strcat(tp,x);
    globfree(&myglob);
    if((c=glob(tp,flags,NULL,&myglob))) {
      ret=NOIR_NOTFOUND;
      goto done; }
    strcpy(np,tp);
    E(2,"retry with \"%s\"\n",np);
    goto retry; }

  if(!(p=s=malloc(n*sizeof(char *)))) {
    ret=NOIR_OOM;
    goto fail; }

  m=0;
  for(c=0;c<n;c++) {
    char *x;
    l=myglob.gl_pathv[c];
    x=l+strlen(l)-1;
    if((*x==DIR_DELIMITER)&&(x>l)) {
      do {
        x--;
      } while((x>l)&&(*x!=DIR_DELIMITER));
      if((*x==DIR_DELIMITER)&&
         ((!strcmp(x,DIR_DELIM_STR DIR_MYSELF DIR_DELIM_STR))||
          (!strcmp(x,DIR_DELIM_STR DIR_PARENT DIR_DELIM_STR))))
        l=NULL; }

    if(l) {
      m++;
      E(3,"adding match \"%s\"...\n",l);
      *(p++)=l; }}

  qsort(s,n,sizeof(char *),tab_sort_fun); /* according to the man-page, glob should return a sorted array to begin with */

  /* yes, it's tabbuffer, typebuffer */
  if((ret=tab(ns,ns->tabbuffer?tb:"",0,m,s))!=NOIR_NOTFOUND) {
    int     o=e-ns->typebuffer;
    size_t  m=strlen(s[ret])+o+4;
    if(m<TB_MAX) {
      if(strchr(s[ret],' '))
        snprintf(ns->typebuffer+o,m-o,"\"%s\"",s[ret]);
      else
        strncpy(ns->typebuffer+o,s[ret],m-o);
      if(!ns->tabbuffer)
        ns->tabbuffer=strdup(np);
      ret=NOIR_SUCC;
      typebuffer_set(ns); }
    else
      ret=NOIR_OOM; }

  free(s);

done:
  globfree(&myglob);
fail:
  if(f)
    free(f);

  return ret; }



/**
 * tab-complete a property (complete from list of properties in session)
 * @return an error-code
 * @param ns the session
 */
static int tab_property(noir_sess *ns) {
  noir_theme_el  *e;
  int             n=0,ret=NOIR_NODATA;
  char          **s=NULL,**p;
  char           *tb;

  if(!ns||!((noir_theme *)ns->theme)||!((noir_theme *)ns->theme)->data)
    return NOIR_NODATA;

  for(e=((noir_theme *)ns->theme)->data;e;e=e->next)
    n++;

  if(!(p=s=malloc(n*sizeof(char *))))
    return NOIR_OOM;

  for(e=((noir_theme *)ns->theme)->data;e;e=e->next)
    *(p++)=e->key;

  qsort(s,n,sizeof(char *),tab_sort_fun);

  if((tb=theme_el_canonize_key(ns->typebuffer+1,FALSE))) {
    ret=tab(ns,tb,ns->typebuffer_entered-1,n,s);

    if(ret!=NOIR_NOTFOUND) {
      size_t  l=strlen(s[ret])+2;

      if(l<TB_MAX) {
        strcat(strcpy(ns->typebuffer,"'"),s[ret]);
        typebuffer_set(ns);
        ret=NOIR_SUCC; }}

    free(tb); }
  else
    ret=NOIR_OOM;

  free(s);

  return ret; }



/**
 * tab-complete an URL (complete from history)
 * @return an error-code
 * @param ns the session
 */
static int tab_url(noir_sess *ns) {
  hist_entry  *h;
  int          n=0,ret=NOIR_NODATA;
  char       **s=NULL,**p;

  if(!history)
    return NOIR_NODATA;

  for(h=history;h;h=h->next)
    n++;

  if(!(p=s=malloc(n*sizeof(char *))))
    return NOIR_OOM;

  for(h=history;h;h=h->next)
    *(p++)=h->line;

  qsort(s,n,sizeof(char *),tab_sort_fun);

  E(0,"\"%s\"\n",ns->typebuffer);
  if((ret=tab(ns,ns->typebuffer,ns->typebuffer_entered,n,s))!=NOIR_NOTFOUND) {
    if(strlen(s[ret])<TB_MAX) {
      ret=NOIR_SUCC;
      strcpy(ns->typebuffer,s[ret]);
      typebuffer_set(ns); }
    else
      ret=NOIR_OOM; }

  free(s);

  return ret; }



/***********************************/

/**
 * find out what's in type-buffer, and tab-complete accordingly
 * @return an error-code
 * @param ns the session
 */
int typebuffer_tab(noir_sess *ns) {
  char *tb;

  if(!(tb=ns->typebuffer))
    return NOIR_FAIL;

  if(!strlen(tb))
    return tab_shell(ns);
  else if((tb[0]==',')||((tb[0]=='\'')))
    return tab_property(ns);
  else {
    char *p,*s=strchr(tb,' ');
    if(((p=strstr(tb,PROTOCOL_DELIMITER))!=NULL)&&
       ((s==NULL)||(p<s)))
      return tab_url(ns); }

  return tab_shell(ns); }



/****************************************************************************/
/* main */



/**
 * process contents of typebuffer
 * @return an error-code
 * @param ns the session
 */
int typebuffer_do(noir_sess *ns) {
  char   *b,*e,*tb,*p;
  char   *x[MAX_PARS+1];
  int     stat,c=0,ret=NOIR_FAIL;
  size_t  l,l_cwd;
  char   *typebuf=ns->typebuffer;

  if(!typebuf||!*typebuf||!ns)
    return NOIR_NODATA;

  if(!(tb=strdup(typebuf)))
    return NOIR_OOM;

  typebuffer_history_add(tb);

  if((tb[0]==',')||(tb[0]=='\'')) {           /* magic comma operator */
    ret=typebuffer_modify_theme(ns,tb);
    free(tb);
    return ret; }

  if(tb[0]=='?') {                            /* magic question mark */
    ret=typebuffer_dump_fi(ns,tb);
    free(tb);
    return ret; }



  /**************************************************************************/

  l=strlen(b=tb);
  l_cwd=strlen(ns->cwd);

  do {
    int done=0;

    typebuffer_get_token(ns,&b,&e,TRUE);
    if(!b)
      continue;

    /* URL ******************************************************************/
    if(strstr(b,PROTOCOL_DELIMITER)) { /* if it contains ://, we assume URL */
      if(!strncmp(b,PROTOCOL_FILE,7)) {
        if((x[c]=strdup(b+7)))
          done=++c; }
      else if(!c) {                  /* if URL is first el, use handler */
        done=1;
        typebuffer_handle_URL(ns,ns->typebuffer);
	goto tb_exit; }
      else {
        if((x[c]=strdup(b)))
          done=++c; }}

    /* file *****************************************************************/
    if(!done) {
      int    gr;
      glob_t myglob;
      char   myfile[PATH_MAX+1]="";

      /* absolute ***********************************************************/
      if(*b=='/') {
        if((ret=handle_relative(myfile,b,NULL))!=NOIR_SUCC)
          goto tb_exit; }

      /* relative to... what? ***********************************************/
      else {
        if(b[0]=='~') {
          if(!b[1]||(b[1]=='/')) {              /* ~ ~/foo */
            if((ret=handle_relative(myfile,home,&b[1]))!=NOIR_SUCC)
              goto tb_exit; }

          else {                                /* ~/azundris */
            struct passwd *pwd;
            char *e=strchr(b,'/');
            if((pwd=getpwnam(&b[1]))) {
              if((ret=handle_relative(myfile,pwd->pw_dir,e))!=NOIR_SUCC)
                goto tb_exit; }}}

        else if(!strcmp(b,DIR_MYSELF)||!strcmp(b,DIR_PARENT)|| /* ./ => current dir */
                strchr(b,DIR_DELIMITER)) {
          strcpy(myfile,ns->cwd);
          if((ret=handle_relative(myfile,"/",b))!=NOIR_SUCC)
            goto tb_exit; }

        else {                                  /* in $PATH? */
          char *bp=NULL;

          if(c) {
            size_t l=strlen(b)+strlen(ns->cwd)+2;
            if(l<PATH_MAX) {
              snprintf(myfile,l,"%s/%s",ns->cwd,b);
              if(file_exists(myfile))
                bp=myfile; }}

          if(bp)
            ;
          else if((bp=get_path_file(ns->cwd,b))) {
            int exc=(strlen(bp)>=PATH_MAX);
            if(!exc)
              strcpy(myfile,bp);
            free(bp);
            if(exc) {
              ret=NOIR_PATHEX;
              goto tb_exit; }}
          else if(strlen(b)<PATH_MAX)           /* neither, keep original */
            strcpy(myfile,b);
          else {
            ret=NOIR_PATHEX;
            goto tb_exit; }}}

      /* handle  /../ and /./ ***********************************************/
#ifdef WITH_DOT_DIR
      path_normalize(myfile,TRUE);
#else
      path_normalize(myfile,FALSE);
#endif

      /* vfolder rewrites ***************************************************/
      if((p=path_rewrite(myfile))) {
        if(strlen(p)<=PATH_MAX)
          strcpy(myfile,p);
        free(p); }

      /* handle glob ********************************************************/
      done=0;
      if(!(gr=glob(myfile,GLOB_BRACE|GLOB_TILDE,NULL,&myglob))) {
        if(myglob.gl_pathc>0) {
          int n,n2=myglob.gl_pathc;
          for(n=0;n<n2;n++) {
            if((x[c]=strdup(myglob.gl_pathv[n]))) {
              done=1;
              E(1,"glob{ %s -> %s }\n",myfile,x[c]);
              c++; }}}
        else
          E(2,"glob produced no results\n");
        globfree(&myglob); }

      if(!done)   /* glob produced no results, keep unchanged input */
        x[c++]=strdup(myfile);
    }     /* file (non-URL) */

    free(b);
    b=e;  /* skip token */
  } while(*e);



  /* done tokenizing, exec! *************************************************/

  x[c]=NULL;

  if(c) {                          /* is there something to do? (metal mix) */

    /* cd *******************************************************************/
    if(!strcmp(x[0],"quit")||!strcmp(x[0],"exit"))
      gui_quit();

    else if(!strcmp(x[0],"cd")||!strcmp(x[0],"ls")||!strcmp(x[0],"ll")) {
      ret=NOIR_SUCC;
      E(1,"built-in cd\n");
      if(!x[1])                           /* cd */
        ret=typebuffer_focus(ns,home);
      else if(!strcmp(x[1],"-")) {        /* cd - */
        if(!ns->pwd) {
          form_alert(ns,"No previous directory...");
          ret=NOIR_IGNORED; }
        else
          ret=typebuffer_focus(ns,ns->pwd); }
      else                                /* cd ~  cd whatever */
        ret=typebuffer_focus(ns,x[1]);
      goto tb_exit; }

    else if(!strcmp(x[0],"rm")) {
      form_alert(ns,"No rm yet...");
      ret=NOIR_IGNORED;
      goto tb_exit; }

    stat=file_exists(x[0]);

    if(S_ISDIR(stat))               /* it's a dir. focus. */
      ret=typebuffer_focus(ns,x[0]);

    else {                          /* it's a file. execute, if we can. */
      if((!stat)||(S_ISREG(stat)&&(stat&(S_IXUSR|S_IXGRP|S_IXOTH)))) {
        int n=0;                    /* it's an executable. execute. */
        while(n<c) {
          E(0,"%2d: \"%s\"\n",n,x[n]);
          n++; }
        file_fork(ns,x);
        ret=NOIR_SUCC; }
      else                          /* it's a data file. run w/ cqt app. */
        ret=file_activate(ns,x[0],NULL,NULL); }}

 tb_exit:

  if(debug>=0) {
    switch(ret) {
      case NOIR_PATHEX:
        E(0,"maximum path length exceeded...\n");
        break;
      default:
        break; }}

  while(--c>=0)
    free(x[c]);

  free(tb);

  return ret; }
