/** @file  ic_nautilus.c
 *  @brief plugin that creates an icon.
 *
 * this is an "advanced" icon-factory that uses the icon_set_state() method
 * to modify icons dynamically rather than to precompute the states.
 * this is the suggested procedure.
 *
 * visually, this will create icons that look a bit like those used by the
 * nautilus file-manager: text-labels appear below the icon-image; when
 * selected, the text-label is underlaid with a grey box with filed off
 * corners.
 *
 * long text-labels are wrapped; for each wrap, the text-size is decreased
 * by one.
 *
 * icons may be overlaid with automatic "emblems" indicating that an item
 * might not be accessible or writable by the current user.  in the default
 * theme, this is indicated by the stop-sign and eye of horus, respectively.
 *
 * icons may be overlaid with additional textual information; at this time,
 * the number of items contained in a folder is superimposed over the folder's
 * icon.  if the exact number of items is not known, an estimate is made
 * ("123+" rather than "4711").  this estimate coincides with the number
 * of folders contained in the folder (normally the number of hardlinks-2).
 *
 * items that are symlinks are "ghosted"; they appear translucent.
 *
 * items that evidence's thumbnailer was able to generate previews for will
 * be show a preview in lieu of a datafile-icon.
 *
 * the factory installs a variety of properties; see below.
 * these properties can be set in themes, or from the typebuffer, e.g.
 * 'icontextcolour #ff00ff
 */

#include <string.h>
#include <Evas.h>
#include <stdio.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>

#include "session.h"    /* noir_sess_focus() */
#include "debug.h"
#include "errors.h"
#include "config.h"
#include "theme.h"
#include "evas_macros.h"
#include "icon_list.h"
#include "icon_path.h"



/* when decreasing the text-size with each wrap,
   don't make the font smaller than this */
#define MIN_HEIGHT 7


#define LISTENER "iconfactory{nautilus}"

typedef enum {
  IC_TEXTFONT,  IC_TEXTSIZE,
  IC_TEXTCOL,   IC_DSCOL,     IC_SELCOL,
  IC_TEXTALPHA, IC_DSALPHA,   IC_SELALPHA,
  IC_DSXOFFS,   IC_DSYOFFS,
  IC_MLBLXOFFS, IC_MLBLYOFFS, IC_MLBLWIDTH,
  IC_GRID } ic_tag;

typedef struct _ic_param {
  icon_list *il;
  char *text_font;
  int   text_size;
  long  text_colour,ds_colour,sel_colour;
  int   text_alpha, ds_alpha, sel_alpha;
  int   ds_x_offset,ds_y_offset;
  int   mlbl_x_offset,mlbl_y_offset,mlbl_width;
  float grid;
  ic_mode icm_save;
} ic_param;



/******************************************************************************/



/**
 * listen for changes in our numeric properties.
 * @return NOIR_SUCC, or an error-code.
 * @param key - the name of the property that has changed
 * @param type - the (new) type of the property that has changed
 * @param tag - our index/tag for the property that has changed
 * @param data - our parameter-struct for this plugin, for faster access to the properties' values
 * @param te - the linked list of properties for this session
 */
static int ic_num_listener(const char *key,const noir_theme_el_type type,const int tag,void *data,noir_theme_el **te) {
  long           val=theme_el_get_int(te,key);
  float          fv;
  ic_param      *param=(ic_param *)data;
  noir_theme    *theme=(noir_theme *)param->il->theme;

  switch(tag) {
    case IC_GRID:
      fv=theme_el_get_float(te,key);
      if(param->grid!=fv) {
        if((param->grid=fv)) {
          param->icm_save=theme->icon_mode;
          theme->icon_mode=IC_AUTO; }
        else
          theme->icon_mode=param->icm_save;
        icon_list_layout_all(param->il); }
      break;

    case IC_TEXTSIZE:
      param->text_size=val; break;

    case IC_TEXTCOL:
      param->text_colour=val; break;
    case IC_TEXTALPHA:
      param->text_alpha=val; break;

    case IC_DSCOL:
      param->ds_colour=val; break;
    case IC_DSALPHA:
      param->ds_alpha=val; break;

    case IC_SELCOL:
      param->sel_colour=val; break;
    case IC_SELALPHA:
      param->sel_alpha=val; break;

    case IC_DSXOFFS:
      param->ds_x_offset=val; break;
    case IC_DSYOFFS:
      param->ds_y_offset=val; break;

    case IC_MLBLXOFFS:
      param->mlbl_x_offset=val; break;
    case IC_MLBLYOFFS:
      param->mlbl_y_offset=val; break;
    case IC_MLBLWIDTH:
      param->mlbl_width=val; break;

    default:
      E(0,"unknown tag %d (\"%s\")!?\n",tag,key?key:"<none>");
      return NOIR_FAIL; }

  E(1,"(%s::=%ld/#%06lx, %p)\n",key,val,val,data);

  return NOIR_SUCC; }



/**
 * listen for changes in our string-properties.
 * @return NOIR_SUCC, or an error-code.
 * @param key - the name of the property that has changed
 * @param type - the (new) type of the property that has changed
 * @param tag - our index/tag for the property that has changed
 * @param data - our parameter-struct for this plugin, for faster access to the properties' values
 * @param te - the linked list of properties for this session
 */
static int ic_str_listener(const char *key,const noir_theme_el_type type,const int tag,void *data,noir_theme_el **te) {
  char          *val=theme_el_get_string(te,key);
  ic_param      *param=(ic_param *)data;

  if(val) {
    switch(tag) {
      case IC_TEXTFONT:
        if(param->text_font)
          free(param->text_font);
        param->text_font=strdup(val);
        noir_sess_focus(param->il->sess,noir_sess_get_cwd(param->il->sess));
        break;
      default:
        E(0,"unknown tag %d (\"%s\")!?\n",tag,key?key:"<none>");
        return NOIR_FAIL; }

    E(1,"(%s::=%s, %p)\n",key,val,data);
    return NOIR_SUCC; }

  return NOIR_IGNORED; }






/******************************************************************************/
/* icon-factory */
/****************/



/**
 * change an icon's state.
 *
 * an icon-factory may create elements for all states when the icon is first
 * created; in this case, icon_set_state() needs not be provided.
 * using icon_set_state() does however allow for more flexibility (and reduced
 * use of memory) by creating the desired elements "on the fly."
 * @return TRUE if evas_icon_set_state() wrapper show show/hide icon-elements
 * based on their states after calling this, FALSE otherwise.  (usually FALSE)
 * @param il - the icon-list
 * @param icon - the icon in question
 * @param s - the desired state
 */
int icon_set_state(icon_list *il,icon *icon,icon_state s) {
  icon_state   c=s^icon->state;
  icon_el     *el;
  void        *o=NULL;
  Evas        *evas=get_evas_from_iconlist(il);
  Evas_Coord   _x,_y,_w,_h;
  static char *noxs=NULL,*readonly=NULL;
  int          tr,tg,tb,ta,
               sr,sg,sb,sa;
  ic_param    *param=((noir_theme *)il->theme)->icon_data;



  if(!noxs) {
    if(!(noxs=icon_path_get(NULL,"emblem/no-access"))) {
      E(0,"icon-set set no no-access emblem, falling back to default...\n");
      noxs=PACKAGE_DATA_DIR "/icons/emblems/noxs.png"; }}
  if(!readonly) {
    if(!(readonly=icon_path_get(NULL,"emblem/read-only"))) {
      E(0,"icon-set set no read-only emblem, falling back to default...\n");
      readonly=PACKAGE_DATA_DIR "/icons/emblems/readonly.png"; }}



  E(3,"%s: $%x -> $%x  ($%x)\n",icon->label,icon->state,s,c);



  /* make symlinks semi-transparent */
  if(icon->user1) {
    int alpha=((posixmd *)icon->user1)->lmode?159:255;
    for(el=icon->els;(el=icon_el_get_by_tag(el,ET_ICON_IMAGE,&o));el=el->next)
      evas_object_color_set(o,255,255,255, alpha); }



  /* noXS */
  if(icon->user2) {
    file_info *fi=(file_info *)icon->user2;
    int        nn=(fi->flags&(IF_NOUSE|IF_NOMODIFY))?1:0,no;

    no=(el=icon_el_get_by_tag(icon->els,ET_EMBLEM_IMAGE,&o))?1:0;

    if(no!=nn) {
      if(nn) {
        int   err;
        char *file=(fi->flags&IF_NOUSE)?noxs:readonly;

        o=evas_object_image_add(evas);
        evas_object_image_file_set(o,file,NULL);
        if((err=evas_object_image_load_error_get(o))!=EVAS_LOAD_ERROR_NONE) {
          E(0,"error #%d loading emblem \"%s\"...\n",err,file);
          evas_object_del(o);
          o=NULL; }
        if(o) {
          int w,h;
          evas_object_image_size_get(o,&w,&h);
          evas_object_resize(o,w,h);
          evas_object_image_fill_set(o,0,0,w,h);

          el=icon->els;

          ICON_EL_ADD(IE_ALL,50,0,0);
          el->tag=ET_EMBLEM_IMAGE;

          icon->els=el;

          if((el=icon_el_get_by_tag(icon->els,ET_EMBLEM_IMAGE,&o))) {
            icon->els->xo=el->xo+el->w/8;
            icon->els->yo=el->yo+el->h/8; }}}
      else
        icon_el_dst_by_tag(icon,ET_EMBLEM_IMAGE); }}



  /* keyboard selection gets gradient as drop-shadow */
  if(c&IE_KBD) {
    if(s&IE_KBD) {
      int         ht=icon->h;

      el=icon->els;

      o=evas_object_gradient_add(evas);
      ICON_EL_ADD(IE_KBD,-10,0,0);
      evas_object_resize(o,icon->w,ht);
      evas_object_gradient_color_add(o,  0,103,127,  0,  8);
      evas_object_gradient_color_add(o,  0,206,255, 63, 16);
      evas_object_gradient_angle_set(o,150);
      el->tag=ET_KEYED_IMAGE;

      icon->els=el; }
    else
      icon_el_dst_by_tag(icon,ET_KEYED_IMAGE); }



  /* selected items get inverted label */
  if(c&IE_SEL) {
    if(s&IE_SEL) {
      Evas_Object *z;
      Evas_Coord   _x,_y,_w,_h;
      noir_theme  *t=il->theme;
      Evas        *evas=get_evas_from_iconlist(il);
      int          tb=t->icon_height,
                   wt=icon->w,
                   ht=icon->h;

      el=icon->els;

      o=evas_object_rectangle_add(evas);
      evas_object_resize(o,wt,ht-tb);
      evas_object_color_set(o,55,55,55, 100);
      ICON_EL_ADD(IE_SEL,10,1,tb);
      el->tag=ET_LABEL_IMAGE;
      z=o;

      o=evas_object_rectangle_add(evas);
      evas_object_resize(o,wt+2,ht-tb-2);
      evas_object_color_set(o,96,96,96, 100);
      ICON_EL_ADD(IE_SEL,10,0,tb+1);
      el->tag=ET_LABEL_IMAGE;

      icon->els=el; }
    else
      icon_el_dst_by_tag(icon,ET_LABEL_IMAGE); }



  /* text / drop-shadow colours */
  if(s&IE_SEL) {
    sr=(param->text_colour>>16);
    sg=(param->text_colour>>8)&0xff;
    sb=(param->text_colour)&0xff;
    sa=param->text_alpha;

    tr=(param->ds_colour>>16);
    tg=(param->ds_colour>>8)&0xff;
    tb=(param->ds_colour)&0xff;
    ta=param->ds_alpha; }
  else {
    tr=(param->text_colour>>16);
    tg=(param->text_colour>>8)&0xff;
    tb=(param->text_colour)&0xff;
    ta=param->text_alpha;

    sr=(param->ds_colour>>16);
    sg=(param->ds_colour>>8)&0xff;
    sb=(param->ds_colour)&0xff;
    sa=param->ds_alpha; }

  for(el=icon->els;(el=icon_el_get_by_tag(el,ET_LABEL_TEXT_HI,&o));el=el->next)
    evas_object_color_set(o,tr,tg,tb, ta);

  for(el=icon->els;(el=icon_el_get_by_tag(el,ET_LABEL_TEXT_LO,&o));el=el->next)
    evas_object_color_set(o,sr,sg,sb, sa);



  /* redraw item */
  icon_move(icon,icon->x-il->sx,icon->y-il->sy);

  return FALSE; }



/**
 * create elements of an icon.
 * @return the first element of a linked list of icon-elements.
 * @param il - the icon-list to create the icon in
 * @param icon - the icon the elements are intended for
 * @param thumb - the path to a thumb-nailed image of the portrayed file's contents, or NULL
 */
icon_el *icon_els_create(icon_list *il,icon *icon,char *thumb) {
  noir_theme  *theme=(noir_theme *)il->theme;
  icon_el     *el=NULL,*ic;
  Evas        *evas=get_evas_from_iconlist(il);
  Evas_Object *o=NULL;
  Evas_Coord   _x,_y,_w,_h;
  int          w_,h_,wt=0,ht=0,tx=2,ty=0,iw,ih,err,links,fs_;
  float        gw;
  char        *fm,*to;
  posixmd     *intrinsic=(posixmd *)(icon?icon->user1:NULL);
  file_info   *fi=(file_info *)(icon?icon->user2:NULL);

  ic_param    *param=NULL;
  int          tr,tg,tb,ta,
               sr,sg,sb,sa;
  int          fs;
  int          mc=0; /* sanity check */

  icon->state&=~IE_THMB;

  if((o=evas_thumbnail_dead(theme,evas,fi))) {
    icon->state|=IE_THMB;
    evas_object_geometry_get(o,&_x,&_y,&_w,&_h);
    w_=_w; h_=_h; }

  if(!o) {
    if(!icon->name||!*icon->name)
      E(0,"cannot thumb-nail, and no fallback icon given...\n");
    else {
      o=evas_object_image_add(evas);
      evas_object_image_file_set(o,icon->name,NULL);
      if((err=evas_object_image_load_error_get(o))!=EVAS_LOAD_ERROR_NONE) {
        E(0,"error #%d loading \"%s\"...\n",err,icon->name);
        evas_object_del(o);
        o=NULL; }
      if(o) {
        evas_object_image_size_get(o,&w_,&h_);
        evas_object_resize(o,w_,h_);
        evas_object_image_fill_set(o,0,0,w_,h_); }}}

  if(!(param=theme->icon_data)) {
    if(!(theme->icon_data=param=malloc(sizeof(ic_param)))) {
      E(0,"out of memory!?\n");
      return NULL; }

    memset(param,0,sizeof(ic_param));

    param->il=il;

    theme_el_default(&theme->data,"iconfont","vera.ttf",0,0,0);
    theme_el_default(&theme->data,"iconfontsize","16",3,128,1);
    theme_el_default(&theme->data,"icontextcolour","#0",0,0,0);
    theme_el_default(&theme->data,"icontextalpha","255",0,255,1);
    theme_el_default(&theme->data,"iconshadowcolour","#ffffff",0,0,0);
    theme_el_default(&theme->data,"iconshadowalpha","129",0,255,1);
    theme_el_default(&theme->data,"iconshadowxoffset","1",-64,64,1);
    theme_el_default(&theme->data,"iconshadowyoffset","1",-64,64,1);
    theme_el_default(&theme->data,"iconmetalabelxoffset","17",0,512,1);
    theme_el_default(&theme->data,"iconmetalabelyoffset","1",0,512,1);
    theme_el_default(&theme->data,"iconmetalabelwidth","92",0,512,1);
    theme_el_default(&theme->data,"iconlistgrid","-1.0",-4,128,0.01);

    theme_el_listen(&theme->data,LISTENER,"iconfont",            ic_str_listener,IC_TEXTFONT,param);
    theme_el_listen(&theme->data,LISTENER,"iconfontsize",        ic_num_listener,IC_TEXTSIZE,param);
    theme_el_listen(&theme->data,LISTENER,"icontextcolour",      ic_num_listener,IC_TEXTCOL,param);
    theme_el_listen(&theme->data,LISTENER,"icontextalpha",       ic_num_listener,IC_TEXTALPHA,param);
    theme_el_listen(&theme->data,LISTENER,"iconshadowcolour",    ic_num_listener,IC_DSCOL,param);
    theme_el_listen(&theme->data,LISTENER,"iconshadowalpha",     ic_num_listener,IC_DSALPHA,param);
    theme_el_listen(&theme->data,LISTENER,"iconshadowxoffset",   ic_num_listener,IC_DSXOFFS,param);
    theme_el_listen(&theme->data,LISTENER,"iconshadowyoffset",   ic_num_listener,IC_DSYOFFS,param);
    theme_el_listen(&theme->data,LISTENER,"iconmetalabelxoffset",ic_num_listener,IC_MLBLXOFFS,param);
    theme_el_listen(&theme->data,LISTENER,"iconmetalabelyoffset",ic_num_listener,IC_MLBLYOFFS,param);
    theme_el_listen(&theme->data,LISTENER,"iconmetalabelwidth",  ic_num_listener,IC_MLBLWIDTH,param);
    theme_el_listen(&theme->data,LISTENER,"iconlistgrid",        ic_num_listener,IC_GRID,param); }

  iw=w_;
  ih=h_;

  fs=param->text_size;

  ICON_EL_PRE((IE_DFL|IE_SEL),0,0,(theme->icon_height-h_)/2.0);
  mc++;
  el->tag=ET_ICON_IMAGE;
  ic=el;

  if(!theme->icon_height) {
    theme->icon_height=ih; }
  if(!theme->icon_width) {
    theme->icon_width=iw; }

  if((gw=param->grid)<0)
    gw=-gw*((float)theme->icon_width);

  ty=theme->icon_height;

  { fs_=fs/6;  /* re-scale font */
    fs_=fs_?fs_:1;
    fm=icon->label;
    while((fm=strchr(fm,'\n'))) {
      if((fs-fs_)>=MIN_HEIGHT)
        fs-=fs_;
      else
        fs=MIN_HEIGHT;
      fm++; }}

  long_to_rgb(param->text_colour,&tr,&tg,&tb);
  ta=param->text_alpha;

  long_to_rgb(param->ds_colour,&sr,&sg,&sb);
  sa=param->ds_alpha;

  fm=icon->label;        /* support text-wrap */
  while(fm) {
    if((to=strchr(fm,'\n')))
      *to='\0';

    o=evas_object_text_add(evas);
    fs_=fs;
  try_text:
    evas_object_text_text_set(o,fm);
    evas_object_text_font_set(o,param->text_font,fs);
    evas_object_geometry_get(o,&_x,&_y,&_w,&_h);
    if(param->grid!=0) {
      if((w_=gw)<=_w) {
        int fd=(fs/6);
        fs-=(fd<1)?1:fd;
        if(fs>=MIN_HEIGHT)
          goto try_text; }}

    if(w_>_w)
      tx=((w_-_w)/2.0);
      
    evas_object_color_set(o,tr,tg,tb, ta);        /* black */
    ICON_EL_PRE((IE_DFL|IE_KBD),20,tx,ty);
    mc++;
    el->tag=ET_LABEL_TEXT_HI;

    o=evas_object_text_add(evas);
    evas_object_text_font_set(o,param->text_font,fs);
    evas_object_text_text_set(o,fm);
    evas_object_color_set(o,sr,sg,sb, sa);  /* white/trans */
    ICON_EL_PRE((IE_DFL|IE_KBD),18,tx+param->ds_x_offset,ty+param->ds_y_offset);
    mc++;
    el->tag=ET_LABEL_TEXT_LO;

    fs=fs_;

    if(to) {
      *(to++)='\n';
      fm=to;
      ty+=fs; }
    else
      fm=NULL; }



/* For the image portion to be of the same size, the icon-mode must be
   "auto" (see config-files).  The icon-factory must then simply make
   sure labels are all the same size; here, this is done by scaling
   down the text until it fits.  Then, an invisible evas box is put
   underneath the icon's image that's the size of one grid box.  If
   the text is shorter, the wider (invisible) box will keep the
   overall bounding box of the icon wide, the same width as all the
   other icons. */

  if(param->grid) {
    o=evas_object_rectangle_add(evas);
    evas_object_resize(o,gw,theme->icon_height);
    evas_object_color_set(o,55,55,55, 0);
    ICON_EL_PRE(IE_SEL,10,0,0);
    el->tag=ET_MISC; }



  /* additional labels (folders contained etc.) ******************************/

  icon->w=wt+4;
  icon->h=ht;

  { char b[9],*p=b;

    if(fi&&fi->children) {
      links=fi->children;
      snprintf(b,7,"%d",links<999999?links:999999); }
    else if(intrinsic&&S_ISDIR(intrinsic->mode)) {
      links=intrinsic->hardlinks-
            (((theme->tf&TF_DOTDIR)&&!(fi->flags&IF_DOTDIR))?1:2);
      snprintf(b,7,"%d+",links<999999?links:999999); }
    else {
      links=-1;
      p=(fi&&fi->sortkey)?fi->sortkey:NULL; } /* item with meta-data sortkey */

    if(p) {
      o=evas_object_text_add(evas);
      evas_object_text_font_set(o,param->text_font,fs);
      evas_object_text_text_set(o,p);
      evas_object_geometry_get(o,&_x,&_y,&_w,&_h);
      tx=ic->xo+((ic->w-_w)/2);

      if(links>=0) {
        ty=(ic->h-_h)/2;
        evas_object_color_set(o,255,255,255, 129); }  /* white/trans */
      else {
        if(_w>param->mlbl_width) {
          evas_object_resize(o,_w=param->mlbl_width,_h);
          tx=ic->xo+((ic->w-_w)/2); }
        if(tx<param->mlbl_x_offset)
          tx=param->mlbl_x_offset;
        ty=param->mlbl_y_offset;
        evas_object_color_set(o,0,0,0, 129); }        /* black/trans */
      ICON_EL_PRE((IE_DFL|IE_KBD),18,tx+1,ty+1);
      mc++;
      if(links<0) {
        el->xo=(int)((wt-el->w)/2); }                 /* center label */
      el->tag=ET_ICON_TEXT_LO; }}



  ic->xo=(int)((wt-ic->w)/2);  /* center image */

#if 0
  /* check whether we built a consistent chain. */
  { icon_el *e2=el;
    while(e2) {
      mc--;
      e2=e2->next; }
    if(mc)
      E(0,"icon-els mismatch: %d\n",mc); }
#endif

  icon->els=el;

  return el; }
