/** @file  clipboard.c
 *  @brief handle the clipboard and the primary selection.
 */

#include <string.h>
#include <stdlib.h>                            /* malloc(), free() */

#include "model.h"                             /* struct client_funs */
#include "errors.h"
#include "src/gui/common/session_private.h"
#include "src/gui/common/alert.h"              /* form_alert() */
#include "src/gui/common/typebuffer.h"         /* TB_MAX, for pasting from clipboard */



/****************************************************************************/



/**
 * set text 'ascii' on clipboard 'cb'.
 * @param cb - the clipboard to modify
 * @param ascii - the text to set
 */
int clipboard_set(GdkAtom cb,char *ascii) {
  gsize   foo;
  gchar  *utf8;
  E(2,"(\"%s\")\n",ascii);
  if((utf8=g_locale_to_utf8(ascii,-1,&foo,&foo,NULL))) {
    gtk_clipboard_set_text(gtk_clipboard_get(cb),utf8,-1);
    g_free(utf8);
    return NOIR_SUCC; }
  else
    return NOIR_FAIL; }



/**
 * get text from clipboard and insert it into type-buffer.
 * @param cb - the clipboard to read from
 * @param ns - the session (text will be set in ns->typebuffer)
 */
int clipboard_get(GdkAtom cb,noir_sess *ns) {
  gchar *cs=gtk_clipboard_wait_for_text(gtk_clipboard_get(cb));
  if(cs) {
    size_t l=strlen(ns->typebuffer)+strlen(cs);
    if(l>ns->typebuffer_max) {
      char *n;
      l+=TB_MAX;
      if(!(n=malloc(l+1))) {
        form_alert(ns,"Cannot paste primary selection, out of memory!?");
        return NOIR_OOM; }
      else {
        strcpy(n,ns->typebuffer);
        free(ns->typebuffer);
        ns->typebuffer=n;
        ns->typebuffer_max=l; }}
    if(l<ns->typebuffer_max) {
      strcat(ns->typebuffer,cs);
      if(ns->typebuffer_open)
        typebuffer_set(ns); }
    g_free(cs);
    return NOIR_SUCC; }

  form_alert(ns,"Could not obtain primary selection -- is it empty?");
  return NOIR_FAIL; }



/**
 * set primary selection for session's selection.
 * @param ns - the session to set the primary selection from
 */
static inline int selection_set_primary(noir_sess *ns) {
  char   *p=ns->selection;
  char   *n,*q,*t,*m;
  size_t  l;

  if(p) {
    l=strlen(p);

    while(*p) {
      switch(*p) {
      case '\n': l+=3; break;
      case '\"': l+=2; break;
      default:   l++; }
      p++; }

    if(!(t=m=malloc(l+3))) {
      form_alert(ns,"Cannot set primary selection, out of memory...");
      return NOIR_OOM; }

    p=ns->selection;
    strcpy(t,"\"");

    while(p&&*p) {
      t=strchr(t,'\0');
      q=strchr(p,'\"');
      n=strchr(p,'\n');
      if(!q||(n<q)) {
        if(n) {
          *n='\0';
          strcat(strcpy(t,p),"\" \"");
          *(n++)='\n'; }
        else
          strcpy(t,p);
        p=n; }
      else {
        if(q) {
          *q='\0';
          strcpy(t,p);
          *(q++)='\"';
          *(t++)='\\';
          *(t++)='\"'; }
        else
          strcpy(t,p);
        p=q; }}

    strcat(t,"\"");

    clipboard_set(GDK_SELECTION_PRIMARY,m);

    free(m);

    return NOIR_SUCC; }

  gtk_clipboard_clear(gtk_clipboard_get(GDK_SELECTION_PRIMARY));

  return NOIR_SUCC; }



/**
 * copy selection to all views in a session.
 * @param ns - the session
 * @param skip - the calling client (don't call setter on that, it already knows)
 */
int selection_update(noir_sess *ns,client_funs *skip) {
#if 0
  client_funs *cf=ns->clients;
#endif

  if(ns->selection)
    E(2,"%s\n",ns->selection);
  else
    E(3,"<empty>\n");

#if 0
  /* deferred, will update when switching pages */
  while(cf) {
    if((cf!=skip)&&(cf->set_selection))
      cf->set_selection((void *)ns);
    cf=cf->next; }
#endif

  return selection_set_primary(ns); }



/**
 * remove 'file' from a session's selection.
 * @param ns - the session
 * @param file - the file
 */
static inline int selection_remove_do(noir_sess *ns,char *file) {
  char   *p,*q;
  size_t  l;
  int     ret=NOIR_NOTFOUND;

  if(!file||!*file)
    return NOIR_NODATA;
  if(!(p=ns->selection))
    return NOIR_IGNORED;

  l=strlen(file);
  while((p=strstr(p,file))) {
    if(((p==ns->selection)||(p[-1]=='\n'))&&
       ((p[l]=='\n')||(!p[l]))) {
      if(p==ns->selection) {          /* is first element */
        if(l==strlen(ns->selection))  /* is only element */
          *p='\0';
        else {                        /* is first of several */
          q=p+l+1;
          memmove(p,q,strlen(q)+1); }}
      else {                          /* is not first element */
        q=p+l;
        memmove(p-1,q,strlen(q)+1); }
      ret=NOIR_SUCC; }
    p++; }

  return ret; }



/**
 * remove 'file' from a session's selection.
 * @param ns - the session, or NULL to remove from all sessions' selections
 * @param file - the file
 */
int selection_remove(noir_sess *ns,char *file) {
  if(ns)
    return selection_remove_do(ns,file);
  else {
    int ret=NOIR_NOTFOUND;
    for(ns=noir_sess_get_1st();ns;ns=noir_sess_get_next(ns)) {
      if(selection_remove_do(ns,file)==NOIR_SUCC)
        ret=NOIR_SUCC; }
    return ret; }}



/**
 * check whether a given session's selection contains a specific file.
 * @return NOIR_SUCC (found), NOIR_NOTFOUND (not found), NOIR_NODATA (no file
 * given), or NOIR_IGNORED (no current selection).
 * @param ns - the session
 * @param file - the file
 */
int selection_contains(noir_sess *ns,char *file) {
  char   *p;
  size_t  l;

  if(!file||!*file)
    return NOIR_NODATA;
  if(!(p=ns->selection))
    return NOIR_IGNORED;

  l=strlen(file);
  while((p=strstr(p,file))) {
    if(((p==ns->selection)||(p[-1]=='\n'))&&
       ((p[l]=='\n')||(!p[l])))
      return NOIR_SUCC;
    p++; }

  return NOIR_NOTFOUND; }



/****************************************************************************/
