/** @file  dnd.c
 *  @brief Drag-N-Drop related functions.
 */

#include "config.h"

#include <gtk/gtk.h>
#include <string.h>
#include <sys/stat.h>   /* S_ISDIR */
#include <stdlib.h>     /* malloc(), free() */

#ifdef HAVE_GNOMECANVAS
#  include "gnome-icon-list.h"
#endif

#include "src/gui/common/session_private.h"
#include "src/gui/common/alert.h"      /* form_alert() */

#include "dnd.h"
#include "errors.h"
#include "util.h"       /* esprintf() */
#include "file.h"       /* file_...() */
#include "file_lcl.h"   /* file_to_URL */



typedef struct _dt_menu {
  noir_sess *ns;
  char      *src,
            *tgt;
  int        typ; } dt_menu;



GdkAtom *target_atoms=NULL;

const GtkTargetEntry target[]={
  { "STRING",              0,DRAGDROP_TARGET_STRING },
  { "_NETSCAPE_URL",       0,DRAGDROP_TARGET_NETSCAPE_URL },
  { "text/uri-list",       0,DRAGDROP_TARGET_URL },
  { "text/plain",          0,DRAGDROP_TARGET_TEXT },
  { "FM_SELECTION",        0,DRAGDROP_TARGET_FM_SELECTION },
  { "application/x-color", 0,DRAGDROP_TARGET_COLOUR },
  { "XdndDirectSave0",     0,DRAGDROP_TARGET_XDS }}; /* XDS always last */
const guint targets=(guint)(sizeof(target)/sizeof(GtkTargetEntry))-1;
const guint targets_xds=
#ifdef WITH_XDS
                    (guint)(sizeof(target)/sizeof(GtkTargetEntry));
#else
                    (guint)(sizeof(target)/sizeof(GtkTargetEntry))-1;
#endif






/*****************************************************************************/



#ifdef WITH_XDS
/* XDS-functions adapted by Azundris from ROX-Filer, by the ROX-Filer team. */

#define MAXURILEN 4096		/* Longest URI to allow */

GdkAtom XdndDirectSave0,xa_text_plain;



/**
 * set XDS property
 * @param context - the drag-context. the XdndDirectSave0-property of context->source_window will be modified.
 * @param text - the string to set the property to.
 */
void xds_set_prop(GdkDragContext *context,const char *text) {
  if(text)
    gdk_property_change(context->source_window,XdndDirectSave0,xa_text_plain,8,
                        GDK_PROP_MODE_REPLACE,text,strlen(text));
  else
    gdk_property_delete(context->source_window,XdndDirectSave0); }



/**
 * read XDS property
 * @return the XdndDirectSave0-property of context->source_window will be returned.
 * @param context - the drag-context.
 */
char *xds_get_prop(GdkDragContext *context) {
  guchar *prop_text;
  int     length;

  if(gdk_property_get(context->source_window,XdndDirectSave0,xa_text_plain,0,
                      MAXURILEN,FALSE,NULL,NULL,&length,&prop_text)&&prop_text)
  {
    prop_text=g_realloc(prop_text,length+1);
    prop_text[length]='\0';
    return prop_text; }

  return NULL; }

#endif



/*****************************************************************************/
/* DND helpers */



/**
 * DND: Is the sender willing/able to supply this target type?
 * @return TRUE/FALSE
 * @param dc   the drag-context
 * @param tgt  the target-type we want to check for
 */
gboolean dnd_src_speaks(GdkDragContext *dc,GdkAtom tgt) {
  GList *tl=dc->targets;
  while(tl&&((GdkAtom)tl->data!=tgt))
    tl=tl->next;
  return (tl!=NULL); }



/**
 * DND: drag completed, negotiate content-type.
 * User has tried to drop some data on us. Decide what format we would like
 * the data in.
 * @param widget     the widget
 * @param dc         the drag-context
 * @param timestamp  when the event happened
 * @param where      file-system path of the target object (item or "background")
 */
gboolean dnd_drag_drop(GtkWidget *widget,GdkDragContext *dc,guint timestamp,char *where) {
  char       *msg=NULL;
  GdkAtom     tgt=GDK_NONE;

  if(!where)
    return TRUE;

#ifdef WITH_XDS
  if(dnd_src_speaks(dc,XdndDirectSave0)) {
    char *what=NULL;
    if((what=xds_get_prop(dc))) {
      if(strchr(what,DIR_DELIMITER)) {
        msg="XDS protocol msg: item may not contain '/'\n";
        g_free(what); }
      else {
        char *x=NULL;
        esprintf(&x,"file://%s%s/%s",get_hostname(),where,what);
        if(x) {
          E(0,"XDS: target \"%s\"\n",x);
          xds_set_prop(dc,x);
          tgt=XdndDirectSave0;
          *x='\0'; /* free() here? */
          free(x); }
        else
          msg="out of memory"; }}
    else
      msg="XdndDirectSave0 target provided, but the atom "
          "XdndDirectSave0 (type text/plain) did not "
          "contain the desired name.\n"; }

  /* no XDS */
  else
#endif
  { GList *tl;
    int    n=0;

    while(n<targets) {
      tl=dc->targets;;
      while(tl&&((GdkAtom)tl->data!=target_atoms[n]))
        tl=tl->next;
      if(!tl)
        n++;
      else
        n=targets; }

    if(tl)
      tgt=(GdkAtom)tl;
    else
      msg="couldn't agree on a data-format..."; }

  if(msg) {
    E(0,"%s",msg);
    gtk_drag_finish(dc,FALSE,FALSE,timestamp); }
/*
  else
    gtk_drag_get_data(widget,dc,tgt,timestamp);
*/
  return TRUE; }



/**
 * make URL from 'file' and append to 'list'.
 * icon_drag_data_get() will send the list to the drop-target.
 * @param list - the list
 * @param file - the file
 */
gchar *dnd_send_URL(gchar *list,gchar *file) {
  gchar *new=NULL;
  if(file) {
    char *myfile=file_to_URL(file);
    if(myfile) {
      if(list) {
        if((new=malloc(strlen(file)+strlen(list)+2))) {
          strcat(strcat(strcpy(new,list),DELIMITER),file);
          g_free(list); }}
      else
        new=strdup(file);
      free(myfile); }}
  return new; }



/**
 * add 'file' to 'list'
 * icon_drag_data_get() will send the list to the drop-target.
 * @param list - the list
 * @param file - the file
 */
gchar *dnd_send_string(gchar *list,gchar *file) {
  gchar *new=NULL;
  if(file) {
    if(list) {
      if((new=malloc(strlen(file)+strlen(list)+2))) {
        strcat(strcat(strcpy(new,list),DELIMITER),file);
        g_free(list); }}
    else
      new=strdup(file); }
  return new; }



/**
 * a file or selection was dropped, all questions (if any) were asked, now
 * take action (and move, copy, link, whatever).
 * @param ns - the session
 * @param dtm - the dt_menu struct; will be released
 */
static int file_drop_do(noir_sess *ns,dt_menu *dtm) {
  int   ret=NOIR_SUCC;
  char *file;

  if(!dtm)
    return NOIR_FAIL;

  if(dtm->src&&dtm->tgt) {
    if(!strcmp(dtm->src,dtm->tgt)) {   /* item was dropped on top of itself */
      E(1,"item dropped on top of itself, ignored...\n");
      ret=NOIR_IGNORED; }

    else if(S_ISDIR(file_exists(dtm->tgt))) {
      switch(dtm->typ) {
      case GDK_ACTION_MOVE:
	ret=file_job_create(ns,NOIR_JOB_MV,dtm->src,dtm->tgt,NULL);
        break;
      case GDK_ACTION_COPY:
	ret=file_job_create(ns,NOIR_JOB_CP,dtm->src,dtm->tgt,NULL);
        break;
      case GDK_ACTION_LINK:
        ret=file_symlink(dtm->src,dtm->tgt);
        break; }}

    else {
      if(dtm->typ!=GDK_ACTION_COPY) {
        form_alert(NULL,"Can only use link and move with folders.");
        ret=NOIR_FAIL; }
      else {
        int r=NOIR_IGNORED;
        while((file=file_get_from_drop(&(dtm->src))))
          file_partial(file_activate(NULL,file,NULL,dtm->tgt),&r);
        ret=(r==NOIR_IGNORED)?NOIR_FAIL:r; }}}
  else
    ret=NOIR_NODATA;

  if(dtm->src)
    free(dtm->src);
  if(dtm->tgt)
    free(dtm->tgt);
  free(dtm);

  return ret; }



/**
 * views can connect "drag-end" to this.
 * releases DND-info.
 */
void file_drop_free(GtkWidget *widget,GdkDragContext *dc,gpointer data) {
  noir_sess *ns=data;

  if(!ns||(ns->drag_context!=dc)||!ns->drag_files)
    return;

  g_free(ns->drag_files);
  ns->drag_context=NULL;
  ns->drag_files=NULL; }



/**
 * views may connect "drag_data_delete" to this.
 */
void file_drop_delete_from_iconlist(GtkWidget *widget,GdkDragContext *dc,gpointer data) {
  noir_sess *ns=data;
  char      *file,*p,*a;

  if(dc==ns->drag_context) {
    if((a=p=strdup(ns->drag_files))) {
      while((file=file_get_from_drop(&p))) {
#if defined(HAVE_GNOMECANVAS)&&!defined(STANDALONE)
        int no;
        char *i=NULL;
        for(no=0;i;no++) {
          if((i=gnome_icon_list_get_icon_filename(GNOME_ICON_LIST(widget),no))&&!strcmp(i,file)) {
            gnome_icon_list_remove(GNOME_ICON_LIST(widget),no);
            i=NULL; }}
#endif
      }
      free(a); }}}



/**
 * remember what data we sent the drop target.
 * @param ns - session
 * @param dc - drag-context
 * @param reply - the info we sent
 */
void file_drop_remember(noir_sess *ns,GdkDragContext *dc,gchar *reply) {
  static int sema=0;

  if(sema>0) { /* we only allow one drag at a time... */
    E(0,"already running, denied.\n");
    return; }

  sema++;

  if(ns->drag_files)
    E(0,"cannot set drag_context to \"%s\", already set to \"%s\"...\n",reply,ns->drag_files);
  while(ns->drag_files)
    mainloop_once();
  ns->drag_context=dc;
  ns->drag_files=reply;

  sema--; }



/**
 * callback for drop-actions menu.
 * @param item - the GTK menu-item that was selected. "data" property should contain action code.
 * @param data - the drop-context
 */
static gint dropactions_menu_selected(GtkWidget *item,gpointer data) {
  dt_menu *dtm=(dt_menu *)data;
  if(dtm) {
    dtm->typ=(int)GPOINTER_TO_INT(g_object_get_data(G_OBJECT(item),"data"));
    return TRUE; }
  else
    return FALSE; }



/**
 * add an entry to drop-actions menu.
 * @param menu - the GTK menu to add to
 * @param dtm - the drop-context
 * @param label - the label for the new item
 * @param code - the action-code this item will emit when activated (cp, ln, mv, ...)
 * this code will be set on the item's "data" property.
 */
static void dropactions_add(GtkWidget *menu,dt_menu *dtm,char *label,int code) {
  GtkWidget *item;
  item=gtk_menu_item_new_with_label(label);
  g_object_set_data(G_OBJECT(item),"data",GINT_TO_POINTER(code));
  g_signal_connect(item,"activate",G_CALLBACK(dropactions_menu_selected),dtm);
  gtk_menu_shell_append((GtkMenuShell *)menu,item); }



/**
 * create and run a drop-actions ("mv, ln, cp?") menu.
 * @param dtm - the drop-context (what sources, what target, ...)
 */
static int dropactions_menu(dt_menu *dtm) {
  GtkWidget *menu=gtk_menu_new();

  dropactions_add(menu,dtm,"Move",GDK_ACTION_MOVE);
  dropactions_add(menu,dtm,"Link",GDK_ACTION_LINK);
  dropactions_add(menu,dtm,"Copy",GDK_ACTION_COPY);

  gtk_widget_show_all(menu);
  gtk_menu_popup(GTK_MENU(menu),NULL,NULL,NULL,NULL,1,0);

  return NOIR_SUCC; }



/**
 * something was dropped on evidence, take action.
 * @param ns - session
 * @param type - what kind of action? (cp, mv, ln, ...)
 * @param files - what was dropped?
 * @param target - where was it dropped?
 */
int file_drop(noir_sess *ns,int type,char *files,char *target) {
  dt_menu *dtm=malloc(sizeof(dt_menu));

  if(!dtm||(type==GDK_ACTION_PRIVATE))
    return FALSE;

  E(3,"#%d: \"%s\" on \"%s\" in \"%s\"\n",type,files,target,ns->cwd);

  dtm->src=strdup(files);
  dtm->tgt=strdup(target);

  switch(type) {
  case GDK_ACTION_DEFAULT:
    dtm->typ=GDK_ACTION_COPY;
    break;
  case GDK_ACTION_MOVE:
  case GDK_ACTION_COPY:
  case GDK_ACTION_LINK:
    dtm->typ=type;
    break;
  case GDK_ACTION_ASK:
  default:
#if 0
    if(S_ISDIR(file_exists(target)))
      dtm->typ=GDK_ACTION_COPY;
    else {
#else
    {
#endif
      dropactions_menu(dtm);
      while(dtm->typ==GDK_ACTION_ASK)
        mainloop_once(); }
    break; }

  return file_drop_do(ns,dtm); }



/**
 * init drag-n-drop subsystem
 * @return an error-code
 * @param ns  the session
 */
int dnd_init(noir_sess *ns) {
  int n;

#ifdef WITH_XDS
  XdndDirectSave0=gdk_atom_intern("XdndDirectSave0",FALSE);
  xa_text_plain  =gdk_atom_intern("text/plain",FALSE);
#endif

  if(!(target_atoms=malloc(sizeof(GdkAtom)*targets)))
    return NOIR_OOM;

  for(n=0;n<targets;n++)
    target_atoms[n]=gdk_atom_intern(target[n].target,FALSE);

  return NOIR_SUCC; }



/**
 * de-init drag-n-drop subsystem
 * @return an error-code
 * @param ns  the session
 */
int dnd_exit(noir_sess *ns) {
  free(target_atoms);
  return NOIR_IGNORED; }
