/** @file  menu_gtk.c
 *  @brief implements the GTK2 specific part of the context-menus.
 */

#include <gtk/gtk.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>                     /* malloc(), free() */

#include "gui/common/session.h"         /* noir_sess_get_selection() */
#include "gui/common/session_private.h" /* struct _noir_sess  (ns->sortflags) */
#include "gui/common/menu.h"            /* struct menu_action */
#include "gui/common/menu_private.h"    /* struct selection_action */
#include "gui/common/alert.h"           /* form_alert() */

#include "gui/gtk2/common/tooltip.h"    /* info_tooltip_cancel() */
#include "gui/gtk2/common/notebook.h"   /* notebook_go2view() */

#include "mime.h"
#include "meta.h"            /* meta_in_group(), meta_add(), meta_del_all() */
#include "util.h"            /* mainloop_once() */
#include "action.h"          /* action_context_*(), action_file_*() */
#include "file.h"            /* file_exists(), file_run(),
                                file_get_meta, file_get_mime() */



extern int go2view_mapping[];



/******************************************************************************/

/* THE CONTEXT-MENU.
 * g_object_*_data(menu_item, "data", meta **meta_data_anchor)
 */


/**
 * closing menu. release all admin data.
 */
static void menu_done(GtkMenuShell *menushell,gpointer data) {
  action_context *ac=(action_context *)data;
  ac->flags|=AF_DONE; }



/******************************************************************************/
/* context-menu for the current selection */



/**
 * apply action from context-menu to a single file.
 *
 * called from menu_selection_apply().
 * @param file  the file to apply the action to
 * @param item  the GTK menu item
 * @param menu  the action to be applied to the file
 * @param ac    the action context
 */
static inline gint menu_selection_apply_one(char *file,GtkWidget *item,menu_action *menu,action_context *ac) {
  gint         ret=FALSE;
  meta        *md=NULL,**mp=NULL;
  noir_sess   *ns=noir_sess_get_by_serial(ac->serial);
  action_file **ap=NULL,*as=NULL;

  if(!strcmp(ac->single->tag,file))   /* try single */
    ap=&ac->single;

  as=ac->selection;
  while(as&&strcmp(as->tag,file))     /* try selection */
    as=as->next;

  /* use file from selection if a) desired file for isn't the file that the
     context-menu was called up from, or b) if it doesn't bring meta-data,
     but the entry in the selection does.
     if neither have meta-data, or the desired file is in neither set,
     we'll read the meta-data on the fly. */
  if(as&&(!ap||!(*ap)->md))
    ap=&as;

  mp=ap?&(*ap)->md:&md;

  if(!*mp) {                          /* still no meta-data? read them now! */
    int r2;
    E(1,"reading MD for \"%s\"...\n",file);
    if((r2=file_get_meta_only(ns,file,mp))!=NOIR_SUCC)  /* read MD if we really must */
      E(0,"could not get meta-data for \"%s\"... (%d)\n",file,r2); }
  else
    E(1,"already have MD for \"%s\"...\n",file);

  if(*mp)
    ret=(menu_action_by_entry(ac->serial,menu,mp,ac)==NOIR_SUCC);

  if(md) { /* if we allocated meta-data on the fly... */
    if(ap&&!(*ap)->md)  /* save it if we have somewhere to save it to... */
      (*ap)->md=md;
    else                /* otherwise release it */
      meta_del_all(&md); }

  return ret; }



/**
 * apply action from context-menu (to one or many).
 *
 * @param item - the GTK menu-item
 * @param data - the action-context
 */
static gint menu_selection_apply_many(GtkWidget *item,gpointer data) {
  menu_action    *m=(menu_action *)g_object_get_data(G_OBJECT(item),"data");
  action_context *ac=(action_context *)data;
  action_file    *af=NULL;

  if(!ac||!(af=ac->selection)) {
    E(0,"insufficient data: ac %p, af %p\n",ac,af);
    return FALSE; }

  ac->flags|=AF_SELECTION; /* we have more than one file, set "selection" flag */
  ac->flags&=(~AF_LAST);   /* we have more than one file, clear "last" flag */

  while(af) {
    if(!(ac->flags&AF_CANCEL_ALL)) {
      if(!af->next)            /* if it's the last item, flag it */
        ac->flags|=AF_LAST;
      menu_selection_apply_one(af->tag,item,m,ac); }
    af=ac->selection?af->next:NULL; }  /* action may steal selection */

  return FALSE; }



/**
 * add action 'title' to context-menu 'menu' for current selection.
 *
 * called from menu_selection_create().
 * @param ac     the action-context
 * @param menu   the GTK menu to add to
 * @param title  the title for the menu entry
 * @param data   the data to set on the entry (an action (menu_action *), or NULL)
 */
static GtkWidget *menu_selection_add(action_context *ac,GtkWidget *menu,char *title,gpointer data) {
  GtkWidget *item=gtk_menu_item_new_with_label(title);
  if(!item) {
    E(0,"failed to create item for \"%s\"!?\n",title);
    return NULL; }
  g_signal_connect(item,"activate",G_CALLBACK(menu_selection_apply_many),ac);
  g_object_set_data(G_OBJECT(item),"data",data);
  gtk_menu_shell_append((GtkMenuShell *)menu,item);
  gtk_widget_show(item);
  gtk_widget_set_sensitive(item,TRUE);
  return item; }



/**
 * set menu item insensitve (disabled)
 * @param item  the GTK menu-item to disable
 */
static GtkWidget *menu_selection_insensitive(GtkWidget *item) {
  GdkColor     c;
  GtkStateType t=GTK_STATE_INSENSITIVE|GTK_STATE_NORMAL;
  gdk_color_parse("blue",&c);
  gtk_widget_modify_fg(gtk_bin_get_child(GTK_BIN(item)),t,&c);
  gtk_widget_modify_fg(item,t,&c);
  gtk_widget_set_sensitive(item,FALSE);
  return item; }



/**
 * create a sub-menu
 * @return the sub-menu, or NULL
 * @param menu  the menu to add the sub-menu to
 * @param mime  the MIME-type the sub-menu is for
 */
static GtkWidget *menu_submenu_new(GtkWidget *menu,char *mime) {
  GtkWidget *sub=gtk_menu_new(),*item;
  if(sub) {
    char *smt=menu_submenu_title_from_mime_type(mime);
    if(smt) {
      if((item=gtk_menu_item_new_with_label(smt))) {
        gtk_menu_shell_append((GtkMenuShell *)menu,item);
        gtk_widget_show(item);
        gtk_menu_item_set_submenu(GTK_MENU_ITEM(item),sub); }
      free(smt); }}
  return sub; }



/**
 * create context-menu for current selection.
 * @return a GTK menu-item
 * @param ac  the action-context
 */
void *menu_selection_create(action_context *ac) {
  noir_sess        *ns=noir_sess_get_by_serial(ac->serial); /* ### */
  GtkWidget        *menu=gtk_menu_new(),*item,*sub=NULL;
  char             *w,*m,*s=strdup(noir_sess_get_selection(ns));
  selection_action *e=NULL,*f;
  int               with_submenus=TRUE;

  if((item=menu_selection_add(ac,menu,"Selection",(gpointer)NULL)))
    menu_selection_insensitive(item);

  if(s) {
    /* determine available actions from files in selection. */
    char *p=s,*b=s;
    while((p=strchr(p,'\n'))) {
      *p='\0';
      menu_selection_file_add(ac,b,&e);
      *(p++)='\0';
      b=p; }
    menu_selection_file_add(ac,b,&e);

    /* add actions to menu. */

    f=e;
    if(with_submenus)
      sub=menu_submenu_new(menu,f->menu->mime);
    while(f) {
      E(2,"%s; %s; %s\n",f->menu->title,f->menu->mime,f->menu->file);
      e=f->next;
      if(!menu_selection_add(ac,sub,f->menu->title,(gpointer)f->menu))
        E(0,"failed to add \"%s\" for \"%s\"\n",f->menu->title,f->menu->mime);
      if(!e||(e->section!=f->section)) {
        if(with_submenus) {
          if(e)
            sub=menu_submenu_new(menu,e->menu->mime); }
        else if((item=gtk_separator_menu_item_new()))
          gtk_menu_shell_append((GtkMenuShell *)menu,item); }

      free(f);
      f=e; }

    free(s); }

  /* add entry "SelectedFilesName only" */
  w=strrchr(ac->single->tag,'/');
  w=w?w+1:ac->single->tag;
  m=NULL;
  esprintf(&m,"\"%s\" only",w);
  if(m) {
    item=menu_selection_add(ac,menu,m,(gpointer)NULL);
    free(m); }
  else
    item=menu_selection_add(ac,menu,"Current file only",(gpointer)NULL);
  menu_selection_insensitive(item);

  return menu; }



/*****************************************************************************/



/**
 * wrapper for menu_selection_apply_one so it can be called by GTK.
 *
 * this is called when a menu-entry pertaining to a single file (rather than
 * a selection of files) is activated; see menu_add.  in this case, NULL will
 * be passed to menu_selected_do in lieu of an action_context.
 * @param item  the menu-item that was selected. its "data" property will contain
 *              the anchor of the file's meta-data.
 * @param data  the menu-action.
 */
static gint menu_single_apply(GtkWidget *item,gpointer data) {
  menu_action    *m=(menu_action *)g_object_get_data(G_OBJECT(item),"data");
  action_context *ac=(action_context *)data;
  action_file    *af=ac->single;
  return menu_selection_apply_one(af->tag,item,m,ac); }



/**
 * add action to context-menu.
 * @param menu  the GTK menu
 * @param m     the menu-action
 * @param ac    the action-context
 */
static void menu_add(GtkWidget *menu,menu_action *m,action_context *ac) {
  GtkWidget *item=gtk_menu_item_new_with_label(m->title);
  if(!item)
    return;
  g_signal_connect(item,"activate",G_CALLBACK(menu_single_apply),ac);
  g_object_set_data(G_OBJECT(item),"data",m);
  gtk_menu_shell_append((GtkMenuShell *)menu,item); }



/**
 * count items in context-menu.
 * @param widget  an item in the menu (ignored)
 * @param data    a pointer to the counter-variable
 */
static void menu_count_items(GtkWidget *widget,gpointer data) {
  int *nc=(int *)data;
  *nc=(*nc+1); }



/**
 * create and display context-menu (according to item this menu is for).
 * @return <0 (error-code) or >=0 (# of items in menu; 0==none,
 * meaning that no menu is shown)
 * @param ac    the action-context
 * @param menu  menu-object (may hand-in previously made menu with extra items;
 *              created inside if NULL)
 */
static int menu_create(action_context *ac,GtkWidget *menu) {
  menu_action *m;
  char        *p,*mime=NULL;
  int          ret=0,internal=0,internal_wd=0;
  meta        *mm;
  char        *mi=NULL;
  GtkWidget   *sub=NULL;

  if(!ac)
    return NOIR_FAIL;

  if(!menu) {
    if(!(menu=gtk_menu_new()))
      return NOIR_OOM; }

  if(ac->single) {
    if((mm=meta_in_group(&ac->single->md,"MIME",MY_NAME)))
      mi=mm->val;

    if(mi) {
      if(!(mime=strdup(mi))) {
        g_object_unref(menu);
        return NOIR_OOM; }

      if(!strcmp(mime,"unknown"))
        *mime='\0';

      /* handlers for this MIME-type: text/plain */
    mc_add_for_mime:
      if(strlen(mime)) {
        m=menu_action_get_anchor();
        if(menu_action_by_mime_first(&m,mime))
          sub=menu_submenu_new(menu,mime);
        while(m) {
          ret++;
          menu_add(sub,m,ac);
          menu_action_by_mime_next(&m,mime); }

        /* handlers for this MIME-group: text/ */
        if((p=strchr(mime,'/'))) {
          internal_wd=!strcmp(mime,MIME_INTL_MB3);
          *(++p)='\0';
          if(strcmp(mime,MIME_INTL)) {
            m=menu_action_get_anchor();
            if(menu_action_by_mime_first(&m,mime))
              sub=menu_submenu_new(menu,mime);
            while(m) {
              ret++;
              menu_add(sub,m,ac);
              menu_action_by_mime_next(&m,mime); }}
          else {
            internal=1;
            if(internal_wd) {
              char *m;
              if((m=strdup(MIME_FOLDER))) {
                free(mime);
                mime=m;
                goto mc_add_for_mime; }}}}
        *mime='\0'; }

      /* wild-card handlers */
      if(!internal) {
        m=menu_action_get_anchor();
        if(menu_action_by_mime_first(&m,mime))
          sub=menu_submenu_new(menu,mime);
        while(m) {
          ret++;
          menu_add(sub,m,ac);
          menu_action_by_mime_next(&m,mime); }}}}

  if(ret>=0) {
    /* count items in menu. can't use ret for that as the caller might have
       handed in a non-empty menu:  count actual items, theirs and ours. */
    gtk_container_foreach(GTK_CONTAINER(menu),menu_count_items,(gpointer)&ret);
    if(ret>0) {
      gtk_widget_show_all(menu);
      gtk_signal_connect(GTK_OBJECT(menu),"selection-done",G_CALLBACK(menu_done),(gpointer)ac);
      gtk_menu_popup(GTK_MENU(menu),NULL,NULL,NULL,NULL,0,0L); }}

  if(mime)
    free(mime);

  return ret; }



/**
 * handle a menu.
 *
 * (called from menu_file_context() for context-menus and from root_menu() for
 * "root"-menus.)
 * @param ac    the action-context
 * @param menu  a pre-made GTK-menu to append entries to, or NULL to create a new menu
 */
int menu_run(action_context *ac,void *menu) {
  int ret;

  info_tooltip_cancel(NULL,NULL,NULL);

  ret=menu_create(ac,(GtkWidget *)menu);

  if(ret<=0) {
    noir_sess *ns=noir_sess_get_by_serial(ac->serial);
    form_alert(ns,"No entries in action menu!?\nDid you make install?");
    ret=NOIR_NOTFOUND; }
  else {
    while(!(ac->flags&AF_DONE))
      mainloop_once();

    action_context_dst(ac);
    ret=NOIR_SUCC; }

  return ret; }






/******************************************************************************/
/* "root-menus" */



/**
 * switch sort-order for icon-view.
 * @param item  the GTK menu-item that was activated.
 *              "data" prop holds sort-flags to set.
 * @param data  the session.
 */
static gint root_menu_sort(GtkWidget *item,gpointer data) {
  noir_sess **sess=(noir_sess **)data;
  gpointer    para=g_object_get_data(G_OBJECT(item),"data");
  char       *cwd=noir_sess_get_cwd(*sess),*cd=NULL;
  int         act=gtk_check_menu_item_get_active((GtkCheckMenuItem *)item);

  if(act)
    (*sess)->sortflags|=(int)para;
  else
    (*sess)->sortflags&=~((int)para);

  E(1,"resort with 0x%x...\n",(*sess)->sortflags);

  /* force resort the ugly way... */
  if(cwd&&(cd=strdup(cwd))) {
    noir_sess_focus((*sess),"/");
    noir_sess_focus((*sess),cd);
    free(cd); }

  return FALSE; }



/**
 * switch views from "root-menu".
 * @param item  the GTK menu-item that was activated.
 *              "data" prop holds index of view to activate.
 * @param data  the session.
 */
static gint root_menu_grid(GtkWidget *item,gpointer data) {
#if 0
  noir_sess **sess=(noir_sess **)data;
  gpointer    para=g_object_get_data(G_OBJECT(item),"data");
  noir_theme *theme=noir_sess_get_theme(*sess);

  /* force resort the ugly way... */

  if(theme_el_set(&theme->data,"iconlistgrid",)==NOIR_SUCC) {
    if(cwd&&(cd=strdup(cwd))) {
      noir_sess_focus((*sess),"/");
      noir_sess_focus((*sess),cd);
      free(cd); }}
#endif
  return FALSE; }



/**
 * switch views from "root-menu".
 * @param item  the GTK menu-item that was activated.
 *              "data" prop holds index of view to activate.
 * @param data  the session.
 */
static gint root_menu_switch2(GtkWidget *item,gpointer data) {
  noir_sess **sess=(noir_sess **)data;
  gpointer    para=g_object_get_data(G_OBJECT(item),"data");
  notebook_go2view(*sess,(int)para);
  return FALSE; }



/**
 * add an item to a "root-menu".
 * @param sess     the session
 * @param menu     the GTK menu
 * @param title    the label for the new entry
 * @param data     the value for the "data" prop (normally index of view to activate)
 * @param handler  the function to call when this item is activated
 */
static GtkMenuItem *root_menu_add(noir_sess **sess,GtkWidget *menu,char *title,gpointer data,gint (*handler)(GtkWidget *,gpointer)) {
  GtkWidget *item;

  if(!title)
    item=gtk_separator_menu_item_new();
  else {
    item=gtk_menu_item_new_with_mnemonic(title);
    if(handler)
      g_signal_connect(item,"activate",G_CALLBACK(handler),sess); }

  g_object_set_data(G_OBJECT(item),"data",data);
  gtk_menu_shell_append((GtkMenuShell *)menu,item);
  return (GtkMenuItem *)item; }



/**
 * add an check-item to a "root-menu".
 * @param sess     the session
 * @param menu     the GTK menu
 * @param title    the label for the new entry
 * @param data     the value for the "data" prop (normally index of view to activate)
 * @param handler  the function to call when this item is activated
 */
static GtkMenuItem *root_sort_add(noir_sess **sess,GtkWidget *menu,char *title,gpointer data,gint (*handler)(GtkWidget *,gpointer)) {
  GtkWidget *item;

  if(!title)
    item=gtk_separator_menu_item_new();
  else {
    item=gtk_check_menu_item_new_with_mnemonic(title);
    gtk_check_menu_item_set_active((GtkCheckMenuItem *)item,(((*sess)->sortflags)&((int)data))); }

  g_object_set_data(G_OBJECT(item),"data",data);
  gtk_menu_shell_append((GtkMenuShell *)menu,item);
  if(handler)
    g_signal_connect(item,"toggled",G_CALLBACK(handler),sess);
  return (GtkMenuItem *)item; }



/**
 * create and show "root-menu".
 * @param ns   session
 * @param cwd  current work-dir (may differ from session's if view can show more than one directory at a time)
 * @param mb   the mouse-button that was pressed to open the menu
 */
int menu_root_context(noir_sess *ns,char *cwd,int mb) {
  action_context *ac;
  action_file    *af=NULL;

  E(2,"menu in \"%s\"...\n",cwd?cwd:"");

  if(!(ac=action_context_new(noir_sess_get_serial(ns),cwd)))
    return NOIR_OOM;
  if(cwd&&!(af=action_file_new(NULL,cwd))) {
    action_context_dst(ac);
    return NOIR_OOM; }
  ac->single=af;

  if(mb==3) {  /* built-in root menu */
    noir_sess   *sess=ns;
    GtkWidget   *menu=gtk_menu_new();
    GtkWidget   *sort=NULL;
    GtkMenuItem *si=NULL;

    if(go2view_mapping[START_IN_ICONVIEW]>=0)
      root_menu_add(&sess,menu,"_Icon-view",(gpointer)START_IN_ICONVIEW,root_menu_switch2);
    if(go2view_mapping[START_IN_BROWSER]>=0)
      root_menu_add(&sess,menu,"_Browser",  (gpointer)START_IN_BROWSER, root_menu_switch2);
    if(go2view_mapping[START_IN_TREEVIEW]>=0)
      root_menu_add(&sess,menu,"_Tree-view",(gpointer)START_IN_TREEVIEW,root_menu_switch2);
    if(go2view_mapping[START_IN_CREDITS]>=0)
      root_menu_add(&sess,menu,"_Credits",(gpointer)START_IN_CREDITS,root_menu_switch2);

#if 0
    root_menu_add(&sess,menu,NULL,NULL,NULL);
    root_sort_add(&sess,menu,"Align using _grid",(gpointer)ns,root_menu_grid);
#endif

    if(cwd) {
      root_menu_add(&sess,menu,NULL,NULL,NULL);

      if((si=root_menu_add(&sess,menu,"Sort",(gpointer)NULL,NULL))) {
        sort=gtk_menu_new(); }
      else
        sort=menu;

      root_sort_add(&sess,sort,"Show _unsorted",(gpointer)SORT_UNSORTED,root_menu_sort);
      root_sort_add(&sess,sort,"_abc = ABC",(gpointer)SORT_CASEFOLD,root_menu_sort);
      root_sort_add(&sess,sort,"_Folders first",(gpointer)SORT_DIRS1ST,root_menu_sort);
      root_sort_add(&sess,sort,"_The Smiths = Smiths",(gpointer)SORT_BANDNAMES,root_menu_sort);
      root_sort_add(&sess,sort,"by _meta-data",(gpointer)SORT_BYMETA,root_menu_sort);
      root_sort_add(&sess,sort,"by _user",(gpointer)SORT_BYUSER,root_menu_sort);
      root_sort_add(&sess,sort,"by _size",(gpointer)SORT_BYSIZE,root_menu_sort);
      root_sort_add(&sess,sort,"by _date",(gpointer)SORT_BYDATE,root_menu_sort);
      root_sort_add(&sess,sort,"by _MIME",(gpointer)SORT_BYMIME,root_menu_sort);

      if(menu!=sort)
        gtk_menu_item_set_submenu(si,sort); }

    if(af) {
      root_menu_add(&sess,menu,NULL,NULL,NULL);

      meta_add(&af->md,META_RO,META_STR,"Tag","File's name",MY_NAME,cwd?cwd:"/",NULL);
      meta_add(&af->md,META_RO,META_STR,"MIME","File's MIME-type",MY_NAME,MIME_INTL_MB3,NULL); }

    menu_run(ac,menu);

    return NOIR_SUCC; }

  else {       /* user-def root menu */
    meta_add(&af->md,META_RO,META_STR,"Tag","File's name",MY_NAME,cwd,NULL);
    meta_add(&af->md,META_RO,META_STR,"MIME","File's MIME-type",MY_NAME,MIME_INTL_MB1,NULL);
    return menu_run(ac,NULL); }}






/******************************************************************************/
