/** @file  tooltip.c
 *  @brief GTK2 file-info tooltips
 */

#include <gtk/gtk.h>
#include <string.h>
#include <stdlib.h>       /* malloc(), free() */

#include "gui/common/session.h"
#include "gui/common/session_private.h"

#include "file.h"         /* file_get_meta_only() */
#include "icon.h"         /* pixbuf_get_by_path_and_resize() */
#include "util.h"         /* timeout_*(), estring */
#include "model.h"        /* file_info_fetch() */

#ifdef WITH_ICONVIEW
#  include "gui/gtk2/icon_view/icon_view.h"    /* struct noir_iconview */
#  include "gui/gtk2/icon_view/gevas_light.h"  /* gevas_queue_redraw() */
#  include "gui/gtk2/icon_view/gevas_bind.h"   /* ICON_LIST_GET_FROM_WIDGET() */
#endif
#ifdef HAVE_EVAS2
#  include "gui/efl/common/theme.h"
#endif


static void *tooltip_timer=NULL;



/******************************************************************************/
/* (GTK2-) tooltips */



/**
 * mouse was moved in time, forget about ever showing the tooltip
 * @param widget the widget the mouse was moved over, or NULL
 * @param ev     the event if widget was left, or NULL otherwise
 * @param data   the session, or NULL
 */
gboolean info_tooltip_cancel(GtkWidget *widget,GdkEventCrossing *ev,gpointer data) {
  if(tooltip_timer)
    timeout_remove(&tooltip_timer);
  return FALSE; }



/**
 * callback for expose events on tooltip
 * @param ns the session
 */
static gint info_tooltip_paint_window(noir_sess *ns) {
  gtk_paint_flat_box(ns->tooltip_window->style,ns->tooltip_window->window,
                     GTK_STATE_NORMAL,GTK_SHADOW_OUT,
                     NULL,GTK_WIDGET(ns->tooltip_window),"tooltip bg",0,0,-1,-1);
  if(!gdk_pointer_grab(ns->tooltip_window->window,FALSE,
                       GDK_BUTTON_PRESS_MASK|GDK_BUTTON_RELEASE_MASK|GDK_LEAVE_NOTIFY_MASK,
                       NULL,NULL,GDK_CURRENT_TIME))
    gtk_grab_add(ns->tooltip_window);
  /* return FALSE so text gets drawn too */
  return FALSE; }



/**
 * callback for leave events of session's window
 * @param widget the widget that was left
 * @param ev     the event
 * @param data   the tooltip-widget
 */
static gboolean info_tooltip_hide_event(GtkWidget *widget,GdkEventCrossing *ev,gpointer data) {
  widget=(GtkWidget *)data;
  if(widget&&GTK_WIDGET_VISIBLE(widget)) {
    gtk_grab_remove(widget); /* this shouldn't be necessary */
    gdk_pointer_ungrab(GDK_CURRENT_TIME);
    gtk_widget_hide(widget); }
  return TRUE; }



/**
 * initialize GTK2 tooltip subsystem for this session
 * @param ns the session
 */
static void info_tooltip_init(noir_sess *ns) {
  if(!ns->tooltip_window) {
      ns->tooltip_window=gtk_window_new(GTK_WINDOW_POPUP);

      gtk_widget_set_app_paintable(ns->tooltip_window,TRUE);
      gtk_window_set_policy(GTK_WINDOW(ns->tooltip_window),FALSE,FALSE,TRUE);
      gtk_widget_set_name(ns->tooltip_window,"tooltip");
      gtk_container_set_border_width(GTK_CONTAINER(ns->tooltip_window),4);

      ns->tooltip_vbox=gtk_vbox_new(FALSE,0);
      ns->tooltip_label=gtk_label_new(NULL);
      gtk_label_set_line_wrap(GTK_LABEL(ns->tooltip_label),TRUE);
      gtk_misc_set_alignment(GTK_MISC(ns->tooltip_label),0.5,0.5);
      
      gtk_container_add(GTK_CONTAINER(ns->tooltip_window),ns->tooltip_vbox);
      gtk_box_pack_start(GTK_BOX(ns->tooltip_vbox),ns->tooltip_label,TRUE,TRUE,0);

      gtk_signal_connect_object(GTK_OBJECT(ns->tooltip_window),"expose_event",
                                GTK_SIGNAL_FUNC(info_tooltip_paint_window),ns);

      gtk_signal_connect_object(GTK_OBJECT(ns->window),"leave-notify-event",
                                G_CALLBACK(info_tooltip_cancel),(gpointer)ns);

      gtk_signal_connect_object(GTK_OBJECT(ns->tooltip_window),"leave-notify-event",
                                G_CALLBACK(info_tooltip_hide_event),ns->tooltip_window);

      gtk_signal_connect_object(GTK_OBJECT(ns->tooltip_window),"button_press_event",
                                G_CALLBACK(info_tooltip_hide_event),ns->tooltip_window);

      gtk_signal_connect_object(GTK_OBJECT(ns->tooltip_window),"key-press-event",
                                G_CALLBACK(info_tooltip_hide_event),ns->tooltip_window);

      gtk_signal_connect(GTK_OBJECT(ns->tooltip_window),"destroy",
                         GTK_SIGNAL_FUNC(gtk_widget_destroyed),
                         &ns->tooltip_window); }}




/**
 * show tooltip *now* (evas or GTK2)
 * @param ns   the session
 * @param file the file to show the tooltip for
 */
void info_tooltip_show_for(noir_sess *ns,char *file) {
  GtkRequisition  requisition;
  gint            x=ns->tooltip_x,y=ns->tooltip_y,w,h,scr_w,scr_h,ret;
  char           *info=NULL;
#ifdef HAVE_LOSTMIND
  GdkBitmap      *mask;
  unsigned char  *d,*c;
  size_t          l,s;
#endif
#ifdef WITH_ICONVIEW
  noir_iconview  *iv;
#endif
  file_info      *fi;

  /* check whether file still points anywhere, session may have disappeared */
  if(!file) {
    E(0,"no file-name given!?\n");
    return; }
  else if(!(fi=file_info_fetch(NULL,file))) {
    E(0,"file_info_fetch(\"%s\") failed!?\n",file);
    return; }
  else if(!fi->meta&&((ret=file_get_meta_only(ns,fi->tag,&fi->meta))!=NOIR_SUCC)) {
    E(0,"file_get_meta_only(%s) failed: %d\n",file,ret);
    return; }

#if defined(WITH_ICONVIEW)&&defined(HAVE_EVAS2)
  iv=(noir_iconview *)noir_sess_client_get_data(ns,CLIENT_ICONVIEW);

  if(iv&&iv->icon_scroll&&GTK_WIDGET_MAPPED(iv->icon_scroll)) {
    int        doshow=0;
    gint       mx=0,my=0;
    icon_list *il=ICON_LIST_GET_FROM_WIDGET(ns->tooltip_origin);

    if((ns->with_tooltips&TOOLTIP_MOUSE)) {
      GdkWindow *win=gdk_window_at_pointer(&mx,&my);
      if(win==ns->tooltip_origin->window) {
        icon *i=icon_list_get_icon_by_file(il,file);
        if(i)
          doshow=1; }}
    else
      doshow=1;

    if(doshow) {
      gdk_window_get_origin(ns->tooltip_origin->window,&mx,&my);
      ns->tt_evas=((noir_theme *)ns->theme)->tt_new(il,x-mx,y-my,fi);
      gevas_queue_redraw((GtkgEvas *)iv->icon_view); }
    ns->with_tooltips=ns->with_tooltips&(~TOOLTIP_MOUSE); }
  else
#endif
  { gsize              foo;
    gchar             *utf8;

    if(!ns->tooltip_window)
      info_tooltip_init(ns);

    if(!(info=meta_to_string(&fi->meta)))
      goto done;

    if((utf8=g_locale_to_utf8(info,-1,&foo,&foo,NULL))) {
      free(info);
      info=utf8; }

    gtk_label_set_text(GTK_LABEL(ns->tooltip_label),info);

    if(ns->tooltip_icon)
      gtk_container_remove(GTK_CONTAINER(ns->tooltip_vbox),ns->tooltip_icon);

    if((ns->tooltip_icon=gtk_image_new_from_pixbuf(pixbuf_get_by_path_and_resize(file,ICON_SIZE_IN_TOOLTIP,GDK_INTERP_HYPER))))
      gtk_box_pack_start(GTK_BOX(ns->tooltip_vbox),ns->tooltip_icon,TRUE,TRUE,0);

    gtk_widget_show_all(ns->tooltip_vbox);
    gtk_widget_size_request(ns->tooltip_window,&requisition);
    w=requisition.width;
    h=requisition.height;

    scr_w=gdk_screen_width()-w; scr_h=gdk_screen_height()-h;
    x-=w/2;                     y-=h/2;

    if(x<0) x=0;                if(y<0) y=0;
    if(x>scr_w) x=scr_w;        if(y>scr_h) y=scr_h;

    gtk_window_move(GTK_WINDOW(ns->tooltip_window),x,y);
    gtk_widget_show_all(ns->tooltip_window); }

 done:
  if(info)
    free(info); }



/**
 * mouse was not moved in time, show tooltip
 * @param data the session
 */
static gint info_tooltip_timeout(gpointer data) {
  noir_sess *ns=(noir_sess *)data;

  GDK_THREADS_ENTER();
  info_tooltip_show_for(ns,ns->tooltip);
  GDK_THREADS_LEAVE();

  return FALSE; }



/**
 * Q tip: if mouse doesn't leave this item soon, show tooltip for it.
 * @param ns   the session
 * @param file the file for which to show the info
 * @param x    the coordinate where the mouse must more or less remain
 * @param y    the coordinate where the mouse must more or less remain
 */
void info_tooltip(noir_sess *ns,char *file,gint x,gint y) {
  if(ns->tooltip)
    free(ns->tooltip);
  ns->tooltip=strdup(file);

  ns->tooltip_x=x;
  ns->tooltip_y=y;

  info_tooltip_cancel(NULL,NULL,NULL);

  tooltip_timer=timeout_add((ns->tooltip_window&&
                            GTK_WIDGET_VISIBLE(ns->tooltip_window))?
                            0:
                            TOOLTIP_DELAY,
                            info_tooltip_timeout,ns); }



/******************************************************************************/
