/** @file  icon_view.c
 *  @brief implements icon-view (shelf in shelf.c)
 *  independent of the actual icon-list (evas/icon_list or gnome-icon-list)
 *  used.
 */

#include "config.h"

#include "gui/common/session_private.h"
#include "gui/common/icon_list.h"
#include "gui/common/menu.h"  /* menu_file_context(), menu_root_context() */

#include "gui/gtk2/common/dnd.h"
#include "gui/gtk2/common/tooltip.h"     /* info_tooltip_cancel() */

#include "gui/gtk2/common/tree/tree.h"

#include "gui/gtk2/tree_view/tree_view.h"

#include "mime.h"
#include "file.h"
#include "file_lcl.h"
#include "util.h"        /* now() */
#include "setup.h"
#include "icon_path.h"

#include <gtk/gtk.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>

#include "icon_view.h"
#if defined(HAVE_EVAS2)
#  include "gui/efl/common/evas_macros.h"
#  include "gui/efl/common/theme.h"
#  include "gui/efl/icon_view/evas_bind.h"
#  include "gevas_light.h"
#  include "gevas_bind.h"
#elif defined(HAVE_GNOMECANVAS)
#  include "gui/gtk2/common/icon.h" /* pixbuf_get_by_path_no_resize() */
#  include "gnome-icon-list.h"
#endif



#define MODULE "iconview"

#define IV_UPDATE_EVERY 2000

#ifdef HAVE_EVAS2
typedef enum { IV_BG, IV_PADX, IV_PADY, IV_PADW, IV_PADH } iv_tag_type;
#endif



/****************************************************************************/



/**
 * set session's selection in icon-view
 * @return an error-code
 * @param data  the session
 */
int icon_view_set_selection(void *data) {
#ifdef HAVE_EVAS2
  noir_sess *ns=(noir_sess *)data;
  return icon_list_set_selection((void *)ns);
#elif defined(HAVE_GNOMECANVAS)
  return NOIR_IGNORED;
#else
  return NOIR_IGNORED;
#endif
}




/****************************************************************************/



/**
 * callback for "exposed"-event
 * the other views don't update our thumbnails for reasons of speed,
 * so if the directory-focus was changed in another view, we'll have
 * to get the thumbnails NOW.
 * @return FALSE
 * @param widget  the gEvas
 * @param ev      the GDK expose-event
 * @param data    the session
 */
gboolean icon_view_exposed(GtkWidget *widget,GdkEventExpose *ev,gpointer data) {
#if defined(HAVE_EVAS2)||defined(HAVE_GNOMECANVAS)
  noir_sess *ns=(noir_sess *)data;
  if(ns) {
#  ifdef HAVE_EVAS2
    { icon_list *il=ICON_LIST_GET_FROM_WIDGET(widget);
      if(il&&il->canvas) {
        gevas_bg_update(il,il->vis_w,il->vis_h);
        evas_obscured_rectangle_add(gevas_get_evas((GtkgEvas *)il->canvas),0,0,il->vis_w,il->vis_h);
        evas_obscured_clear((gevas_get_evas((GtkgEvas *)il->canvas)));
        if(GTK_WIDGET_MAPPED(widget)) {
          if(!il->bg) {
            noir_theme *theme=il->theme;

            gtk_widget_hide(widget); gtk_widget_show(widget); gtk_widget_grab_focus(widget);

            theme->bg_new(theme,il,theme->iconview_bg);

            E(1,"%dx%d\n",il->vis_w,il->vis_h);
            E(1,"delayed background creation %s\n",il->bg?"succeeded.":"failed!?"); }
          evas_render(gevas_get_evas((GtkgEvas *)il->canvas)); }}}
#  endif
  }
#endif
  return FALSE; }



/****************************************************************************/



#if defined(HAVE_EVAS2)||defined(HAVE_GNOMECANVAS)

/**
 * get path at given position
 * @param ns  the session
 * @param x   mouse-x
 * @param y   mouse-y
 */
static char *icon_path_at(noir_sess *ns,int x,int y) {
  noir_iconview *iv;
  int            no;
#  ifdef HAVE_EVAS2
  icon_list *il;
#  endif

  if(!ns||!(iv=(noir_iconview *)noir_sess_client_get_data(ns,CLIENT_ICONVIEW)))
    return NULL;

#  ifdef HAVE_GNOMECANVAS
  no=gnome_icon_list_get_icon_at(GNOME_ICON_LIST(iv->icon_view),x,y);
#  else
  il=ICON_LIST_GET_FROM_WIDGET(iv->icon_view);
  no=il?icon_list_get_index_at(il,x,y):-1;
#  endif

  if(no<0)                                          /* dropped on background */
    return ns->cwd;
  else {                                                  /* dropped on icon */
#  ifdef HAVE_GNOMECANVAS
    return gnome_icon_list_get_icon_filename(GNOME_ICON_LIST(iv->icon_view),no);
#  elif defined(HAVE_EVAS2)
    icon *i=(il>=0)?icon_list_get_icon_by_index(il,no):NULL;
    return ICON_GET_FILENAME(i);
#  endif
  }}



/**
 * DND: drag completed, negotiate content-type.
 * @param widget     the widget
 * @param dc         the drag-context
 * @param x          where the data was dropped
 * @param y          where the data was dropped
 * @param timestamp  when the event happened
 * @param data       session
 */
gboolean icon_drag_drop(GtkWidget *widget,GdkDragContext *dc,gint x,gint y,guint timestamp,gpointer data) {
  return dnd_drag_drop(widget,dc,timestamp,icon_path_at(data,x,y)); }



/**
 * DND: drop source sent data. act upon them.
 * @param widget     the widget
 * @param dc         the drag-context
 * @param x          where the data was dropped
 * @param y          where the data was dropped
 * @param selection  the received data
 * @param info       type of the data as per GtkTargetList
 * @param timestamp  when the event happened
 * @param data       session
 */
static void icon_drag_data_received(GtkWidget *widget,GdkDragContext *dc,gint x,gint y,GtkSelectionData *selection,guint info,guint timestamp,gpointer data) {
  noir_sess     *ns=(noir_sess *)data;
  gchar         *d,*a;
  noir_iconview *iv;

  if(!ns)
    return;

  iv=(noir_iconview *)noir_sess_client_get_data(ns,CLIENT_ICONVIEW);

  if(!selection||(selection->length<1)) {
    E(0,"no data (%p,%d)\n",selection,selection?selection->length:-1234); }
  else if(selection->format==8) {
#  ifdef HAVE_GNOMECANVAS
    int no=gnome_icon_list_get_icon_at(GNOME_ICON_LIST(iv->icon_view),x,y);
#  else
    icon_list *il=ICON_LIST_GET_FROM_WIDGET(iv->icon_view);
    int        no;
    void      *rp;

    if((no=evas_ext_evt(il,x,y,1,EV_DROPPED,ns,NULL,&rp))!=NOIR_NOTFOUND) {
      file_info *fi=(file_info *)rp;
      E(4,"dropped on extension: %d for \"%s\".\n",no,fi?fi->tag:"<NULL>");
      if((no==NOIR_SUCC)&&fi&&fi->tag)
        file_drop(ns,dc->action,selection->data,fi->tag);
      return; }
    else
      no=il?icon_list_get_index_at(il,x,y):-1;
#  endif

    if(info==DRAGDROP_TARGET_XDS) {
#  ifdef WITH_XDS
      char *r=selection->data;
      switch(*r) {
        case 'S':
          E(0,"XDS: success!\n");
          break;
        case 'F':
          E(0,"XDS: the application can't save in remote location, it's up to us.\n");
          xds_set_prop(dc,"");  /* we can't remote-safe, neither -- for now */
          break;
        case 'E':
          E(0,"XDS: the application refuses to save there.\n");
          break;
        default:
          E(0,"XDS: invalid op-code '%c'!?\n",*r);
          break; }
#  endif
      return; }

    a=d=g_strdup(selection->data);

    if(no<0)                                        /* dropped on background */
      file_drop(ns,dc->action,a,ns->cwd);
    else {                                                /* dropped on icon */
#  ifdef HAVE_GNOMECANVAS
      char *dt=gnome_icon_list_get_icon_filename(GNOME_ICON_LIST(iv->icon_view),no);
#  elif defined(HAVE_EVAS2)
      icon *i=(il>=0)?icon_list_get_icon_by_index(il,no):NULL;
      char *dt=ICON_GET_FILENAME(i);
#  endif

      if(dt&&*dt)
        file_drop(ns,dc->action,a,dt);
      else
        E(0,"breakage in 'use'!\n"); }
    g_free(a); }

  else if(selection->format==16) {                       /* GIMP colour drop */
    char buf[8];
    guint16 *color_vals=(guint16 *)selection->data;
    guchar r,g,b;
    r=color_vals[0]>>8;  g=color_vals[1]>>8;  b=color_vals[2]>>8;
    snprintf(buf,8,"#%02x%02x%02x",r,g,b);
    theme_el_set(&((noir_theme *)ns->theme)->data,MODULE  "backgroundtintcolour",buf);
    E(0,"setting " MODULE "-background-tint-colour to %s...\n",buf); }}



/**
 * DND: drop target asks for data. collect and send.
 * @param widget         the widget that was dragged from, or NULL
 * @param dc             the drag-context
 * @param selectionData  the selection we're to set the data on
 * @param info           type of the data as per GtkTargetList
 * @param timestamp      when the event happened
 * @param data           session
 */
static void icon_drag_data_get(GtkWidget *widget,GdkDragContext *dc,GtkSelectionData *selectionData,guint info,guint timestamp,gpointer data) {
  noir_sess      *ns=(noir_sess *)data;
  noir_iconview  *iv=(noir_iconview *)noir_sess_client_get_data(ns,CLIENT_ICONVIEW);

  if(!ns)
    return;

  if(selectionData) {
    gchar *reply=NULL,*file;

#  ifdef HAVE_GNOMECANVAS
    GList *l;
    for(l=gnome_icon_list_get_selection(GNOME_ICON_LIST(iv->icon_view));l;l=l->next) {
      file=gnome_icon_list_get_icon_filename(GNOME_ICON_LIST(iv->icon_view),GPOINTER_TO_INT(l->data));
#  elif defined(HAVE_EVAS2)
    icon_list *il=ICON_LIST_GET_FROM_WIDGET(iv->icon_view);
    icon      *i=il->icons;

    while(i) {
      if(i->state&IE_SEL) {
        file=ICON_GET_FILENAME(i);
#  endif

      if(info==DRAGDROP_TARGET_FM_SELECTION) {
        puts("fm, forthcoming?"); }
      else if((info==DRAGDROP_TARGET_STRING)||(info==DRAGDROP_TARGET_TEXT))
        reply=dnd_send_string(reply,file);
      else if((info==DRAGDROP_TARGET_NETSCAPE_URL)||(info==DRAGDROP_TARGET_URL))
        reply=dnd_send_URL(reply,file);
#  ifdef HAVE_EVAS2
      }
      i=i->next; }

 /* gdk_drag_context_unref(il->dc); */
    il->dc=NULL;
#  else
    }
#  endif

    if(reply) {
      gtk_selection_data_set(selectionData,selectionData->target,8,reply,strlen(reply));
      file_drop_remember(ns,dc,reply); }
    else
      E(0,"got nothing to say\n"); }
  else
    E(0,"no selection data!?\n"); }



/****************************************************************************/



/**
 * set callbacks for icon-view
 * @param ns  the session
 */
void icon_view_connect(noir_sess *ns) {
  noir_iconview  *iv=NULL;

  if(!ns||!(iv=(noir_iconview *)noir_sess_client_get_data(ns,CLIENT_ICONVIEW))) {
    E(0,"failed! %p, %p\n",ns,iv);
    return; }

  gtk_signal_connect(GTK_OBJECT(iv->icon_view),
                     "drag_data_get",G_CALLBACK(icon_drag_data_get),(gpointer)ns);
  gtk_signal_connect(GTK_OBJECT(iv->icon_view),
                     "drag_data_received",G_CALLBACK(icon_drag_data_received),(gpointer)ns);
  gtk_signal_connect(GTK_OBJECT(iv->icon_view),
                     "drag_drop",G_CALLBACK(icon_drag_drop),(gpointer)ns);
  gtk_signal_connect(GTK_OBJECT(iv->icon_view),
                     "drag_data_delete",G_CALLBACK(file_drop_delete_from_iconlist),(gpointer)ns);
  gtk_signal_connect(GTK_OBJECT(iv->icon_view),
                     "drag-end",G_CALLBACK(file_drop_free),(gpointer)ns);

  gtk_drag_dest_set(GTK_WIDGET(iv->icon_view),GTK_DEST_DEFAULT_ALL,target,targets_xds,GDK_ACTION_LINK|GDK_ACTION_COPY|GDK_ACTION_MOVE);
#  ifdef HAVE_GNOMECANVAS
  gtk_drag_source_set(GTK_WIDGET(iv->icon_view),GDK_BUTTON1_MASK|GDK_SHIFT_MASK|GDK_CONTROL_MASK,target,targets,GDK_ACTION_LINK|GDK_ACTION_COPY|GDK_ACTION_MOVE);
#  elif defined(HAVE_EVAS2)
  { /* icon_list *il=ICON_LIST_GET_FROM_WIDGET(iv->icon_view);
    sb_update(il,TRUE); */
  }
#  endif
 }



/****************************************************************************/



#  ifdef HAVE_GNOMECANVAS

/**
 * icon-view/GNOME: handle clicks on icon-view's icon
 * @param il    the icon-list
 * @param no    the index of the icon that was selected
 * @param ev    the event that led to the selection of this icon
 * @param data  the session
 */
void icon_view_selected(GnomeIconList *il,gint no,GdkEvent *ev,gpointer data) {
  noir_sess *ns=(noir_sess *)data;

  if(!ns)
    return;
  if(!ns->cwd) {
    ns->cwd=strdup("/"); /* failsafe, should never happen */
    E(0,"icon_view: ns->cwd not set!?\n"); }

  if(ev) {
    char *file=strdup(gnome_icon_list_get_icon_filename(il,no));
    int is_dir;

    info_tooltip_cancel(NULL,NULL,NULL);

    if(((GdkEventButton *)ev)->button==2)
      menu_path(ns,((GdkEventButton *)ev)->button,((GdkEventButton *)ev)->time);
    else {
      if(!strcmp(gnome_icon_list_get_icon_text_item(il,no)->text,DIR_PARENT)) {
        char *p;
        size_t l=strlen(ns->cwd)-1;

        strcpy(file,ns->cwd);

        if(file[l]=='/')
          file[l--]='\0';  /* trailing slash */

        file_monitor_remove(ns,file);

        if((p=strrchr(file,'/'))) {  /* slash left (should always be true!) */
          if(IS_HIDDEN(((char *)(p+1)))&&!IS_DOTDIR(((char *)(p+1))))
            p[2]='\0';
          else
            p[(p>file)?0:1]='\0'; }}
      if((is_dir=file_exists(file))>0) {
        if(ev->type==GDK_2BUTTON_PRESS) {
          if(S_ISDIR(is_dir)) {
            E(1,"icon_view: focussing on %s\n",file);
            if(((GdkEventButton *)ev)->state&(GDK_MOD1_MASK|GDK_CONTROL_MASK))
              session_new(SESSION_COPY,ns,file);
            else {
              focus(ns,file);
              file_monitor_add(ns,NOIR_MONITOR_DIR,file); }} /* clear semaphore */
          else
            file_activate(ns,file,NULL,NULL); }
        else if((ev->type==GDK_BUTTON_RELEASE)&&((GdkEventButton*)ev)->button==3)
          menu_file_context(ns,file); }
      else if(((GdkEventButton *)ev)->button!=2) {
        if(ev->type==GDK_BUTTON_RELEASE)
          menu_root_context(ns,((GdkEventButton *)ev)->button); }
      else
        E(0,"evidence::icon_view: file \"%s\" (%d) not found!?\n",file,no);
      free(file); }}}



/****************************************************************************/



/**
 * icon-view/GNOME: callback for mouse-motion (prepare tooltip)
 * @return FALSE
 * @param widget  widget that mouse was moved over
 * @param ev      GDK motion-event
 * @param data    the session
 */
gboolean icon_view_make_tooltip(GtkWidget *widget,GdkEventMotion *ev,gpointer data) {
  noir_sess     *ns=(noir_sess *)data;
  noir_iconview *iv;
  int no;

  if(!ns||!ev||
     !(iv=(noir_iconview *)noir_sess_client_get_data(ns,CLIENT_ICONVIEW)))
    return FALSE;

  if(((no=gnome_icon_list_get_icon_at(GNOME_ICON_LIST(iv->icon_view),ev->x,ev->y))>=0)&&(no!=ns->icon_view_last_tooltip)) {
    char *file=strdup(gnome_icon_list_get_icon_filename(GNOME_ICON_LIST(iv->icon_view),no));
    int is_dir;

    ns->icon_view_last_tooltip=no;

    if(file) {
      if(!strcmp(gnome_icon_list_get_icon_text_item(GNOME_ICON_LIST(iv->icon_view),no)->text,DIR_PARENT)) {
        char *p;
        strcpy(file,ns->cwd);
        if((p=strrchr(file,'/'))) {
          if(!p[1]&&(p>file)) { /* bah, trailing slash */
            *p='\0';
            p=strrchr(file,'/'); }
           if(p) {
             if(p>file)
               *p='\0';
             else
               p[1]='\0'; }}}
      if((is_dir=file_exists(file))>0) {
        gint x,y;
        gdk_window_get_pointer(NULL,&x,&y,NULL);
        info_tooltip(ns,file,x,y); }
      else
        E(0,"evidence::icon_view_make_tooltip: file \"%s\" not found!?\n",file);
      free(file); }}
  else if(no<0)
    info_tooltip_cancel(NULL,NULL,NULL);

  return FALSE; }

#  endif

#endif



/****************************************************************************/



/**
 * rerender icon-view
 * (should be inline'd once debugged)
 * @param iv      the icon-view
 * @param il      the icon-list
 * @param caller  debug-info
 */
static inline void icon_view_rerender(noir_iconview *iv,icon_list *il,const char *caller) {
#ifdef HAVE_EVAS2
  Evas *evas;

  if(il&&(evas=get_evas_from_iconlist(il))) {
    E(1,"%s  (%p, %p)\n",caller,il,evas);
    iv->next_update=now(IV_UPDATE_EVERY);
    icon_list_layout_all(il); }
  else
    E(0,"evas has disappeared, not rendering : )\n");
#endif
}



/**
 * voluntary rerender?
 * @param iv  icon-view
 * @param il  icon-list
 */
static inline void icon_view_rerender_maybe(noir_iconview *iv,icon_list *il) {
  if(iv->next_update<now(0)) {
    E(1,"VOLUNTARY RERENDER!  (every %1.3fs)  @%lu\n",((float)IV_UPDATE_EVERY)/1000,now(0));
    icon_view_rerender(iv,il,__FUNCTION__); }}



/****************************************************************************/



/**
 * handle model-event -- any locks must be handled outside of this
 * @return an error-code 
 * @param fi       the fs object this event is about
 * @param type     event-type
 * @param changed  what parts of 'fi' changed
 * @param iv       the icon-view
 * @param pa       parent of 'fi'
 */
static int icon_view_evt_do(file_info *fi,model_event type,file_flags changed,noir_iconview *iv,file_info *pa) {
  /* temporary outer #ifdef, need major fixes for gnome-canvas */
#ifdef HAVE_EVAS2 /* bogusity, hot-fix */
  icon_list     *il=ICON_LIST_GET_FROM_WIDGET(iv->icon_view);
  icon          *i=icon_list_get_icon_by_file(il,fi->tag);
  static int     warned_nofallback=0;

  switch(type) {
    case MODEL_CWD:
      if(changed==FF_NONE) {   /* pre */
#ifdef HAVE_EVAS2
        if(il) {
          iv->next_update=now(IV_UPDATE_EVERY/4);
          icon_list_rst(il); }

#elif defined(HAVE_GNOMECANVAS)
        if(iv->icon_view)
          gtk_container_remove(GTK_CONTAINER(iv->icon_scroll),iv->icon_view);

        iv->icon_view=gnome_icon_list_new(ICON_SIZE_IN_LIST,iv->icon_adj,0);
        iv->session->icon_view_last_tooltip=-1;

        gtk_signal_connect(GTK_OBJECT(iv->icon_view),"select-icon",G_CALLBACK(icon_view_selected),(gpointer)iv->session);
        gtk_signal_connect(GTK_OBJECT(iv->icon_view),"motion-notify-event",G_CALLBACK(icon_view_make_tooltip),(gpointer)iv->session);

        gnome_icon_list_set_selection_mode(GNOME_ICON_LIST(iv->icon_view),GTK_SELECTION_MULTIPLE);
        gnome_icon_list_set_separators(GNOME_ICON_LIST(iv->icon_view),"- \'\"/ \t\n\r_.");

        gtk_widget_set_name(iv->icon_view,"icon_view");

        gtk_container_add(GTK_CONTAINER(iv->icon_scroll),iv->icon_view);
#endif
      }

      else if(!(fi=file_info_fetch(NULL,fi->tag)))
        E(0,"MODEL_POST -- no item!?\n");
      else {                      /* post */
#ifdef HAVE_EVAS2
        /* update ".." icon */
        if((pa=fi->parent)&&strcmp(fi->tag,DIR_DELIM_STR))
          icon_list_icon_insert_sorted(il,icon_path_get_always(pa->tag,DIR_PARENT),pa->tag,DIR_PARENT,NULL,pa->intrinsic,pa,FALSE);

        /* replay from store */
        if(fi->child) {
          file_info  *fc=fi->child;
          int         cnt=0,evnt;

          E(1,"replaying \"%s\"...\n",fi->tag);
          while(fc) {
            E(4,"replaying \"%s\": $%x  %d\n",fc->tag,fc->have,fc->have&FF_THUMB);
/*          icon_view_evt_do(fc,MODEL_UPD,fc->have&~(FF_STAT|FF_MIME|FF_THUMB),iv,fc->parent); */
            icon_view_evt_do(fc,MODEL_UPD,fc->have,iv,fc->parent);
            if((evnt=gdk_events_pending())||(cnt++>64)) {
              cnt=0;
              icon_view_rerender_maybe(iv,il); }
            fc=fc->next; }}

        icon_view_rerender(iv,il,"MODEL_CWD");


        /* extensions -- toolbars etc. */
        if(il) {
          icon *i=((noir_theme *)il->theme)->areas;
          while(i) {
            ((theme_area *)i->user1)->ext_upd(il,i,(void *)fi);
            i=i->next; }}
#endif
      }
      break;



    case MODEL_MSC:
      if(changed&(FF_NAME_|FF_THUMB_|FF_META_)) {
        E(2,"\n\nMODEL_MSC %s %x\n\n",fi?fi->tag:"",changed);
        if(changed&FF_META_&&il->sortflags&&(*il->sortflags&SORT_BYMETA))
          icon_list_resort(il);

        icon_view_set_selection(iv->session);

        { icon *i;
          for(i=il->icons;i;i=i->next)
            evas_icon_set_state(il,i,i->state); }

        icon_view_rerender(iv,il,"MODEL_MSC"); }
      break;



    case MODEL_UPD:
/*    E(0,"UPD %s, $%x, %s\n",type==MODEL_UPD?"UPD":"???",changed,fi&&fi->tag?fi->tag:"???"); */

      /* icon does not yet exist, create & add */
      if((changed&FF_NAME)&&!i) {
        char *ifile=icon_path_get_always(fi->tag,fi->mime);
        char *name=fi->name;

/*      E(0,"%s: %p, $%x  %d\n",fi->tag,pa,pa?pa->have:0,0); */

        if(!ifile||!*ifile) {
          E(0,"no icon set for \"%s\"/\"%s\"...  your " MIME_ICONS " has no icon set for the \"default\" and \"unknown\" fallbacks.\n",fi->tag,fi->mime); }
        else if(!warned_nofallback&&!file_exists(ifile)) {
          warned_nofallback++;
          E(0,"icon-file \"%s\" not found...\n",ifile); }

#ifdef HAVE_EVAS2
        /* update icon for file */
        i=icon_list_icon_insert_sorted(il,ifile,fi->tag,name,fi->thumbfile,fi->intrinsic,fi,FALSE);

#elif defined(HAVE_GNOMECANVAS)
        { GdkPixbuf *pb=NULL;
          if(iv->session->thumbnails_in_iconlist) {
            if(!(pb=pixbuf_get_by_path_and_resize(fi->tag,ICON_SIZE_IN_LIST,GDK_INTERP_BILINEAR)))
              pb=pixbuf_get_by_path_and_resize(icon_path_get_always(fi->tag,fi->mime),ICON_SIZE_IN_LIST,GDK_INTERP_BILINEAR); }
          else
            pb=pixbuf_get_by_path_no_resize(icon_path_get_always(fi->tag,fi->mime));
          if(pb)
            gnome_icon_list_append_pixbuf(GNOME_ICON_LIST(iv->icon_view),pb,fi->tag,name); }
#endif
        goto done; }


#if 1
      if((changed&(FF_STAT|FF_MIME))&&i) {
#ifdef HAVE_EVAS2
        if((evas_icon_chg_event(il,i,fi)!=NOIR_FAIL)&&
           (!pa||(pa->have&FF_MIME_))) {
/* E(0,"%s, %x (%s, $%x)\n",fi->tag,fi->have,pa?pa->tag:"???",pa?pa->have:0); */
          icon_view_rerender(iv,il,"MODEL_UPD -- STAT/MIME"); }
#elif defined(HAVE_GNOMECANVAS)
#endif
      }
#endif

#if 1
      if(changed&FF_THUMB) {
#ifdef HAVE_EVAS2
        /* are thumbnails enabled?  and is this file supposed to have one? */
        if(iv->session->thumbnails_in_iconlist&&(fi->have&FF_THUMB)) {
          /* find pre-existing icon for this file. */
          if((i=icon_list_get_icon_by_file(il,fi->tag))) {
            /* try to modify icon. */
            if(evas_icon_chg_event(il,i,fi)!=NOIR_FAIL) {
              /* did we actually get a thumbnail? */
              if(i->state&IE_THMB) {
                /* if this is a monitor-event rather than the initial read so
                   we don't expect more of the same soon, update view now. */
                if(!pa||(pa->have&FF_THUMB_))
                  icon_view_rerender(iv,il,"MODEL_UPD -- THUMB"); }

              /* got no thumbnail, so don't pretend to have one. */
              else {
                if(fi->intrinsic&&!S_ISDIR(fi->intrinsic->mode)) {
                  E(0,"failed to get thumbnail for \"%s\"!? ($%x)\n",fi->tag,i->state); }
                fi->have&=(~FF_THUMB); }}}
          /* event was unrelated to this view!? */
          else
            E(0,"FF_THUMB ignored for \"%s\"!?\n",fi->tag); }
#elif defined(HAVE_GNOMECANVAS)
#endif
      }
#endif
      break;



    case MODEL_DEL:
      if((i=icon_list_get_icon_by_file(il,fi->tag))) {
        icon_list_icon_remove(il,i,TRUE);
        icon_view_rerender(iv,il,"MODEL_DEL"); }
      break;

    default:
      break; }

 done:

#ifdef HAVE_EVAS2
  evas_damage_rectangle_add(get_evas_from_iconlist(il),0,0,il->vis_w,il->vis_h);
#endif

#if 1
  if(pa&&!(pa->flags&&FF_NAME_))
    icon_view_rerender_maybe(iv,il);
#endif

#endif  /* bogusity, hot-fix */

  return NOIR_SUCC; }



/**
 * take care of locks, then handle model-event
 * @return an error-code 
 * @param fi       the fs object this event is about
 * @param type     event-type
 * @param changed  what parts of 'fi' changed
 * @param data     the icon-view
 */
static int icon_view_evt(file_info *fi,model_event type,file_flags changed,void *data) {
  noir_iconview *iv=(noir_iconview *)data;
  file_info     *pa=fi->parent;
  client_funs   *cf=noir_sess_client_lock(iv->serial,CLIENT_ICONVIEW);
  int            ret;

  if(!cf) {
    E(4,"ignoring event for newly-destroyed view...\n");
    ret=NOIR_IGNORED;
    goto done; }

  /* temporary fix so we'll only get events for stuff in the CWD.
     at a later date, we'll only get what we register for, which
     will make this kludge unnecessary. */
  if(type!=MODEL_CWD) {
    if(type==MODEL_MSC)
      pa=fi;
    if(pa!=iv->session->cd) {
      ret=NOIR_IGNORED;
      goto done; }}

#  if 0
  { char *mon[]={ "???", "UPD", "DEL", "CWD", "MSC" };
    E(0,"%s/$%4x   \"%s\"\n",mon[type],changed,fi?fi->tag:"<NULL>"); }
#  endif

  ret=icon_view_evt_do(fi,type,changed,data,pa);

 done:
  noir_sess_client_unlock(cf);

  return ret; }






/****************************************************************************/



/**
 * destruct icon-view
 * @return an error-code
 * @param data  the icon-view
 */
static int icon_view_dst(void **data) {
  noir_iconview *iv=*((noir_iconview **)data);

  if(iv) {
    memset(iv,0,sizeof(noir_iconview));
    free(iv);
    return NOIR_SUCC; }
  else
    return NOIR_IGNORED; }



#ifdef HAVE_EVAS2

/**
 * listener for numeric properties
 * @return an error-code
 * @param key   the name of the property
 * @param type  the data-type of the property
 * @param tag   the numeric tag of the property
 * @param data  the icon-list
 * @param te    linked list of theme-elements
 */
static int icon_view_num_listener(const char *key,const noir_theme_el_type type,const int tag,void *data,noir_theme_el **te) {
  long           val=theme_el_get_int(te,key);
  icon_list     *il=(icon_list *)data;

  switch(tag) {
    case IV_PADX:
      il->xpad=val;
      icon_list_layout_all(il);
      break;
    case IV_PADY:
      il->ypad=val;
      icon_list_layout_all(il);
      break;
    case IV_PADW:
      il->wpad=val;
      icon_list_layout_all(il);
      break;
    case IV_PADH:
      il->hpad=val;
      icon_list_layout_all(il);
      break;
    default:
      E(0,"unknown tag %d (\"%s\")!?\n",tag,key?key:"<none>");
      return NOIR_FAIL; }

  E(1,"(%s::=%ld, %p)\n",key,val,data);

  return NOIR_SUCC; }



/**
 * listener for string properties
 * @return an error-code
 * @param key   the name of the property
 * @param type  the data-type of the property (string? : )
 * @param tag   the numeric tag of the property
 * @param data  the icon-list
 * @param te    linked list of theme-elements
 */
static int icon_view_string_listener(const char *key,const noir_theme_el_type type,const int tag,void *data,noir_theme_el **te) {
  char          *val=theme_el_get_string(te,key),*x=NULL;
  noir_sess     *ns=(noir_sess *)data;
  noir_theme    *theme=(noir_theme *)((noir_theme *)ns->theme);
  noir_iconview *iv=(noir_iconview *)noir_sess_client_get_data(ns,CLIENT_ICONVIEW);
  icon_list     *il=ICON_LIST_GET_FROM_WIDGET(iv->icon_view);

  if(val) {
    void *o=il->bg;

    if((theme->bg_new==NULL)||(theme->bg_dst==NULL))
      return NOIR_FAIL;

    /* empty values are permissable, invalid file-paths are not */
    if(val&&*val&&!file_exists(val)) {
      char *y=file_if_local(val);
      if((y==NULL)||(y[0]==DIR_DELIMITER))
        return NOIR_NOTFOUND;
      if(!(x=get_path_file(NULL,y))) {
        if(!(x=get_setup_file(y)))
          return NOIR_NOTFOUND; }
      val=x; }

    if((GTK_WIDGET_MAPPED(iv->icon_view))&&!theme->bg_new(theme,il,val)) {
      E(0,"failed to update background...\n");
      il->bg=o; }
    else {
      if(theme->iconview_bg)
        free(theme->iconview_bg);
      theme->iconview_bg=strdup(val);
      if(theme->bg_dst)
	theme->bg_dst(&o); }

    E(1,"(%s->%s::=\"%s\", %p): $%p/$%p\n",
      ((noir_theme *)theme)->name,key,val?val:"<null>",data,il->bg,o);

    gevas_icon_list_force(il);

    if(x)
      free(x);

    return NOIR_SUCC; }

  return NOIR_IGNORED; }
#endif



/**
 * create icon-view in given session
 * @return an error-code
 * @param ns  the session
 */
int icon_view_init(noir_sess *ns) {
  noir_iconview *iv;
#ifdef HAVE_GNOMECANVAS
  GtkAdjustment *va=(GtkAdjustment*)gtk_adjustment_new(0,0,100,1,10,10);
  GtkWidget     *scroll=gtk_scrolled_window_new(NULL,va);

  gtk_widget_set_name(scroll,"icon_scroll");

  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
                                 GTK_POLICY_NEVER,GTK_POLICY_AUTOMATIC);

  if(!ns)
    return NOIR_FAIL;

  if(!(iv=(noir_iconview *)malloc(sizeof(noir_iconview))))
    return NOIR_OOM;
  memset(iv,0,sizeof(noir_iconview));

  iv->icon_scroll=scroll;
  iv->icon_adj=va;

  gtk_signal_connect(GTK_OBJECT(scroll),"expose-event",G_CALLBACK(icon_view_exposed),(gpointer)ns);

  iv->client=noir_sess_client_register_full(ns,CLIENT_ICONVIEW,(void *)iv,icon_view_evt,icon_view_dst,icon_view_set_selection);

  return NOIR_SUCC;

#elif defined(HAVE_EVAS2)
  GtkWidget  *wtoy;
  icon_list  *il;

  if(!(iv=(noir_iconview *)malloc(sizeof(noir_iconview))))
    return NOIR_OOM;
  memset(iv,0,sizeof(noir_iconview));

  if(!ns||!(wtoy=gevas_icon_list_new(ns,MODULE, 24,16, 0))) {
    free(iv);
    return NOIR_FAIL; }

  il=ICON_LIST_GET_FROM_WIDGET(wtoy);

/*  ((noir_theme *)ns->theme)->bg_new(((noir_theme *)ns->theme),il,((noir_theme *)ns->theme)->iconview_bg); */

  iv->icon_view=il->canvas;
  iv->icon_scroll=wtoy;
  iv->session=ns;
  iv->serial=noir_sess_get_serial(ns);

  gtk_widget_set_name(iv->icon_scroll,"icon_scroll");
  gtk_widget_set_name(iv->icon_view,  "icon_view");

  gtk_signal_connect(GTK_OBJECT(iv->icon_scroll),"expose-event",G_CALLBACK(icon_view_exposed),(gpointer)ns);

  iv->client=noir_sess_client_register_full(ns,CLIENT_ICONVIEW,(void *)iv,icon_view_evt,icon_view_dst,icon_view_set_selection);

  icon_view_connect(ns);

  theme_el_default(&((noir_theme *)ns->theme)->data,MODULE "padx","12",-64,64,1);
  theme_el_default(&((noir_theme *)ns->theme)->data,MODULE "pady", "8",-64,64,1);
  theme_el_default(&((noir_theme *)ns->theme)->data,MODULE "padw","24",-32,32,1);
  theme_el_default(&((noir_theme *)ns->theme)->data,MODULE "padh","16",-32,32,1);

#  define LISTENER "icon_view{evas}"
  theme_el_listen(&((noir_theme *)ns->theme)->data,LISTENER,MODULE,icon_view_string_listener,IV_BG,ns);
  theme_el_listen(&((noir_theme *)ns->theme)->data,LISTENER,MODULE "padx",icon_view_num_listener,IV_PADX,il);
  theme_el_listen(&((noir_theme *)ns->theme)->data,LISTENER,MODULE "pady",icon_view_num_listener,IV_PADY,il);
  theme_el_listen(&((noir_theme *)ns->theme)->data,LISTENER,MODULE "padw",icon_view_num_listener,IV_PADW,il);
  theme_el_listen(&((noir_theme *)ns->theme)->data,LISTENER,MODULE "padh",icon_view_num_listener,IV_PADH,il);

  il->extensions=(void *)((noir_theme *)ns->theme)->areas;

  return NOIR_SUCC;
#else
  return NOIR_FAIL;
#endif
}



/**
 * drop icon-view in given session
 * @return an error-code
 * @param ns  the session
 */
int icon_view_exit(noir_sess *ns) {
  /* icons are destroyed in gevas' call-back (cf. gevas_icon_list_new()) */
  return NOIR_SUCC; }
