/** @file  ipc_main.c
 *  @brief common functions for IPC-plugins.  plugin-loader.
 */

#include <signal.h>
#include <dlfcn.h>
#include <stdio.h>
#include <glob.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>     /* malloc(), free() */

#include "gui/efl/common/theme.h"

#include "debug.h"
#include "errors.h"
#include "util.h"
#include "ipc.h"
#include "model.h"
#include "file.h"
#include "file_lcl.h"

#ifndef TRUE
#  define FALSE 0
#  define TRUE (!FALSE)
#endif



typedef struct _ipc {
  void *lib;
  void *data;
  int (*ipc_init)(void **data);
  int (*ipc_exit)(void **data);
  int (*ipc_poll)(void **data);
  struct _ipc *next;
} ipc;



static ipc   *ipc_modules=NULL;
static void  *ipc_timer=NULL;



/*****************************************************************************/
/* INTERFACE FOR IPC MODULES */
/*****************************/



/**
 * give all props for session identified by 'serial'.
 * @param serial  session identifier
 */
char *ipc_prop_list(const long serial) {
  noir_theme    *theme=noir_sess_get_theme(noir_sess_get_by_serial(serial));
  noir_theme_el *e=theme?theme->data:NULL;
  estring       *s=estring_new(8192);
  int            f=0;

  while(e) {
    estring_appendf(s,"%s%s: %s",f?"\n":"",e->key,theme_el_get_type(e));
    if(e->flags&EF_BOUNDS) {
      if(e->type==ET_INT)
        estring_appendf(s,", range %d..%d, step %d",e->lo,e->hi,e->step);
      else if(e->type==ET_FLT)
        estring_appendf(s,", range %f..%f, step %f",(float)e->lo/NOIR_THEME_FLOAT_PRECISION,(float)e->hi/NOIR_THEME_FLOAT_PRECISION,(float)e->step/NOIR_THEME_FLOAT_PRECISION);
      else
        E(0,"bounds not expect on type %d prop \"%s\"...\n",e->type,e->key); }
    f=1;
    e=e->next; }

  return estring_disown(s); }



/**
 * describe a specific property in a given session.
 * @param serial  session identifier
 * @param key     property-name
 */
char *ipc_prop_desc(const long serial,const char *key) {
#ifdef HAVE_EVAS2
  noir_theme    *theme=noir_sess_get_theme(noir_sess_get_by_serial(serial));
  noir_theme_el *e=theme_el_get(&theme->data,key);

  if(e) {
    estring *s=estring_new(512);
    estring_appendf(s,"%s: %s",e->key,theme_el_get_type(e));
    if(e->flags&EF_BOUNDS) {
      if(e->type==ET_INT)
        estring_appendf(s,", range %d..%d, step %d",e->lo,e->hi,e->step);
      else if(e->type==ET_FLT)
        estring_appendf(s,", range %d..%d, step %d",e->lo/NOIR_THEME_FLOAT_PRECISION,e->hi/NOIR_THEME_FLOAT_PRECISION,e->step/NOIR_THEME_FLOAT_PRECISION);
      else
        E(0,"bounds not expect on type %d prop \"%s\"...\n",e->type,e->key); }
    return estring_disown(s); }
#endif
  return strdup("<undefined>"); }



/**
 * get the value of a specific property in a given session.
 * @param serial  session identifier
 * @param key     property-name
 */
char *ipc_prop_get(const long serial,const char *key) {
#ifdef HAVE_EVAS2
  char       *ret=NULL;
  noir_theme *theme=noir_sess_get_theme(noir_sess_get_by_serial(serial));
  if((ret=theme_el_get_as_string(&theme->data,key)))
    return ret;
#endif
  return strdup("<undefined>"); }



/**
 * set the value of a specific property in a given session.
 * @param serial  session identifier
 * @param key     property-name
 * @param val     the new value for the property
 */
int ipc_prop_set(const long serial,const char *key,const char *val) {
#ifdef HAVE_EVAS2
  int ret;
  noir_theme *theme=noir_sess_get_theme(noir_sess_get_by_serial(serial));
  ret=theme_el_set(&theme->data,key,(char *)val);
  E(1,"set(%s->%s,\"%s\") => %d\n",theme->name,key,val,ret);
  return ret;
#else
  return NOIR_NOTSUPP;
#endif
}






/*****************************************************************************/



/**
 * get meta-data for a given file.
 * @param file  the file to get info on
 */
char *ipc_cmd_info(char *file) {
  noir_sess   *ns=noir_sess_get_by_serial(0);
  meta        *m=NULL;
  int          ret;
  char        *f,*r;

  if(!file||!*file||!ns)
    return strdup("<must_provide_file>");

  if((r=file_if_local(file))) {
    if(*r==DIR_DELIMITER)
      f=strdup(r);
    else
      esprintf(&f,"%s/%s",noir_sess_get_cwd(ns),r); }
  else
    f=strdup(file);

  r=NULL;
  if((ret=file_get_meta_only(ns,f,&m))==NOIR_SUCC) {
    if(!(r=meta_to_string(&m)))
      r=strdup("<internal_error>"); }
  else
    esprintf(&r,"<error_%d>",ret);

  meta_del_all(&m);
  if(f)
    free(f);

  return r; }



/**
 * list a given directory.
 * @param file  the directory to list
 */
char *ipc_cmd_ls(char *file) {
  noir_sess   *ns=noir_sess_get_by_serial(0);
  char        *f,*r=NULL;
  file_info   *fi;

  if(!file||!*file||!ns)
    return strdup("<must_provide_dir>");

  if((file=file_if_local(file))) {
    if(*file==DIR_DELIMITER)
      f=strdup(file);
    else
      esprintf(&f,"%s/%s",noir_sess_get_cwd(ns),file); }
  else
    f=strdup(file);

  if((fi=file_info_fetch(NULL,f))) {
    if(!(fi=fi->child))
      r=strdup("<file_not_a_directory>");
    else {
      estring       *s=estring_new(8192);
      int            f=0;

      while(fi) {
        estring_appendf(s,"%s%s",f?"\n":"",fi->name);
        f=1;
        fi=fi->next; }

      r=estring_disown(s); }}
  else
    r=strdup("<file_not_found>");

  if(f)
    free(f);

  return r; }






/*****************************************************************************/



/**
 * list all windows/sessions.
 */
char *ipc_win_list(void) {
  noir_sess   *ns=noir_sess_get_1st();
  estring     *s=estring_new(8192);
  int          f=0;

  if(!ns)
    return strdup("<no_windows_open>");

  while(ns) {
    estring_appendf(s,"%s%d: %s",f?"\n":"",noir_sess_get_serial(ns),noir_sess_get_cwd(ns));
    f=1;
    ns=noir_sess_get_next(ns); }

  return estring_disown(s); }



/**
 * open a new window showing a specified folder.
 * @param cwd  the folder to show
 */
int ipc_win_new(const char *cwd) {
  noir_sess *ns=noir_sess_get_by_serial(0);
  return noir_sess_create(SESSION_COPY,ns,cwd); }



/**
 * get the current selection of a given window.
 * @param serial  a session-identifier
 */
char *ipc_win_sel_get(const long serial) {
  noir_sess *ns=noir_sess_get_by_serial(serial);
  if(!ns)
    return "<no such window>";
  else {
    char      *sel=noir_sess_get_selection(ns);
    return strdup(sel?sel:""); }}



/**
 * set the current selection of a given window.
 * @param serial  session identifier
 * @param sel     the selection to set
 */
int ipc_win_sel_set(const long serial,const char *sel) {
  noir_sess *ns=noir_sess_get_by_serial(serial);
  return noir_sess_set_selection(ns,(char *)sel); }



/**
 * get the current working directory of a given window.
 * @param serial  session-identifier
 */
char *ipc_win_cwd_get(const long serial) {
  noir_sess *ns=noir_sess_get_by_serial(serial);
  char      *cwd=noir_sess_get_cwd(ns);
  return strdup(cwd?cwd:"<no such window>"); }



/**
 * set the current working directory of a given window.
 * @param serial  session identifier
 * @param cwd     the folder to show
 */
int ipc_win_cwd_set(const long serial,const char *cwd) {
  noir_sess *ns=noir_sess_get_by_serial(serial);
  return noir_sess_focus(ns,(char *)cwd); }



/**
 * find the (first) session showing a given working directory.
 * @param cwd  the folder to show
 */
long ipc_win_cwd_find(const char *cwd) {
  noir_sess *ns=noir_sess_get_by_cwd(cwd);
  return ns?noir_sess_get_serial(ns):-1; }






/*****************************************************************************/
/* INTERFACE FOR EVIDENCE */
/**************************/



/**
 * check all IPC-handlers for new data.
 * @param data  list of IPC handlers
 */
static int ipc_poll(void *data) {
  ipc *m=(ipc *)data;

  while(m) {
    m->ipc_poll(&m->data);
    m=m->next; }

  return TRUE; }



/**
 * deinitialize IPC sub-system.
 */
int ipc_exit(void) {
  ipc *m;
  if(ipc_timer)
    timeout_remove(&ipc_timer);
  while(ipc_modules) {
    m=ipc_modules;
    ipc_modules=ipc_modules->next;
    m->ipc_exit(&m->data);
    free(m); }
  return NOIR_IGNORED; }



/**
 * initialize IPC sub-system.
 */
int ipc_init(void) {
  char         buf[PATH_MAX];
  glob_t       globbuf;
  int          ret;
  unsigned int c;
  ipc         *nm=NULL;

  if(((ret=snprintf(buf,PATH_MAX,PACKAGE_PREFIX "/lib/evidence/ipc/ipc_*.so"))<0)||
     (ret>=PATH_MAX))
    return NOIR_FAIL;

  glob(buf,0,NULL,&globbuf);
  if(!globbuf.gl_pathc)
    return NOIR_NOTFOUND;

  for(c=0;c<globbuf.gl_pathc;c++) {
    if(!(nm=malloc(sizeof(ipc)))) {
      ret=NOIR_OOM;
      goto done; }
    memset(nm,0,sizeof(ipc));

    E(1,"checking \"%s\"...\n",globbuf.gl_pathv[c]);
    ret=dlmulti("IPC-plugin",globbuf.gl_pathv[c],RTLD_NOW,&nm->lib,
                "!ipc_module_init !ipc_module_exit !ipc_module_poll",
                &nm->ipc_init,&nm->ipc_exit,&nm->ipc_poll);
    if(ret==NOIR_NODATA)
      E(0,"could not load \"%s\": %s...\n",globbuf.gl_pathv[c],dlerror());
    else if(ret==NOIR_SUCC) {
      if((ret=nm->ipc_init(&nm->data))!=NOIR_SUCC)
        E(0,"could not initialize \"%s\": %d\n",globbuf.gl_pathv[c],ret);
      else {
        char *p=globbuf.gl_pathv[c];
        if(!debug) {
          char *q=strrchr(p,DIR_DELIMITER);
          if(q)
            p=++q; }
        E(1,"adding \"%s\"...\n",p);
        nm->next=ipc_modules;
        ipc_modules=nm; }}
    if(ret!=NOIR_SUCC)
      free(nm); }

 done:
  globfree(&globbuf);

  if(ipc_modules) {
    /* ### temporary evilness */
    if((debug>0)||(getenv("USER")&&!strcmp(getenv("USER"),"azundris"))) {
      signal(SIGINT,SIG_DFL);
      signal(SIGSEGV,SIG_DFL); }
    ipc_timer=timeout_add(20,ipc_poll,ipc_modules); }

  return ret; }



/*****************************************************************************/
