/** @file  model.h
 *  @brief internal model of file-system
 */

#ifndef FS_MODEL

#  define FS_MODEL

#if defined(__cplusplus)
extern "C" {
#endif

#  include "meta.h"
#  include "userconfig.h"
#  include <sys/types.h>
#  include <pthread.h>

#ifdef HIDE_FILES_STARTING_WITH_HASH
#  define IS_HIDDEN(name) (name&&((name[0]=='.')||(name[0]=='#')||(name[0]=='%')||(name[strlen(name)-1]=='~')))
#else
#  define IS_HIDDEN(name) (name&&((name[0]=='.')||(name[0]=='%')||(name[strlen(name)-1]=='~')))
#endif

#define IS_DOTDIR(name) (name&&(!strcmp(name,DIR_MYSELF)))
#define IS_IGNOBLE(name) (name&&(!strcmp(name,DIR_PARENT)))

#define DELIMITER          "\n"
#define DELIMITER_         '\n'

#define DIR_DELIMITER      '/'
#define DIR_DELIM_STR      "/"
#define DIR_PARENT         ".."
#define DIR_MYSELF         "."
#define DIR_DOT            ".files"

#ifndef FALSE
#  define FALSE 0
#  define TRUE (!(FALSE))
#endif



/** what state a queue's worker is in */
typedef enum {
  QUEUE_NONE=0,       /**< thread does not exist */
  QUEUE_SCANNING=1,   /**< thread is busy */
  QUEUE_ZOMBIE=2      /**< thread has terminated, must be joined */
} queue_flags;



/** how to sort */
typedef enum {
  SORT_UNSORTED=1,    /**< don't sort */
  SORT_CASEFOLD=2,    /**< sort is case-insensitive */
  SORT_DIRS1ST=4,     /**< don't put directories in front */
  SORT_BANDNAMES=8,   /**< musicshop sort: ignore "the" etc. in band-names */
  SORT_BYMIME=16,     /**< sort as "mp3.mysong" rather than "mysong.mp3" */
  SORT_BYSIZE=32,     /**< by size */
  SORT_BYDATE=64,     /**< by date */
  SORT_BYUSER=128,    /**< by user */
  SORT_BYMETA=256     /**< by meta */
} sort_flags;



/** bit-vector describing what data we already have */
typedef enum {
  FF_NONE  = 0,       /**< got nothing */
  FF_NAME  = 1,       /**< got name */
  FF_STAT  = 2,       /**< got stat()-info */
  FF_MIME  = 4,       /**< got MIME-type */
  FF_META  = 8,       /**< got meta-data */
  FF_THUMB = 0x10,    /**< got thumbnail */

  FF_ANY   = 0x1f,    /**< any of the "got"-flags? */

  FF_FIRST = 0x20,    /**< issued when about to read first item in a folder */
  FF_LAST  = 0x100,   /**< issued after last item in a folder was read */

  FF_NAME_ = 0x100,   /**< got names for all items in folder */
  FF_STAT_ = 0x200,   /**< got stat-info for all items in folder */
  FF_MIME_ = 0x400,   /**< got mime-type for all items in folder */
  FF_META_ = 0x800,   /**< got other meta-data for all items in folder */
  FF_THUMB_= 0x1000,  /**< got thumbnails for all (applicable) items in folder */
  FF_ANY_  = 0x1f00,  /**< any of the "got all"-flags? */

  FF_SEMA  = 0x10000, /**< for directories */
  FF_UNSEEN= 0x20000, /**< this item already clocked in? */

  FF_SORTED= 0x40000  /**< deprecated */
} file_flags;



/** non-POSIX flags for a file-info */
typedef enum {
  IF_NONE    =0,      /**< no flags */
  IF_NOUSE   =1,      /**< no 'x' */
  IF_NOMODIFY=2,      /**< no 'w' */
  IF_VIRTUAL =4,      /**< virtual item */
  IF_DOTDIR  =8       /**< ".folder" */
} item_flags;



/** flags for a client/view */
typedef enum {
  CF_NONE  =0,        /**< no flags */
  CF_PAUSED=1         /**< client/view isn't interested in updates right now */
} client_flags;



/** path_get() mode */
typedef enum {
  PG_ENQUEUE_NOT,     /**< do note enqueue this directory */
  PG_ENQUEUE_ALL,     /**< read all directories along the path */
  PG_ENQUEUE_LAST     /**< read only deepest directory in path */
} pg_mode;



/**< what kind of events the model can send */
typedef enum {
  MODEL_UPD=1,        /**< update an item */
  MODEL_DEL=2,        /**< remove an item */
  MODEL_CWD=3,        /**< change directory */
  MODEL_MSC=4         /**< miscellaneous, see file_flags for sub-type */
} model_event;



/*****************************************************************/



#define CLIENT_PAUSED(cf)     (((cf)->flags)&CF_PAUSED)
// #define CLIENT_LISTENS(cf,fi) ((!(((cf)->flags)&CF_PAUSED))&&(dcd_get((fi),(cf))||(((fi)->parent)&&dcd_get((fi)->parent,(cf)))))
#define CLIENT_LISTENS(cf,fi)  (!CLIENT_PAUSED(cf))



/*****************************************************************/



/** POSIX compatible meta-data */
typedef struct _posixmd {
  mode_t  mode,lmode;                   /**< mode (cf stat()) of file, symlink to file */
  nlink_t hardlinks;                    /**< hardlinks (# of names) */
  uid_t   uid;                          /**< ID of owning user */
  gid_t   gid;                          /**< ID of owning group */
  off_t   size;                         /**< the content's size in bytes */
  time_t  atime;                        /**< when the data was last accessed */
  time_t  mtime;                        /**< when the file's content was last changed */
  time_t  ctime; } posixmd;             /**< when the (POSIX) meta-data were last changed */



typedef struct _dir_client_data dcd;



/** a file-info entry */
typedef struct _file_info {
  char              *name,*tag,         /**< file name, full path/UUID/URL */
                    *mime,              /**< MIME-type */
                    *sortkey;           /**< sort key */
  file_flags         have;              /**< what info is already set on f/i */
  item_flags         flags;             /**< "read-only" etc. */
  posixmd           *intrinsic;         /**< stat()-info */
  struct _meta      *meta;              /**< meta-data */
  char              *thumbfile;         /**< thumbnail (name and path) */
  dcd               *clientdata;        /**< list of client-data items */
  int                children;          /**< children (count) */
  int                refcount;          /**< references to this item */
  struct _file_info *parent,            /**< the parent file_info */
                    *child;             /**< the first child's file_info */
  struct _file_info *next;              /**< next sibling (in same folder) */
} file_info;



/** a client/view object */
typedef struct _client_funs {
  const char          *name;            /**< the view's name (usually type) */
  void                *data;            /**< view's data (struct _myview *) */
  client_flags         flags;           /**< flags: paused etc. */
  int                (*evt_view)(file_info *,model_event type,file_flags changed,void *); /**< event callback */
  int                (*dst_data)(void **);     /**< destructor for private 'data' */
  int                (*set_selection)(void *); /**< method to highlight session's selection in this view */
  void              *(*dcd_set_data)(void *);  /**< DCD setter */
  int                (*dcd_dst_data)(void **); /**< DCD destructor */
  pthread_mutex_t      viewlock;        /**< view mutex */
  void                *user1,           /**< next view in per-session chain */
                      *user2;           /**< session this view belongs to */
  struct _client_funs *next; } client_funs;



/** client/view-specific data to associate with a single file-entry */
struct _dir_client_data {
  client_funs             *client;      /**< the client this data is for */
  void                    *data;        /**< a client-specific pointer */
  struct _dir_client_data *next; };     /**< next DCD for this file-info */



/** entry in (directories-to-be-read) queue */
typedef struct _q_element {
  file_info         *dir;               /**< the file-info of the directory we're reading */
  file_flags         want,wanted;       /**< we want some things, we got some things */
  unsigned long      td,tt;             /**< int'l benchmarking: time-delta, time-total */
  void              *backend_data;      /**< backend-specific pointer */
  struct _q_element *next; } q_element; /**< next dir in to-read queue */



/** a queue */
typedef struct _q_head {
  pthread_mutex_t  qchg;                /**< mutex - frontend and backend shouldn't modify the Q at the same time */
  queue_flags      flags;               /**< busy/idle/dead etc. */
  pthread_t        scanner;             /**< the worker-thread associated with this queue */
  q_element       *anchor; } q_head;    /**< list of items in queue */



/*****************************************************************/



/* file_info   *file_info_dst(file_info *); */
file_info   *file_info_new(const char *name,const char *tag);
file_info   *file_info_ref(file_info *info);
file_info   *file_info_unref(file_info *info);
file_info   *file_info_fetch(file_info *root,const char *tag);
file_info   *file_info_insert(file_info *root,char *tag);
file_info   *file_info_copy(file_info *src);

void        *dcd_get_data(file_info *,client_funs *);
dcd         *dcd_get(file_info *,client_funs *);
dcd         *dcd_dst(file_info *,client_funs *);
dcd         *dcd_set(file_info *,client_funs *,void *);

q_element   *dir_dequeue(q_head *);
q_element   *dir_requeue(q_head *);

int          client_chain_lock(const char *);
int          client_chain_unlock(const char *);

int          client_lock(client_funs *,const char *);
int          client_unlock(client_funs *,const char *);

#define      client_get_data(a) ((a)->data)
void         client_unregister(client_funs *);
client_funs *client_register(const char *name,void *data);
client_funs *client_register_full(const char *name,void *data,
                                  int (*evt_view)(file_info *,model_event type,file_flags changed,void *),
                                  int (*dst_data)(void **),
                                  int (*set_selection)(void *));

char        *path_rewrite(const char *path);

int          path_get(q_head *qh,file_info *root,char *path,pg_mode m);

void         model_ls(file_info *root,int indent);
int          model_find_by_name(file_info *root,char *name);

int          model_evt(file_info *,model_event etype,file_flags eflags);

inline int   model_sort_fun(sort_flags,char *,file_info *,char *,file_info *);

q_head *     model_init(char *);
int          model_exit(q_head *);



/*****************************************************************/

#if defined(__cplusplus)
}
#endif

#endif
