/** @file  properties.c
 *  @brief handle properties (theme-elements)
 */

#include "evidence.h"
#include "setup.h"
#include "util.h"          /* esprintf() */
#include "file.h"
#include "properties.h"

#include <stdlib.h>        /* atof() */
#include <string.h>
#include <dlfcn.h>
#include <ctype.h>
#include <sys/stat.h>      /* S_ISDIR() */



static char *el_type[]={ "undefined", "integer", "float", "string", "pointer", "colour" };






/*****************************************************************************/
/* theme-elements ("properties") */



/**
 * split an RGB-value contained in a long into three ints
 * @param rgb  the RGB value combined into a long integer
 * @param r    where to store the red   fraction of the colour
 * @param g    where to store the green fraction of the colour
 * @param b    where to store the blue  fraction of the colour
 */
void long_to_rgb(long rgb,int *r,int *g,int *b) {
  *r=(rgb>>16)&0xff;
  *g=(rgb>>8)&0xff;
  *b=rgb&0xff; }



/**
 * canonize property name -- convert to lower case, weed out _-./
 * @return the canonized name
 * @param pn      a buffer holding the property's name
 * @param modify  if FALSE, new memory will be allocated for the canonized
 *                name, otherwise, the handed-in buffer will be modified
 */
char *theme_el_canonize_key(char *pn,int modify) {
  char *v,*r;

  if(modify)
    r=pn;
  else if(!(r=strdup(pn)))
    return NULL;

  v=r;
  while(*v) {
    if(strchr("-_/.",*v))
      memmove(v,&v[1],strlen(v));
    else {
      if(isupper(*v))
	*v=tolower(*v);
      v++; }}
  return r; }



/**
 * destruct a theme-element and remove it from a list (if list is given and element present therein)
 * @param te  the list of theme-elements, or NULL
 * @param e   the element to destroy
 */
noir_theme_el *theme_el_dst(noir_theme_el **te,noir_theme_el *e) {
  noir_theme_el       *p=NULL,*c=e;
  noir_theme_listener *l;

  if(!e||!e->key)
    return NULL;
  if(te) {
    if(*te==e)
      *te=e->next;
    else {
      do {
        p=c;
        c=c->next;
      } while(c&&(c!=e));
      if(c)    /* remove from list if even in there */
        p->next=c->next; }}

  while(e->listeners) {
    l=e->listeners;
    e->listeners=e->listeners->next;
    free(l); }

  if(e->key)
    free(e->key);
  if(e->ptr&&(e->type==ET_STR))
    free(e->ptr);

  memset(e,0,sizeof(noir_theme_el));
  free(e);

  return NULL; }



/**
 * find theme-element with property-name "key"
 * @return a theme-element
 * @param te   the list of theme-elements
 * @param key  the name of the property to look for
 */
noir_theme_el *theme_el_get(noir_theme_el **te,const char *key) {
  noir_theme_el *e;
  if(!te||!key)
    return NULL;
  e=*te;
  while(e) {
    if(!strcmp(key,e->key))
      return e;
    e=e->next; }
  return NULL; }



/**
 * get the type of a theme-element
 * @return textual description of the type (static memory caller mustn't free())
 * @param e  the theme-element we want to know the type of
 */
const char *theme_el_get_type(noir_theme_el *e) {
  if(e) {
    return el_type[e->type]; }
  return "not found"; }



/**
 * get pointer at the value of a theme-element
 * @return a pointer to either a pointer (types string, pointer) or a value
 * @param te   the list of theme-elements
 * @param key  the name of the property to get
 */
void *theme_el_get_data(noir_theme_el **te,const char *key) {
  noir_theme_el *e=theme_el_get(te,key);
  return (e?(((e->type==ET_STR)||(e->type==ET_PTR))?((void *)&e->ptr)
                                                   :((void *)&e->val))
           :NULL); }



/**
 * get character-string value from theme-element
 * @return a pointer to a C-string
 * @param te   the list of theme-elements
 * @param key  the name of the property to get
 */
char *theme_el_get_string(noir_theme_el **te,const char *key) {
  noir_theme_el *e=theme_el_get(te,key);
  return (e&&(e->type==ET_STR))?e->ptr:NULL; }



/**
 * get long integer value from theme-element
 * @return a long integer value
 * @param te   the list of theme-elements
 * @param key  the name of the property to get
 */
long theme_el_get_int(noir_theme_el **te,const char *key) {
  noir_theme_el *e=theme_el_get(te,key);
  return (e&&((e->type==ET_INT)||(e->type==ET_RGB)))?e->val:0L; }



/**
 * get floating-point value from theme-element
 * @return a floating-point value
 * @param te   the list of theme-elements
 * @param key  the name of the property to get
 */
float theme_el_get_float(noir_theme_el **te,const char *key) {
  noir_theme_el *e=theme_el_get(te,key);
  return (e&&(e->type==ET_FLT))?((float)e->val/NOIR_THEME_FLOAT_PRECISION):0.0; }



/**
 * get integer-part of a floating-point value from theme-element
 * @return an integer value
 * @param e  the theme-element
 */
int theme_el_get_float_int(noir_theme_el *e) {
  return (e&&(e->type==ET_FLT))?((int)e->val/NOIR_THEME_FLOAT_PRECISION):0.0; }



/**
 * get decimal places of a floating-point value from theme-element
 * @return an integer value
 * @param e  the theme-element
 */
int theme_el_get_float_decimals(noir_theme_el *e) {
  return (e&&(e->type==ET_FLT))?((int)e->val%NOIR_THEME_FLOAT_PRECISION):0.0; }



/**
 * get individual R, G and B values from theme-element
 * @return an error-code
 * @param te   the list of theme-elements
 * @param key  the name of the property to get
 * @param r    where to store the red   fraction of the colour
 * @param g    where to store the green fraction of the colour
 * @param b    where to store the blue  fraction of the colour
 */
int theme_el_get_rgb(noir_theme_el **te,const char *key,int *r,int *g,int *b) {
  noir_theme_el *e=theme_el_get(te,key);
  if(e&&((e->type==ET_RGB))) {
    long_to_rgb(e->val,r,g,b);
    return NOIR_SUCC; }
  return NOIR_FAIL; }



/**
 * get textual representation of the value of a theme-element
 * @return a newly allocated buffer containing the textual represenation of
 *         the property's value, or NULL
 * @param te   the list of theme-elements
 * @param k    the name of the property to get
 */
char *theme_el_get_as_string(noir_theme_el **te,const char *k) {
  noir_theme_el *e;
  char          *r=NULL;

  if(!(e=theme_el_get(te,k)))
    E(0,"no such property, \"%s\"...\n",k);
  else {
    const char *type=theme_el_get_type(e);

    switch(e->type) {
      case ET_NIL:
        esprintf(&r,"%s:%s=<nil>",k,type); break;
      case ET_INT:
        esprintf(&r,"%s:%s=%ld",k,type,theme_el_get_int(te,k)); break;
      case ET_FLT:
        esprintf(&r,"%s:%s=%lf",k,type,theme_el_get_float(te,k)); break;
      case ET_STR:
        esprintf(&r,"%s:%s=\"%s\"\n",k,type,theme_el_get_string(te,k)); break;
      case ET_PTR:
        esprintf(&r,"%s:%s=%p\n",k,type,e->ptr); break;
      case ET_RGB:
        esprintf(&r,"%s:%s=#%06lx\n",k,type,theme_el_get_int(te,k)); break;
      default:
        esprintf(&r,"%s:unknown_type",k); break; }}

  return r; }



/**
 * ensure that if a property has a range and/or a step-setting, the value adheres to this
 * called from theme_el_val() and theme_el_default()
 * @return an error-code
 * @param e  the theme-element
 */
static int theme_el_bound(noir_theme_el *e) {
  int  ret=NOIR_SUCC;
  long v;

  if(!e)
    return NOIR_FAIL;
  if(e->flags&EF_BOUNDS) {
    if((e->val<e->lo)) {
      E(0,"(\"%s\",%ld): value out of range; adjusted to %ld...\n",e->key,e->val,e->lo);
      e->val=e->lo; }
    else if((e->val>e->hi)) {
      E(0,"(\"%s\",%ld): value out of range; adjusted to %ld...\n",e->key,e->val,e->hi);
      e->val=e->hi; }
    else
      ret=NOIR_IGNORED; }
  else
    ret=NOIR_IGNORED;

  if(e->step) {
    v=((int)(e->val/e->step))*e->step;
    if(v!=e->val) {
      if(e->type==ET_FLT)
        E(1,"(\"%s\"): float value %f not a multiple of %f, adjusted to %f...\n",
          e->key,((double)e->val)/NOIR_THEME_FLOAT_PRECISION,((double)e->step)/NOIR_THEME_FLOAT_PRECISION,((double)v)/NOIR_THEME_FLOAT_PRECISION);
      else
        E(1,"(\"%s\"): integer value %ld not a multiple of %ld, adjusted to %ld...\n",
          e->key,e->val,e->step,v);
      ret=NOIR_SUCC;
      e->val=v; }}

  return ret; }



/**
 * set a property from a string. will guess type from value.
 * called from theme_el_add() and theme_el_set().
 * @return an error-code
 * @param e    the theme-element to set
 * @param val  the value (as C-string) to set the element to
 */
static int theme_el_val(noir_theme_el *e,char *val) {
  char          *l=NULL;
  long           v=0;

  e->type=ET_NIL;

  if(!val)
    e->ptr=NULL;

  else if(!*val) {
    if(!(e->ptr=strdup(val)))
      return NOIR_OOM;
    e->type=ET_STR; }

  else if(val[0]=='#') {   /* rgb */
    if((v=strtol(&val[1],&l,16))<0) {
      v=0;
      E(0,"(\"%s\"): hexadecimal value less than zero, bound to zero...\n",val);
      l=val; }
    if(*l)
      E(0,"key \"%s\" -- value \"%s\" not a valid hexadecimal RGB value?\n",e->key,val);
    else {
      e->val=v;
      e->type=ET_RGB; }}

  else {
    char *p=val;
    int   dp=0;
    int   nd=0;
    int   di=0;
    int   mi=0;

    while(*p) {    /* must ignore locale on this */
      if(!isdigit(*p)) {
        if(*p=='.')
          dp++;
        else if(*p=='-')
          mi++;
        else
          nd++; }
      else
        di++;
      p++; }

    if(di&&!nd&&(dp<2)&&(!mi||((mi==1)&&(val[0]=='-')))) {
      if(dp) {
        p=&val[mi];
        float f=(float)strtol(p,&p,10),f2;

        if(p&&*p) {
          int t=1;
          di=strlen(++p);
          f2=strtol(p,NULL,10);
          while(di--)
            t*=10;
          f+=(f2/t); }

        e->val=(long)(f*(float)NOIR_THEME_FLOAT_PRECISION);
        if(mi)
          e->val=-e->val;
/*      E(0,"val: %ld, f: %f, s: %s|%s\n",e->val,f,val,p); */
        e->type=ET_FLT; }
      else {
        e->val=strtol(val,&l,10);
        e->type=ET_INT; }}

    else if(!(e->ptr=strdup(val)))
      return NOIR_OOM;

    else
      e->type=ET_STR; }

  theme_el_bound(e);

  e->flags|=EF_MODIFIED;

  return NOIR_SUCC; }



/**
 * add a property to a list. will not check if property already present.
 * called from theme_el_set(), theme_el_default() and theme_el_listen().
 * @return an error-code
 * @param te   the list of theme-elements
 * @param key  the name of the property to add
 * @param val  the value of the property to set using theme_el_val() magic
 */
static inline int theme_el_add(noir_theme_el **te,const char *key,char *val) {
  noir_theme_el *e;

  if(!key)
    return NOIR_NODATA;

  if(!(e=malloc(sizeof(noir_theme_el))))
    goto ret;
  memset(e,0,sizeof(noir_theme_el));

  if(!(e->key=strdup(key)))
    goto ret_free_nte;

  if(!val)
    e->type=ET_NIL;
  else if(theme_el_val(e,val)==NOIR_OOM)
    goto ret_free_key;

  e->next=te?*te:NULL;
  if(te)     *te=e;

  return NOIR_SUCC;

 ret_free_key:
  free(e->key);
 ret_free_nte:
  free(e);
 ret:
  return NOIR_OOM; }



/**
 * add a property to a list, or modify it if it already exists.
 * @return an error-code
 * @param te   the list of theme-elements
 * @param key  the name of the property to add
 * @param val  the value of the property to set using theme_el_val() magic
 */
int theme_el_set(noir_theme_el **te,const char *key,char *val) {
  noir_theme_el       *e,f;
  noir_theme_listener *l;
  int                  ret=NOIR_SUCC;

  if(!te||!key)
    return NOIR_NODATA;

  if(!(e=theme_el_get(te,key)))
    return theme_el_add(te,key,val);

  if((e->ptr)&&(e->type==ET_STR)) {
    e->type=ET_NIL;
    free(e->ptr);
    e->ptr=NULL; }

  memcpy(&f,e,sizeof(noir_theme_el));
  if((ret=theme_el_val(&f,val))==NOIR_SUCC) {
    if((f.type==e->type)||(f.type==ET_NIL)||(e->type==ET_NIL)) {
/*    value=((e->type==ET_STR)||(e->type==ET_PTR))?((void *)&e->ptr):((void *)&e->val); */
      memcpy(e,&f,sizeof(noir_theme_el));
      for(l=e->listeners;l;l=l->next)
        l->listener(e->key,e->type,l->tag,l->data,te); }
    else
      E(0,"(\"%s\"): attempt to change variable type from %s to %s denied!\n",key,el_type[e->type],el_type[f.type]); }
  else
    E(0,"(\"%s\"): theme_el_val() failed: %d\n",key,ret);

  return ret; }



/**
 * set a default value for a property.
 * @return an error-code
 * @param te    the list of theme-elements
 * @param key   the name of the property to add
 * @param val   the value of the property to set using theme_el_val() magic
 * @param lo    if the type warrants this: the minimum value
 * @param hi    if the type warrants this: the maximum value
 * @param step  if the type warrants this: a step-value, 0 for default (1)
 */
int theme_el_default(noir_theme_el **te,const char *key,char *val,float lo,float hi,float step) {
  int            ret=NOIR_SUCC;
  noir_theme_el *e;

  if(!(e=theme_el_get(te,key))) {                  /* prop doesn't exist yet */
    if((ret=theme_el_add(te,key,val))!=NOIR_SUCC)  /* try to add it */
      return ret;                                  /* ...failed */
    if(!(e=theme_el_get(te,key)))                  /* get handle */
      return NOIR_FAIL; }

  if(e->type==ET_INT) {
    e->step=(int)step?step:1;
    e->flags|=EF_BOUNDS;
    e->lo=(int)lo;
    e->hi=(int)hi; }
  else if(e->type==ET_FLT) {
    e->step=(int)((step?step:1)*NOIR_THEME_FLOAT_PRECISION);
    e->flags|=EF_BOUNDS;
    e->lo=(int)(lo*NOIR_THEME_FLOAT_PRECISION);
    e->hi=(int)(hi*NOIR_THEME_FLOAT_PRECISION); }

  theme_el_bound(e);

  e->flags=e->flags&~EF_MODIFIED;

  return ret; }



/**
 * install a listener
 * @return an error-code
 * @param te        the list of theme-elements
 * @param name      the name of the listener (usually that of the plugin the listener lives in)
 * @param key       the name of the property the listener is for
 * @param listener  the function to call back
 * @param tag       an arbitrary tag -- when one listener listens for changes
 *                  on several properties, this helps distinguishing them
 *                  (without having to resort to strcmp())
 * @param data      some user-data (often the session) to give the listener
 *                  when called back
 */
int theme_el_listen(noir_theme_el **te,const char *name,const char *key,
                    property_listener listener,int tag,void *data) {
  noir_theme_el       *e;
  noir_theme_listener *l;

  if(!te||!key)
    return NOIR_NODATA;

  if(!(e=theme_el_get(te,key))) {
    int ret=theme_el_add(te,key,NULL);
    if(ret!=NOIR_SUCC) {
      E(0,"theme_el_add(\"%s\") failed: %d\n",key,ret);
      return ret; }
    if(!(e=theme_el_get(te,key))) {
      E(0,"list of theme-elements corrupted!?\n");
      return NOIR_FAIL; }}

  for(l=e->listeners;l;l=l->next)
    if(!strcmp(l->name,name)||(l->listener==listener)) {
      E(1,"%s is already listening for changes of %s...\n",
        name,key);
      return NOIR_IGNORED; }

  if(!(l=malloc(sizeof(noir_theme_listener))))
    return NOIR_OOM;

  E(2,"registering listener \"%s\" for \"%s\" (%d)...\n",name,key,e->type);

  memset(l,0,sizeof(noir_theme_listener));

  l->listener =listener;
  l->name     =name;
  l->data     =data;
  l->tag      =tag;
  l->next     =e->listeners;
  e->listeners=l;

  if(e->type!=ET_NIL)     /* call right on creation if el exists and has val */
    listener(key,e->type,tag,data,te);

  return NOIR_SUCC; }



/**
 * destall a listener (by name, or by listener-struct)
 * @return an error-code
 * @param te        the list of theme-elements
 * @param name      name of the listener, or NULL if listener-function is given
 * @param key       the property the listener was for
 * @param listener  the listener-function, or NULL if name is given
 */
int theme_el_deaf(noir_theme_el **te,const char *name,const char *key,
                  property_listener listener) {
  noir_theme_el       *e;
  noir_theme_listener *l,*p;
  int                  ret=NOIR_NOTFOUND;

  if(!te||!key)
    return NOIR_NODATA;

  if(!(e=theme_el_get(te,key)))
    return NOIR_NOTFOUND;

  for(p=NULL,l=e->listeners;l;p=l,l=l->next) {
    if((name&&!strcmp(l->name,name))||(l->listener==listener)) {
      ret=NOIR_SUCC;
      if(!p)
        e->listeners=e->listeners->next;
      else
        p->next=l->next;
      memset(l,0,sizeof(noir_theme_listener));
      free(l); }}

  return ret; }
