/** @file  vorbis.c
 *  @brief meta-data provider plug-in "vorbis"
 *
 *  get/set Ogg Vorbis tags
 */

#include "config.h"

#include "meta.h"
#include "errors.h"
#include "util.h"

#include "vcedit.h"

#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <vorbis/codec.h>
#include <vorbis/vorbisfile.h>

#include <unistd.h>

#include <stdio.h>

#define MY_NAME "Vorbis"



/**
 * write meta-data for this file
 * extractor framework cannot write meta-data, only read it
 * @return an error-code
 * @param met       the list of (all) meta-data
 * @param filename  the file to set the meta-data on
 * @param prov      the provider-entry for this plugin
 */
int meta_set_for_file(meta **met,char *filename,provider *prov) {
  vcedit_state *state=vcedit_new_state();
  vorbis_comment *vc;
  int ret=NOIR_SUCC;
  meta *r;
  FILE *src;
  FILE *dst;
  char *nn;

  if(!met||!filename)
    return NOIR_FAIL;
  if(!(r=*met))
    return NOIR_NODATA;

  if(!(nn=malloc(strlen(filename)+8)))
    return NOIR_OOM;

  if(!(src=fopen(filename,"rb"))) {
    ret=NOIR_META_FILE;
    goto free_nn; }

  strcat(strcpy(nn,filename),".evitmp");
  if(!(dst=fopen(nn,"wb"))) {
    ret=NOIR_META_FILE;
    goto close_src; }

  if(vcedit_open(state,src)<0) {
    fprintf(stderr,"Failed to open file \"%s\" as vorbis: %s\n",filename,vcedit_error(state));
    ret=NOIR_META_FILE;
    goto close_dst; }

  /* grab and clear the exisiting comments */

  vc=vcedit_comments(state);
  vorbis_comment_clear(vc);
  vorbis_comment_init(vc);

  /* add comments */
  while(r) {
    if(((r->flg==META_DO)||(r->flg==META_RW))&&(r->grp&&!strcmp(MY_NAME,r->grp))&&strcasecmp(r->key,"encoder")&&r->val&&strlen(r->val)) {
      vorbis_comment_add_tag(vc,r->key,r->val); }
    r=r->next; }

  /* write out the modified stream */
  if(vcedit_write(state,dst)<0)
    fprintf(stderr,"Failed to write comments to output file \"%s\": %s\n",filename,vcedit_error(state));
  else
    rename(nn,filename);

  /* done */
  vcedit_clear(state);

  close_dst:
  fclose(dst);
  unlink(nn);
  close_src:
  fclose(src);
  free_nn:
  free(nn);

  return ret; }



/**
 * get meta-data for this file
 * @return an error-code
 * @param ret   the list of (all) meta-data
 * @param file  the file to set the meta-data on
 * @param prov  the provider-entry for this plugin
 */
int meta_get_for_file(meta **ret,char *file,provider *prov) {
  OggVorbis_File  vf;
  FILE           *fh;

  if(!ret)
    return NOIR_FAIL;

  if(!(fh=fopen(file,"r")))
    return NOIR_META_FILE;

  if(ov_open(fh,&vf,NULL,0)<0) {
    fprintf(stderr,"Input does not appear to be an Ogg bitstream.\n");
    return NOIR_META_FORMAT; }
  else {
    char        **ptr=ov_comment(&vf,-1)->user_comments;
    char         *p,*s=NULL;
    vorbis_info  *vi=ov_info(&vf,-1);
    long          ml=ov_pcm_total(&vf,-1);
    int           hh,mm,ss;
    meta         *m=NULL;

    while(*ptr){
      if((p=strchr(*ptr,'='))) {
        *(p++)='\0';
        meta_add(ret,META_RW,META_STR,*ptr,NULL,MY_NAME,p,prov); }
      ++ptr; }

    esprintf(&s,"%d",vi->channels);
    meta_add(ret,META_RO,META_INT,"channels",NULL,MY_NAME,s,prov);

    esprintf(&s,"%ld",vi->rate);
    meta_add(ret,META_RO,META_INT,"rate",NULL,MY_NAME,s,prov);

    ml=(long)(ml/vi->rate);
    ss=(ml%60);
    ml=(long)(ml/60);
    mm=(ml%60);
    ml=(long)(ml/60);
    hh=ml;

    if(hh)
      esprintf(&s,"%d:%2d:%2d",hh,mm,ss);
    else
      esprintf(&s,"%d:%2d",mm,ss);
    meta_add(ret,META_RO,META_STR,"length",NULL,MY_NAME,s,prov);

    meta_add(ret,META_RW,META_STR,"encoder",NULL,MY_NAME,ov_comment(&vf,-1)->vendor,prov);

    meta_init_str(ret,"title","The song's title");
    meta_init_str(ret,"artist","The artist/band who created or performed this song");
    meta_init_str(ret,"album","The album this song appears on");
    meta_init_str(ret,"tracknumber","The track-number of this song (1..)");
    meta_init_str(ret,"date","The date this song was created");
    meta_dflt(ret,META_RW,META_STR,"genre","The genre of this song",MY_NAME,"Other",prov);

    if((m=meta_in_group(ret,"title",MY_NAME))&&m->val)
      meta_add(ret,META_RO,META_STR,"sortkey","If you enable 'Sort by meta-data', files will be sorted according to this field.",MY_NAME,(char *)m->val,prov);

    if(s)
      free(s); }

  ov_clear(&vf); /* takes care of fclose() */
  return NOIR_SUCC; }



/**
 * let info-dialog do the layout for us
 * @return NOIR_SUCC to do our own layout, NOIR_FAIL to have the caller do it
 */
int layoutp(void) {
  return NOIR_FAIL; }
