/** @file  setup.c
 *  @brief find data-files.
 */

#include "evidence.h"

#include <glib.h>

#include <stdlib.h>      /* getenv() */
#include <string.h>
#include <unistd.h>      /* access() */
#include <sys/stat.h>    /* stat */
#include <ctype.h>       /* isspace() */
#include <pwd.h>         /* getpwuid() */
#include "file.h"
#include "file_lcl.h"    /* file_if_local() */
#include "util.h"        /* esprintf() */



eslist *paths_setup=NULL;
eslist *paths_env=NULL;

char   *prefix=NULL;
char   *path_process=NULL,
       *path_setup=NULL;



/*****************************************************************************/



/**
 * get parent-directory of a given object.
 * @param buf - buffer to hold the resulting dirname
 * @param dir - the path and name of the fs object
 */
static char *edirname(char *buf,const char *dir) {
  char *p;
  if(!buf||!dir||!*dir||(strlen(dir)>=PATH_MAX))
    return NULL;
  strcpy(buf,dir);
  if((p=strrchr(buf,DIR_DELIMITER)))
    *p='\0';
  else
    buf[0]='\0';
  return buf; }



/*****************************************************************************/



/**
 * find 'file' in $PATH.
 * @return a buffer that must be free()d, or NULL for failure.
 * @param cwd - current working directory of session
 * @param file - name of file to find
 */
char *get_path_file(const char *cwd,const char *file) {
  char         *ret;
  const char   *pre;
  eslist       *e;

  if(!file||!*file)
    return NULL;

  for(e=paths_env;e;ESLIST_NEXT(e)) {
    pre=(char*)ESLIST_PAYLOAD(e);
    if(!strcmp(pre,".")) {
      E(4,"(%s,%s): pre <- cwd\n",cwd?cwd:"",file?file:"");
      pre=cwd; }
    if(pre) {
      ret=NULL;
      esprintf(&ret,"%s/%s",pre,file);
      if(file_exists(ret))
        return ret;
      free(ret); }}
  return NULL; }



/**
 * find 'url' in one of the setup-dirs.
 * @return a buffer that must be free()d, or NULL for failure.
 * @param url - name of file to find
 */
char *get_setup_file(const char *url) {
  char         *ret,*file;
  eslist       *e;

  if(!url||!*url)                       /* empty?  fail */
    return NULL;

  if(!((ret=file_if_local(url))&&(ret[0]!=DIR_DELIMITER))) {
    if(file_exists(url))                /* absolute (URL or other)?  use. */
      return strdup(url); }

  if(!(file=file_if_local(url)))        /* strip URL. */
    return NULL;

  for(e=paths_setup;e;ESLIST_NEXT(e)) {
    char *p=(char*)ESLIST_PAYLOAD(e);
    ret=NULL;
    esprintf(&ret,"%s%s%s",p,*p?"/":"",file);
    if((ret[0]==DIR_DELIMITER)&&file_exists(ret)) {
      return ret; }
    free(ret); }
  return NULL; }



/**
 * add a path to search-paths for setup-files.
 * called e.g. from theme.c to add theme->path.
 * @param p - path to add
 */
int paths_add(char *p) {
  int ret;
  if(!p||!*p) {
    E(0,"called with no path!?  (%p)\n",p);
    return NOIR_NODATA; }
  else if(!(ret=file_exists(p))) {
    E(0,"(\"%s\"): directory does not exist!?\n",p);
    return NOIR_NOTFOUND; }
  else if(!S_ISDIR(ret)) {
    E(0,"(\"%s\"): parameter is not a directory.\n",p);
    return NOIR_FAIL; }
  return eslist_prepend(&paths_setup,p); }



/*****************************************************************************/



/**
 * deinitialize paths sub-system.
 */
void paths_exit(void) {
  eslist_free(&paths_setup,free); }



/**
 * initialize paths sub-system.
 * determine and set search-paths.
 * @param iconset - an icon-set (if one was passed on the command-line), or NULL
 * @param cwd - the current working-directory
 * @param argv0 - the name the program was invoked under
 */
void paths_init(const char *iconset,const char *cwd,const char *argv0) {
  char        *p=NULL;
  char        *g,*e;
  const char  *path_home;
  static char  buf[PATH_MAX+1];

  /* ----------------------------------------------------------------------- */

  if((e=getenv("PATH"))&&(e=strdup(e))) {
    char *a,*b;
    a=e;
    do {
      if((b=strchr(a,':')))
        *(b++)='\0';
      eslist_append(&paths_env,strdup(a));
      E(3,"$PATH.=%s\n",a);
      a=b;
    } while(b);
    free(e); }

  /* ----------------------------------------------------------------------- */

  if(argv0[0]!=DIR_DELIMITER) {         /* relative path? */
    if(!strchr(argv0,DIR_DELIMITER)) {  /* file-name only, look in $PATH */
      char *f=get_path_file(cwd,argv0);

      if(f) {
        path_process=strdup(edirname(buf,f));
        free(f); }}

    else {                              /* relative path, make absolute */
      char *f,*p;
      esprintf(&path_process,"%s/%s",cwd,edirname(buf,argv0));

      while((f=strstr(path_process,"/./")))
        memmove(f,f+2,strlen(f+2)+1);

      while((f=strstr(path_process,"/../"))) {
        *f='\0';
        if((p=strrchr(path_process,'/')))
          memmove(p,f+3,strlen(f+3)+1);
        else
          memmove(f,f+3,strlen(f+3)+1); }}}

  else {   /* absolute path */
    path_process=strdup(edirname(buf,argv0)); }

  /* ----------------------------------------------------------------------- */

  if(!(path_home=getenv("HOME"))) {
    struct passwd *pwe=getpwuid(getuid());
    if(!pwe)
      E(0,"Go away, you don't exist!\n");
    else
      path_home=pwe->pw_dir; }

  esprintf(&path_setup,"%s/" DIR_USER,path_home);

  /* ----------------------------------------------------------------------- */

  /* default search paths */
  if(path_process&&(p=strrchr(path_process,'/'))) {
    if(!strcmp(p,"/bin")) {
      *p='\0';
      prefix=strdup(path_process);
      *p='/'; }
    else
      p=NULL; }
  if(!prefix)
    prefix=strdup(PACKAGE_PREFIX);

  { char *g2=NULL,*g3;
    g=strdup(prefix);

    /* /usr/local/share/evidence */
    g3=strdup(PACKAGE_DATA_DIR);
    eslist_prepend(&paths_setup,g3);

    /* /usr/local/share/evidence/icons/default */
    if(iconset&&(iconset[0]=='/'))
      eslist_prepend(&paths_setup,strdup(iconset));
    else {
      esprintf(&g2,"%s/" DIR_ICONS "/%s",g3,iconset?iconset:"default");
      eslist_prepend(&paths_setup,g2); }}

  if(!p)
    eslist_prepend(&paths_setup,strdup(path_process));

  free(g);
  g=NULL;

  /* /home/azundris/.evidence/icons/default */
  esprintf(&g,"%s/" DIR_ICONS "/%s",path_setup,iconset?iconset:"default");
  eslist_prepend(&paths_setup,g);

  /* /home/azundris/.evidence */
  eslist_prepend(&paths_setup,strdup(path_setup));

  /* / */
  eslist_append(&paths_setup,strdup(""));

  /* /usr/local -- needed for stuff in PREFIX/lib/evidence/... */
  eslist_append(&paths_setup,strdup(prefix));


#if 0
  /* no longer free these, they're globally available now */
  free(path_process);
  free(path_setup);
#endif

  { eslist *e;
    char   *p;
    for(e=paths_setup;e;ESLIST_NEXT(e)) {
      p=(char*)ESLIST_PAYLOAD(e);
      E(2,"SearchDir: %s\n",p&&strlen(p)?p:"."); }}}



/*****************************************************************************/
