/** @file  imlib.c
 *  @brief imlib2-loaders thumbnailer
 */

#include "config.h"

#include <stdio.h>
#include <sys/stat.h>

#include "debug.h"
#include "errors.h"
#include "util.h"
#include "file.h"            /* file_exists() */

#include "../thumbnail_private.h"

#include <stdlib.h>      /* malloc(), free() */



/*****************************************************************************/

/**
 * thumbnail an image (using Imlib2)
 *
 * @return an error-code
 * @param e    - the thumbnail struct
 */
static Imlib_Image evidence_thumb_imlib_image_preview(Evidence_Thumb *e) {
  Imlib_Image      tmp=NULL;
  Imlib_Load_Error err;
  int              try=0;
  struct stat      st;
  static char *em[]={ "","file not found","dir","may not read file",
                      "no loader for format","path too long",
                      "non-existant element in path","non-dir in path",
                      "funky path","nlinks","out of memory","out of FDs",
                      "read-only","out of space","unknown error" };

  E(3,"loading \"%s\"...\n",e->origfile);

 try_load:
  try++;
  if((tmp=imlib_load_image_with_error_return(e->origfile,&err))) {
    file_info *fi;

    imlib_context_set_image(tmp);

    if(!(tmp=evidence_thumb_imlib_standardize(e))) {
      E(0,"evidence_thumb_imlib_standardize() failed!? trying again!\n");
      goto try_load; }

    /* check if file has changed while we thumbnailed it.
       this happens if ev notices a new file before the creator is done
       writing it -- there are some pigs that take forever to write a file
       and don't place advisory locks on the file while doing it. */
    if(!stat(e->origfile,&st)&&(fi=file_info_fetch(NULL,e->origfile))&&
       fi->intrinsic&&(fi->intrinsic->size!=st.st_size)) {
      E(1,"file \"%s\" changed size (%ld->%ld) while thumbnailing!?\n",
        e->origfile,fi->intrinsic->size,st.st_size);
      imlib_free_image_and_decache();
      fi->intrinsic->size=st.st_size;
      if(try>=8) {   /* patience running out */
        E(0,"I looked at \"%s\" %d times, and it had a different size each time. Whoever is writing this file sure takes their time. I give up.\n",e->origfile,try);
        return NULL; }
      thread_sleep(1);
      goto try_load; }

    return tmp; }

  /* fix misleading error occurring when a loader in stacked loaders
     (.xcf.gz etc.) is missing */
  if((err==IMLIB_LOAD_ERROR_PATH_COMPONENT_NON_EXISTANT)&&
     file_exists(e->origfile))
    err=IMLIB_LOAD_ERROR_NO_LOADER_FOR_FILE_FORMAT;

  E((err==IMLIB_LOAD_ERROR_NO_LOADER_FOR_FILE_FORMAT)?1:0,
    "Imlib2 wouldn't load \"%s\" (%d -- \"%s\").\n",e->origfile,(int)err,
    (err<=IMLIB_LOAD_ERROR_UNKNOWN)?em[err]:"???");
  return NULL; }






/*****************************************************************************/



/**
 * initialize the thumbnailer-plugin and register it for the MIME-types and extensions it knows.
 * @return an error-code <0, or a set of features supported by the plugin
 * @param tp  an opaque thumbnailer-plugin struct
 */
int thumbnail_plugin_init(thumbnail_plugin *tp) {
  thumbnailer_register_mime(tp,"image/*");
  return (TP_IMAGE|TP_IMLIB); }



/**
 * deinitialize the thumbnailer-plugin and free any resources it may hold
 * @return an error-code
 * @param tp  an opaque thumbnailer-plugin struct
 */
int thumbnail_plugin_exit(thumbnail_plugin *tp) {
  return NOIR_IGNORED; }



/**
 * create a thumbnail
 * @return an error-code
 * @param tp    an opaque thumbnailer-plugin struct
 * @param e     the struct describing the thumbnail
 * @param data  the imlib-image (if TP_IMLIB)
 */
int thumbnail_plugin_make(thumbnail_plugin *tp,Evidence_Thumb *e,void **data) {
  Imlib_Image im=evidence_thumb_imlib_image_preview(e);
  *((Imlib_Image *)data)=im;
  return im?NOIR_SUCC:NOIR_FAIL; }



/*****************************************************************************/
